unit UPhysics2D;

{ This unit is written based on Box2D whose author is Erin Catto (http://www.gphysics.com)
  All type names follow the Delphi custom Txxx and xxx means the corresponding
  type in cpp source.

  Because versions before Delphi 2007 don't support operator overloading, so
  I write two versions of all math operations for vector and matrix, etc. But
  later I found that the version without operator overloading runs faster.
  So if you want a better performance, DEFINE BETTER_PERFORMANCE in Physics2D.inc
  which will UNDEFINE OP_OVERLOAD even if you are using Delphi 2010.

  This library supports three kinds of floats, Single(32bit), Double(64bit) and
  Extended(80bit).
       flags        EXTENDED_PRECISION        DOUBLE_PRECISION
     Extended               ON                     whatever
      Double(default)       OFF                       ON
      Single                OFF                       OFF
  There is also a flag SINGLE_PRECISION in the include file but it doesn't affect
  Float type definition.

  All assertions are ignored.

  Translator:  Qianyuan Wang(ǬԪ)
  Contact me: http://hi.baidu.com/wqyfavor
              wqyfavor@163.com
              QQ: 466798985
}

interface
{$I Physics2D.inc}
{$IFDEF D2009UP}
{$POINTERMATH ON}
{$ENDIF}

uses
   UPhysics2DTypes,
   Math,
   Classes,
   SysUtils,
   Dialogs;

type
   RGBA = array[0..3] of Single;
   TRGBA = packed record
      red, green, blue, alpha: Single;
   end;

   Tb2ContactID = record
      /// The features that intersect to form the contact point
      case Integer of
         0: (referenceEdge: UInt8; ///< The edge that defines the outward contact normal.
            incidentEdge: UInt8; ///< The edge most anti-parallel to the reference edge.
            incidentVertex: UInt8; ///< The vertex (0 or 1) on the incident edge that was clipped.
            flip: UInt8); ///< A value of 1 indicates that the reference edge is on shape2.
         1: (key: UInt32); ///< Used to quickly compare contact ids.
   end;

   /// A manifold point is a contact point belonging to a contact
   /// manifold. It holds details related to the geometry and dynamics
   /// of the contact points.
   /// The point is stored in local coordinates because CCD
   /// requires sub-stepping in which the separation is stale.
   Pb2ManifoldPoint = ^Tb2ManifoldPoint;
   Tb2ManifoldPoint = record
      localPoint1: TVector2; ///< local position of the contact point in body1
      localPoint2: TVector2; ///< local position of the contact point in body2
      separation: Float; ///< the separation of the shapes along the normal vector
      normalImpulse: Float; ///< the non-penetration impulse
      tangentImpulse: Float; ///< the friction impulse
      id: Tb2ContactID; ///< uniquely identifies a contact point between two shapes
   end;

   /// A manifold for two touching convex shapes.
   Pb2Manifold = ^Tb2Manifold;
   Tb2Manifold = record
      points: array[0..b2_maxManifoldPoints - 1] of Tb2ManifoldPoint; ///< the points of contact
      normal: TVector2;	///< the shared unit normal vector
      pointCount: Int32; ///< the number of manifold points
   end;

   /// A line segment.
   Tb2Segment = record
      p1, p2: TVector2; // The starting and ending points

      {$IFDEF OP_OVERLOAD}
      /// Ray cast against this segment with another segment.
      function TestSegment(var lambda: Float; var normal: TVector2;
         const segment: Tb2Segment; maxLambda: Float): Boolean;
      {$ENDIF}
   end;

   /// An axis aligned bounding box.
   Tb2AABB = record
      lowerBound, upperBound: TVector2; // The lower and upper vertices

      {$IFDEF OP_OVERLOAD}
    	/// Verify that the bounds are sorted.
    	function IsValid: Boolean;
      {$ENDIF}
   end;

   /// An oriented bounding box.
   Tb2OBB = record
      R: TMatrix22; ///< the rotation matrix
      center: TVector2; ///< the local centroid
      extents: TVector2; ///< the half-widths
   end;

   Tb2BodyDef = class;
   Pb2Body = ^Tb2Body;
   Tb2Body = class;
   Tb2JointDef = class;
   Pb2Joint = ^Tb2Joint;
   Tb2Joint = class;
   Pb2Contact = ^Tb2Contact;
   Tb2Contact = class;
   Pb2Shape = ^Tb2Shape;
   Tb2Shape = class;
   Tb2BroadPhase = class;
   Tb2ContactFilter = class;
   Tb2ContactListener = class;
   Tb2ContactManager = class;
   Tb2PairManager = class;
   Tb2PairCallback = class;
   Tb2Island = class;

   Pb2PolyVertices = ^Tb2PolyVertices;
   Tb2PolyVertices = array[0..b2_maxPolygonVertices - 1] of TVector2;

   //////////////////////////////////////////////////////////////
   // World

   Tb2TimeStep = record
      dt: Float; // time step
      inv_dt: Float; // inverse time step (0 if dt == 0).
	    dtRatio: Float; // dt * inv_dt0
      maxIterations: Int32;
      warmStarting, positionCorrection: Boolean;
   end;

   /// Joints and shapes are destroyed when their associated
   /// body is destroyed. Implement this listener so that you
   /// may nullify references to these joints and shapes.
   Tb2DestructionListener = class
   public
      /// Called when any joint is about to be destroyed due
      /// to the destruction of one of its attached bodies.
      procedure SayGoodbye(joint: Tb2Joint); overload; virtual; abstract;

      /// Called when any shape is about to be destroyed due
      /// to the destruction of its parent body.
      procedure SayGoodbye(shape: Tb2Shape); overload; virtual; abstract;
   end;

   /// This is called when a body's shape passes outside of the world boundary.
   Tb2BoundaryListener = class
   public
      /// This is called for each body that leaves the world boundary.
      /// @warning you can't modify the world inside this callback.
      procedure Violation(body: Tb2Body); virtual; abstract;
   end;

   Tb2DebugDrawBits = (e_shapeBit, e_jointBit, e_coreShapeBit, e_aabbBit,
      e_obbBit, e_pairBit, e_centerOfMassBit);
   Tb2DebugDrawBitsSet = set of Tb2DebugDrawBits;

   Tb2DebugDraw = class
   public
      m_drawFlags: Tb2DebugDrawBitsSet;
      m_shapeColor_Static, m_shapeColor_Sleeping, m_shapeColor_Normal,
      m_pairColor, m_aabbColor, m_obbColor, m_world_aabbColor, m_coreColor,
      m_jointLineColor: RGBA;

      (* !!! *)
      m_shapeColor_Sensor: RGBA;

      constructor Create;

      procedure DrawPolygon(const vertices: Tb2PolyVertices; vertexCount: Int32; const color: RGBA); virtual; abstract;
      procedure DrawPolygon4(const vertices: TVectorArray4; vertexCount: Int32; const color: RGBA); virtual; abstract;
      procedure DrawSolidPolygon(const vertices: Tb2PolyVertices; vertexCount: Int32; const color: RGBA); virtual; abstract;
      procedure DrawCircle(const center: TVector2; radius: Float; const color: RGBA); virtual; abstract;
      procedure DrawSolidCircle(const center, axis: TVector2; radius: Float; const color: RGBA); virtual; abstract;
      procedure DrawSegment(const p1, p2: TVector2; const color: RGBA); virtual; abstract;
      procedure DrawXForm(const xf: Tb2XForm); virtual; abstract;
   end;

   /// The world class manages all physics entities, dynamic simulation,
   /// and asynchronous queries. The world also contains efficient memory
   /// management facilities.
   Tb2World = class
   private
      procedure Solve(const step: Tb2TimeStep);
      procedure SolveTOI(const step: Tb2TimeStep);
      procedure DrawShape(shape: Tb2Shape; const xf: Tb2XForm; const color: RGBA; core: Boolean);
      procedure DrawJoint(joint: Tb2Joint);

   public
      m_lock: Boolean;

      m_broadPhase: Tb2BroadPhase;
      m_contactManager: Tb2ContactManager;

      m_bodyList: Tb2Body;
      m_groundBody: Tb2Body;

      m_jointList: Tb2Joint;

      // Do not access
      m_contactList: Tb2Contact;

      m_bodyCount,
      m_contactCount,
      m_jointCount: Int32;

      m_gravity: TVector2;
      m_allowSleep: Boolean;

      m_destructionListener: Tb2DestructionListener;
      m_boundaryListener: Tb2BoundaryListener;
      m_contactFilter: Tb2ContactFilter;
      m_contactListener: Tb2ContactListener;
      m_debugDraw: Tb2DebugDraw;

      m_inv_dt0: Float;

      m_positionIterationCount: Int32;

      m_positionCorrection: Boolean; // This is for debugging the solver.
      m_warmStarting: Boolean; // This is for debugging the solver.
      m_continuousPhysics: Boolean; // This is for debugging the solver.

      /// Construct a world object.
      /// @param worldAABB a bounding box that completely encompasses all your shapes.
      /// @param gravity the world gravity vector.
      /// @param doSleep improve performance by not simulating inactive bodies.
      constructor Create(const worldAABB: Tb2AABB; const gravity: TVector2; doSleep: Boolean);

      /// Destruct the world. All physics entities are destroyed and all heap memory is released.
      destructor Destroy; override;

      /// Create a rigid body given a definition. No reference to the definition is retained.
      /// @warning This function is locked during callbacks.
      function CreateBody(def: Tb2BodyDef; AutoFreeBodyDef: Boolean = True): Tb2Body;

      /// Destroy a rigid body given a definition. No reference to the definition
      /// is retained. This function is locked during callbacks.
      /// @warning This automatically deletes all associated shapes and joints.
      /// @warning This function is locked during callbacks.
      procedure DestroyBody(body: Tb2Body; DoFree: Boolean = True);

      /// Create a joint to constrain bodies together. No reference to the definition
      /// is retained. This may cause the connected bodies to cease colliding.
      /// @warning This function is locked during callbacks.
      function CreateJoint(def: Tb2JointDef; AutoFreeJointDef: Boolean = True): Tb2Joint;

      /// Destroy a joint. This may cause the connected bodies to begin colliding.
      /// @warning This function is locked during callbacks.
      procedure DestroyJoint(j: Tb2Joint);

      /// Take a time step. This performs collision detection, integration,
      /// and constraint solution.
      /// @param timeStep the amount of time to simulate, this should not vary.
      /// @param iterations the number of iterations to be used by the constraint solver.
      procedure Step(timeStep: Float; iterations: Int32; drawThisStep: Boolean = True);
      procedure DrawDebugData;

      /// Query the world for all shapes that potentially overlap the
      /// provided AABB. You provide a shape pointer buffer of specified
      /// size. The number of shapes found is returned.
      /// @param aabb the query box.
      /// @param shapes a user allocated shape pointer array of size maxCount (or greater).
      /// @param maxCount the capacity of the shapes array.
      /// @return the number of shapes found in aabb.
      function Query(const aabb: Tb2AABB; shapes: TList; maxCount: Int32): Int32; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}

      /// Re-filter a shape. This re-runs contact filtering on a shape.
      procedure Refilter(shape: Tb2Shape); {$IFDEF INLINE_AVAIL}inline;{$ENDIF}

      /// Perform validation of internal data structures.
      procedure Validate; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}

      function GetProxyCount: Int32; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}/// Get the number of broad-phase proxies.
      function GetPairCount: Int32; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}/// Get the number of broad-phase pairs.

      /// Change the global gravity vector.
      procedure SetGravity(const gravity: TVector2); {$IFDEF INLINE_AVAIL}inline;{$ENDIF}
      procedure WakeAllSleepingBodies;

      //////////////////////////////////////////////////////////////////////
      property DestructionListener: Tb2DestructionListener read m_destructionListener write m_destructionListener;
      property BoundaryListener: Tb2BoundaryListener read m_boundaryListener write m_boundaryListener;
      property ContactFilter: Tb2ContactFilter read m_contactFilter write m_contactFilter;
      property ContactListener: Tb2ContactListener read m_contactListener write m_contactListener;
      property DebugDraw: Tb2DebugDraw read m_debugDraw write m_debugDraw;

      property GetGroundBody: Tb2Body read m_groundBody;
      property GetBodyList: Tb2Body read m_bodyList;
      property GetJointList: Tb2Joint read m_jointList;
      property GetBodyCount: Int32 read m_bodyCount;
      property GetJointCount: Int32 read m_jointCount;
      property GetContactCount: Int32 read m_contactCount;

      property WarmStarting: Boolean read m_warmStarting write m_warmStarting;
      property PositionCorrection: Boolean read m_positionCorrection write m_warmStarting;
      property ContinuousPhysics: Boolean read m_continuousPhysics write m_continuousPhysics;
   end;

   ////////////////////////////////////////////////////
   // Contact

   /// A contact edge is used to connect bodies and contacts together
   /// in a contact graph where each body is a node and each contact
   /// is an edge. A contact edge belongs to a doubly linked list
   /// maintained in each attached body. Each contact has two contact
   /// nodes, one for each attached body.
   Pb2ContactEdge = ^Tb2ContactEdge;
   Tb2ContactEdge = record
      other: Tb2Body;	///< provides quick access to the other body attached.
      contact: Tb2Contact; ///< the contact
      prev, next: Pb2ContactEdge;
   end;

   /// This structure is used to report contact points.
   Tb2ContactPoint = record
      shape1, shape2: Tb2Shape;
      position: TVector2; ///< position in world coordinates
      velocity: TVector2;	///< velocity of point on body2 relative to point on body1 (pre-solver)
      normal: TVector2;	///< points from shape1 to shape2
      separation: Float; ///< the separation is negative when shapes are touching
      friction: Float ;		///< the combined friction coefficient
      restitution: Float ;	///< the combined restitution coefficient
      id: Tb2ContactID;			///< the contact id identifies the features in contact
   end;

   /// This structure is used to report contact point results.
   Tb2ContactResult = record
      shape1, shape2: Tb2Shape;
      position: TVector2; ///< position in world coordinates
      normal: TVector2;	///< points from shape1 to shape2
      normalImpulse: Float;	///< the normal impulse applied to body2
      tangentImpulse: Float; ///< the tangent impulse applied to body2
      id: Tb2ContactID;	///< the contact id identifies the features in contact
   end;

   Pb2ContactConstraintPoint = ^Tb2ContactConstraintPoint;
   Tb2ContactConstraintPoint = record
      localAnchor1, localAnchor2: TVector2;
      r1, r2: TVector2;

      normalImpulse, tangentImpulse, positionImpulse: Float;
      normalMass, tangentMass, equalizedMass: Float;
      separation, velocityBias: Float;
   end;

   Pb2ContactConstraint = ^Tb2ContactConstraint;
   Tb2ContactConstraint = record
      points: array[0..b2_maxManifoldPoints - 1] of Tb2ContactConstraintPoint;
      normal: TVector2;
      manifold: Pb2Manifold;
      body1, body2: Tb2Body;
      friction, restitution: Float;
      pointCount: Int32;
   end;

   /// Implement this class to provide collision filtering. In other words,
   /// you can implement this class if you want finer control over contact creation.
   Tb2ContactFilter = class
   public
      /// Return True if contact calculations should be performed between
      /// these two shapes. @warning for performance reasons this is only
      /// called when the AABBs begin to overlap.
      function ShouldCollide(shape1, shape2: Tb2Shape): Boolean; virtual;
   end;

   /// Implement this class to get collision results. You can use these results for
   /// things like sounds and game logic. You can also get contact results by
   /// traversing the contact lists after the time step. However, you might miss
   /// some contacts because continuous physics leads to sub-stepping.
   /// Additionally you may receive multiple callbacks for the same contact in a
   /// single time step.
   /// You should strive to make your callbacks efficient because there may be
   /// many callbacks per time step.
   /// @warning The contact separation is the last computed value.
   /// @warning You cannot create/destroy Box2D entities inside these callbacks.
   Tb2ContactListener = class
   public
      /// Called when a contact point is added. This includes the geometry and the forces.
      procedure Add(var point: Tb2ContactPoint); virtual;

      /// Called when a contact point persists. This includes the geometry and the forces.
      procedure Persist(var point: Tb2ContactPoint); virtual;

      /// Called when a contact point is removed. This includes the last
      /// computed geometry and forces.
      procedure Remove(var point: Tb2ContactPoint); virtual;

      /// Called after a contact point is solved.
      procedure Result(var point: Tb2ContactResult); virtual;
   end;

   /// The class manages contact between two shapes. A contact exists for each overlapping
   /// AABB in the broad-phase (except if filtered). Therefore a contact object may exist
   /// that has no contact points.
   Tb2ContactClass = class of Tb2Contact;
   Tb2Contact = class
   public
      m_flags: UInt32;

      /// The number of manifolds. This is 0 or 1 between convex shapes.
      /// This may be greater than 1 for convex-vs-concave shapes. Each
      /// manifold holds up to two contact points with a shared contact normal.
      m_manifoldCount: Int32;

      m_prev, m_next: Tb2Contact; // World pool and list pointers.
      m_node1, m_node2: Tb2ContactEdge; // Nodes for connecting bodies.
      m_shape1, m_shape2: Tb2Shape;

      // Combined friction
      m_friction: Float;
      m_restitution: Float;

      m_toi: Float;

      constructor Create; overload;
      constructor Create(shape1, shape2: Tb2Shape); overload; virtual;

      class function CreateContact(Shape1, Shape2: Tb2Shape): Tb2Contact;

      procedure Update(listener: Tb2ContactListener);
      procedure Evaluate(listener: Tb2ContactListener); virtual; abstract;

      /// Get the manifold array.
      function GetManifolds: Pb2Manifold; virtual; abstract;

      /// @return True if this contact should generate a response.
      function IsSolid: Boolean; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}

      property GetManifoldCount: Int32 read m_manifoldCount;
      property GetShape1: Tb2Shape read m_shape1;
      property GetShape2: Tb2Shape read m_shape2;
      property GetNext: Tb2Contact read m_next;
   end;

   Tb2ContactSolver = class
   public
      m_step: Tb2TimeStep;
      m_constraints: Pb2ContactConstraint;
      m_constraintCount: Int32;

	    constructor Create(const step: Tb2TimeStep; contacts: TList; contactCount: Int32);
      destructor Destroy; override;

      procedure InitVelocityConstraints(const step: Tb2TimeStep);
      procedure SolveVelocityConstraints;
      procedure FinalizeVelocityConstraints;

      function SolvePositionConstraints(baumgarte: Float): Boolean;
   end;

   Tb2NullContact = class(Tb2Contact)
   public
	    procedure Evaluate(listener: Tb2ContactListener); override;
	    function GetManifolds(): Pb2Manifold; override;
   end;

   Tb2PairCallback = class
   public
      // This should return the new pair user data. It is ok if the user data is null.
      function PairAdded(proxyUserData1, proxyUserData2: Pointer): Pointer; virtual; abstract;

      // This should free the pair's user data. In extreme circumstances, it is possible
      // this will be called with null pairUserData because the pair never existed.
      procedure PairRemoved(proxyUserData1, proxyUserData2, pairUserData: Pointer); virtual; abstract;
   end;

   // Delegate of b2World.
   Tb2ContactManager = class(Tb2PairCallback)
   public
      m_world: Tb2World;

      // This lets us provide broadphase proxy pair user data for
      // contacts that shouldn't exist.
      m_nullContact: Tb2NullContact;

      m_destroyImmediate: Boolean;

      constructor Create(world: Tb2World);

      // Implements PairCallback
      function PairAdded(proxyUserData1, proxyUserData2: Pointer): Pointer; override;

      // Implements PairCallback
      procedure PairRemoved(proxyUserData1, proxyUserData2, pairUserData: Pointer); override;

      procedure Destroy(c: Tb2Contact);
      procedure Collide;
   end;

   ////////////////////////////////////////////////////
   // Island

   Tb2Island = class
   public
      m_listener: Tb2ContactListener;

      m_bodies: TList;
      m_contacts: TList;
      m_joints: TList;

      m_bodyCount: Int32;
      m_jointCount: Int32;
      m_contactCount: Int32;

      m_bodyCapacity, m_contactCapacity, m_jointCapacity: Int32;
      m_positionIterationCount: Int32;

      constructor Create(bodyCapacity, contactCapacity, jointCapacity: Int32;
         listener: Tb2ContactListener);
      destructor Destroy; override;

      procedure Clear;
      procedure Solve(const step: Tb2TimeStep; const gravity: TVector2;
         correctPositions, allowSleep: Boolean);
      procedure SolveTOI(const subStep: Tb2TimeStep);

      procedure Add(body: Tb2Body); overload; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}
      procedure Add(contact: Tb2Contact); overload; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}
      procedure Add(joint: Tb2Joint); overload; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}

      procedure Report(constraints: Pb2ContactConstraint);
   end;

   ////////////////////////////////////////////////////
   //

   Pb2Bound = ^Tb2Bound;
   Tb2Bound = record
      value, proxyId, stabbingCount: UInt16;

      {$IFDEF OP_OVERLOAD}
	    function IsLower: Boolean; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}
      function IsUpper: Boolean; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}
      {$ENDIF}
   end;

   Pb2Proxy = ^Tb2Proxy;
   Tb2Proxy = record
      lowerBounds, upperBounds: array[0..1] of UInt16;
      overlapCount: UInt16;
      timeStamp: UInt16;
      userData: Pointer;

      {$IFDEF OP_OVERLOAD}
      function GetNext: UInt16; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}
      procedure SetNext(Next: UInt16); {$IFDEF INLINE_AVAIL}inline;{$ENDIF}
      function IsValid: Boolean; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}
      {$ENDIF}
   end;

   Tb2BoundValuesArray = array[0..1] of UInt16;
   Tb2BoundValues = record
      lowerValues, upperValues: Tb2BoundValuesArray;
   end;

   Pb2AxialBoundsArray = ^Tb2AxialBoundsArray;
   Tb2AxialBoundsArray = array[0..2 * b2_maxProxies - 1] of Tb2Bound;
   Tb2BroadPhase = class
   private
      procedure ComputeBounds(var lowerValues, upperValues: Tb2BoundValuesArray; const aabb: Tb2AABB);
      function TestOverlap(var p1, p2: Tb2Proxy): Boolean; overload;
      function TestOverlap(const b: Tb2BoundValues; var p: Tb2Proxy): Boolean; overload;

      procedure Query(var lowerIndex, upperIndex: Int32; lowerValue,
         upperValue: UInt16; var bounds: Tb2AxialBoundsArray; boundCount, axis: Int32); overload;
      procedure IncrementOverlapCount(proxyId: Int32);
      procedure IncrementTimeStamp;
   public
      m_pairManager: Tb2PairManager;
      m_proxyPool: array[0..b2_maxProxies - 1] of Tb2Proxy;
      m_freeProxy: UInt16;
      m_bounds: array[0..1] of Tb2AxialBoundsArray;
      m_queryResults: array[0..b2_maxProxies - 1] of UInt16;
      m_queryResultCount: Int32;
      m_worldAABB: Tb2AABB;
      m_quantizationFactor: TVector2;
      m_proxyCount: Int32;
      m_timeStamp: UInt16;

      constructor Create(const worldAABB: Tb2AABB; callback: Tb2PairCallback);
      destructor Destroy; override;

      // Use this to see if your proxy is in range. If it is not in range,
      // it should be destroyed. Otherwise you may get O(m^2) pairs, where m
      // is the number of proxies that are out of range.
      function InRange(const aabb: Tb2AABB): Boolean;

      // Create and destroy proxies. These call Flush first.
      function CreateProxy(const aabb: Tb2AABB; userData: Pointer): UInt16;
      procedure DestroyProxy(proxyId: Int32);

      // Call MoveProxy as many times as you like, then when you are done
      // call Commit to finalized the proxy pairs (for your time step).
      procedure MoveProxy(proxyId: Int32; const aabb: Tb2AABB);
      procedure Commit; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}

      // Get a single proxy. Returns NULL if the id is invalid.
      function GetProxy(proxyId: Int32): Pb2Proxy;

      // Query an AABB for overlapping proxies, returns the user data and
      // the count, up to the supplied maximum count.
      function Query(const aabb: Tb2AABB; userData: TList;
         maxCount: Int32): Int32; overload;

      procedure Validate;
      procedure ValidatePairs;

   {$IFDEF CLASSVAR_AVAIL}
   class var
      s_validate: Boolean;
   {$ENDIF}
   end;

   ////////////////////////////////////////////////////
   // Pair
   Pb2Pair = ^Tb2Pair;
   Tb2Pair = record
      userData: Pointer;
      proxyId1: UInt16;
      proxyId2: UInt16;
      next: UInt16;
      status: UInt16;

      {$IFDEF OP_OVERLOAD}
	    procedure SetBuffered; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}		{ status |= e_pairBuffered; }
	    procedure ClearBuffered; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}	{ status &= ~e_pairBuffered; }
	    function IsBuffered: Boolean; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}		{ return (status & e_pairBuffered) == e_pairBuffered; }

	    procedure SetRemoved; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}		{ status |= e_pairRemoved; }
	    procedure ClearRemoved; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}		{ status &= ~e_pairRemoved; }
	    function IsRemoved: Boolean; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}		{ return (status & e_pairRemoved) == e_pairRemoved; }

	    procedure SetFinal; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}		{ status |= e_pairFinal; }
	    function IsFinal: Boolean; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}		{ return (status & e_pairFinal) == e_pairFinal; }
      {$ENDIF}
   end;

   Tb2BufferedPair = record
      proxyId1: UInt16;
      proxyId2: UInt16;
   end;

   Tb2PairManager = class
   private
      function Find(proxyId1, proxyId2: Int32): Pb2Pair; overload;
      function Find(proxyId1, proxyId2: Int32; hashValue: UInt32): Pb2Pair; overload;

      function AddPair(proxyId1, proxyId2: Int32): Pb2Pair; // Returns existing pair or creates a new one.
      function RemovePair(proxyId1, proxyId2: Int32): Pointer;

      procedure ValidateBuffer;
      procedure ValidateTable;
   public
      m_broadPhase: Tb2BroadPhase;
      m_callback: Tb2PairCallback;
      m_pairs: array[0..b2_maxPairs - 1] of Tb2Pair;
      m_freePair: UInt16;
      m_pairCount: Int32;
      m_pairBuffer: array[0..b2_maxPairs - 1] of Tb2BufferedPair;
      m_pairBufferCount: Int32;
      m_hashTable: array[0..b2_tableCapacity - 1] of UInt16;

      constructor Create;

      procedure Initialize(broadPhase: Tb2BroadPhase; callback: Tb2PairCallback);
      procedure AddBufferedPair(proxyId1, proxyId2: Int32);
      procedure RemoveBufferedPair(proxyId1, proxyId2: Int32);
      procedure Commit;
   end;

   //////////////////////////////////////////////////////////////
   // Shapes

   /// This holds the mass data computed for a shape.
   Tb2MassData = record
	    mass: Float; /// The mass of the shape, usually in kilograms.
      I: Float; 	/// The rotational inertia of the shape.
      center: TVector2; /// The position of the shape's centroid relative to the shape's origin.
   end;

   /// This holds contact filtering data.
   Pb2FilterData = ^Tb2FilterData;
   Tb2FilterData = record
      categoryBits: UInt16; /// The collision category bits. Normally you would just set one bit.

      /// The collision mask bits. This states the categories that this
      /// shape would accept for collision.
      maskBits: UInt16;

      /// Collision groups allow a certain group of objects to never collide (negative)
      /// or always collide (positive). Zero means no collision group. Non-zero group
      /// filtering always wins against the mask bits.
      groupIndex: Int16;
   end;

   /// The various collision shape types supported by Box2D.
   Tb2ShapeType = (e_unknownShape = -1, e_circleShape, e_polygonShape);

   /// A shape definition is used to construct a shape. This class defines an
   /// abstract shape definition. You can reuse shape definitions safely.
   Tb2ShapeDef = class
   public
      ShapeType: Tb2ShapeType; /// Holds the shape type for down-casting.
      userData: Pointer; /// Use this to store application specify shape data.

      friction: Float; /// The shape's friction coefficient, usually in the range [0,1].
      restitution: Float; /// The shape's restitution (elasticity) usually in the range [0,1].
      density: Float; /// The shape's density, usually in kg/m^2.

      isSensor: Boolean; /// A sensor shape collects contact information but never generates a collision response.
      filter: Tb2FilterData; /// Contact filtering data.

      constructor Create;
   end;

   /// A shape is used for collision detection. Shapes are created in b2World.
   /// You can use shape for collision detection before they are attached to the world.
   /// @warning you cannot reuse shapes.
   Tb2Shape = class
   private
      destructor Destroy2; 
   protected
      procedure CreateProxy(broadPhase: Tb2BroadPhase; const xf: Tb2XForm);
      procedure DestroyProxy(broadPhase: Tb2BroadPhase);
      function Synchronize(broadPhase: Tb2BroadPhase; const xf1, xf2: Tb2XForm): Boolean;
      procedure RefilterProxy(broadPhase: Tb2BroadPhase; const xf: Tb2XForm);

      procedure UpdateSweepRadius(const center: TVector2); virtual; abstract;
   public
      m_type: Tb2ShapeType;
      m_next: Tb2Shape;
      m_body: Tb2Body;

      // Sweep radius relative to the parent body's center of mass.
      m_sweepRadius: Float;
      m_density: Float;
      m_friction: Float;
      m_restitution: Float;

      m_proxyId: UInt16;
      m_filter: Tb2FilterData;
      m_isSensor: Boolean;
      m_userData: Pointer;

      constructor Create(def: Tb2ShapeDef);
      destructor Destroy; override;

      /// Test a point for containment in this shape. This only works for convex shapes.
      /// @param xf the shape world transform.
      /// @param p a point in world coordinates.
      function TestPoint(const xf: Tb2XForm; const p: TVector2): Boolean; virtual; abstract;

      /// Perform a ray cast against this shape.
      /// @param xf the shape world transform.
      /// @param lambda returns the hit fraction. You can use this to compute
      /// the contact point p = (1 - lambda) * segment.p1 + lambda * segment.p2.
      /// @param normal returns the normal at the contact point. If there is no
      /// intersection, the normal is not set.
      /// @param segment defines the begin and end point of the ray cast.
      /// @param maxLambda a number typically in the range [0,1].
      /// @return True if there was an intersection.
      function TestSegment(const xf: Tb2XForm; var lambda: Float; var normal: TVector2;
         const segment: Tb2Segment; maxLambda: Float): Boolean; virtual; abstract;

      /// Given a transform, compute the associated axis aligned bounding box for this shape.
      /// @param aabb returns the axis aligned box.
      /// @param xf the world transform of the shape.
      procedure ComputeAABB(var aabb: Tb2AABB; const xf: Tb2XForm); virtual; abstract;

      /// Given two transforms, compute the associated swept axis aligned bounding box for this shape.
      /// @param aabb returns the axis aligned box.
      /// @param xf1 the starting shape world transform.
      /// @param xf2 the ending shape world transform.
      procedure ComputeSweptAABB(var aabb: Tb2AABB; const xf1, xf2: Tb2XForm); virtual; abstract;

      /// Compute the mass properties of this shape using its dimensions and density.
      /// The inertia tensor is computed about the local origin, not the centroid.
      /// @param massData returns the mass data for this shape.
      procedure ComputeMass(var massData: Tb2MassData); virtual; abstract;

      property GetType: Tb2ShapeType read m_type;
      property GetBody: Tb2Body read m_body;
      property GetFriction: Float read m_friction;
      property GetRestitution: Float read m_restitution;
      property GetSweepRadius: Float read m_sweepRadius;
      property IsSensor: Boolean read m_isSensor write m_isSensor;
   end;

   //////////////////////////////////////////////////////////////
   // Joints    
   Tb2JointType = (e_unknownJoint, e_revoluteJoint, e_prismaticJoint, 
      e_distanceJoint, e_pulleyJoint, e_mouseJoint, e_gearJoint, e_fixedJoint);
   Tb2LimitState = (e_inactiveLimit, e_atLowerLimit, e_atUpperLimit, e_equalLimits);

   Tb2Jacobian = record
      linear1, linear2: TVector2;
      angular1, angular2: Float;

      {$IFDEF OP_OVERLOAD}
      procedure SetZero;
      procedure SetValue(const x1, x2: TVector2; a1, a2: Float);
      function Compute(const x1, x2: TVector2; a1, a2: Float): Float; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}
      {$ENDIF}      
   end;

   /// A joint edge is used to connect bodies and joints together
   /// in a joint graph where each body is a node and each joint
   /// is an edge. A joint edge belongs to a doubly linked list
   /// maintained in each attached body. Each joint has two joint
   /// nodes, one for each attached body.
   Pb2JointEdge = ^Tb2JointEdge;
   Tb2JointEdge = record
      other: Tb2Body; ///< provides quick access to the other body attached.
      joint: Tb2Joint; ///< the joint
      prev, next: Pb2JointEdge;
   end;

   /// Joint definitions are used to construct joints.
   Tb2JointDef = class
   public            
      JointType: Tb2JointType; /// The joint type is set automatically for concrete joint types.
      userData: Pointer; /// Use this to attach application specific data to your joints.

      body1, body2: Tb2Body ; /// The attached bodies.
      collideConnected: Boolean; /// Set this flag to True if the attached bodies should collide.

      constructor Create;
   end;  

   /// The base joint class. Joints are used to constraint two bodies together in
   /// various fashions. Some joints also feature limits and motors.
   Tb2Joint = class
   protected     
      procedure InitVelocityConstraints(const step: Tb2TimeStep); virtual; abstract;
      procedure SolveVelocityConstraints(const step: Tb2TimeStep); virtual; abstract;

      // This returns True if the position errors are within tolerance.
      procedure InitPositionConstraints; virtual;
      function SolvePositionConstraints: Boolean; virtual; abstract;
   
   public
      m_type: Tb2JointType;
      m_prev, m_next: Tb2Joint;
      m_node1, m_node2: Tb2JointEdge; 
      m_body1, m_body2: Tb2Body;

      m_inv_dt: Float;
      m_islandFlag, m_collideConnected: Boolean;

      m_userData: Pointer;

      constructor Create(def: Tb2JointDef);

      /// Get the anchor point on body1 in world coordinates.
      function GetAnchor1: TVector2; virtual; abstract;

      /// Get the anchor point on body2 in world coordinates.
      function GetAnchor2: TVector2; virtual; abstract;

      /// Get the reaction force on body2 at the joint anchor.
      function GetReactionForce: TVector2; virtual; abstract;

      /// Get the reaction torque on body2.
      function GetReactionTorque: Float; virtual; abstract;

      property GetType: Tb2JointType read m_type;
      property GetBody1: Tb2Body read m_body1;
      property GetBody2: Tb2Body read m_body2;
   end;

   //////////////////////////////////////////////////////////////
   // Body

   /// A body definition holds all the data needed to construct a rigid body.
   /// You can safely re-use body definitions.
   Tb2BodyDef = class
   public
      /// You can use this to initialized the mass properties of the body.
      /// If you prefer, you can set the mass properties after the shapes
      /// have been added using b2Body::SetMassFromShapes.
      massData: Tb2MassData;

      userData: Pointer; /// Use this to store application specific body data.

      /// The world position of the body. Avoid creating bodies at the origin
      /// since this can lead to many overlapping shapes.
      position: TVector2;

      angle: Float; // The world angle of the body in radians.

      /// Linear damping is use to reduce the linear velocity. The damping parameter
      /// can be larger than 1.0f but the damping effect becomes sensitive to the
      /// time step when the damping parameter is large.
      linearDamping: Float;

      /// Angular damping is use to reduce the angular velocity. The damping parameter
      /// can be larger than 1.0f but the damping effect becomes sensitive to the
      /// time step when the damping parameter is large.
      angularDamping: Float;

      /// Set this flag to false if this body should never fall asleep. Note that
      /// this increases CPU usage.
      allowSleep: Boolean;

      isSleeping: Boolean; /// Is this body initially sleeping?
      fixedRotation: Boolean; /// Should this body be prevented from rotating? Useful for characters.

      /// Is this a fast moving body that should be prevented from tunneling through
      /// other moving bodies? Note that all bodies are prevented from tunneling through
      /// static bodies.
      /// @warning You should use this flag sparingly since it increases processing time.
      isBullet: Boolean;

      constructor Create;
   end;

   /// A rigid body.
   Tb2BodyType = (e_staticType, e_dynamicType, e_maxTypes);
   Tb2Body = class
   private
      destructor Destroy2; // Only free heap
   
      function SynchronizeShapes: Boolean;
      procedure SynchronizeTransform;

      // This is used to prevent connected bodies from colliding.
      // It may lie, depending on the collideConnected flag.
      function IsConnected(other: Tb2Body): Boolean;
      procedure Advance(t: Float);
   public
      m_flags: UInt16;
      m_type: Tb2BodyType;

      m_xf: Tb2XForm; // the body origin transform
      m_sweep: Tb2Sweep; // the swept motion for CCD

      m_linearVelocity: TVector2;
      m_angularVelocity: Float;

      m_force: TVector2;
      m_torque: Float;

      m_world: Tb2World;
      m_prev, m_next: Tb2Body;

      m_shapeList: Tb2Shape;
      m_shapeCount: Int32;

      m_jointList: Pb2JointEdge;
      m_contactList: Pb2ContactEdge;

      m_mass, m_invMass: Float;
      m_I, m_invI: Float;

      m_linearDamping: Float;
      m_angularDamping: Float;

      m_sleepTime: Float;
      m_userData: Pointer;

      constructor Create(bd: Tb2BodyDef; world: Tb2World);
      destructor Destroy; override;        

      /// Creates a shape and attach it to this body.
      /// @param shapeDef the shape definition.
      /// @warning This function is locked during callbacks.
      function CreateShape(shapeDef: Tb2ShapeDef; AutoFreeShapeDef: Boolean = True): Tb2Shape;

      /// Destroy a shape. This removes the shape from the broad-phase and
      /// therefore destroys any contacts associated with this shape. All shapes
      /// attached to a body are implicitly destroyed when the body is destroyed.
      /// @param shape the shape to be removed.
      /// @warning This function is locked during callbacks.
      procedure DestroyShape(s: Tb2Shape; DoFree: Boolean = True);

      /// Set the mass properties. Note that this changes the center of mass position.
      /// If you are not sure how to compute mass properties, use SetMassFromShapes.
      /// The inertia tensor is assumed to be relative to the center of mass.
      /// @param massData the mass properties.
      procedure SetMass(const massData: Tb2MassData);

      /// Compute the mass properties from the attached shapes. You typically call this
      /// after adding all the shapes. If you add or remove shapes later, you may want
      /// to call this again. Note that this changes the center of mass position.
      procedure SetMassFromShapes;

      /// Set the position of the body's origin and rotation (radians).
      /// This breaks any contacts and wakes the other bodies.
      /// @param position the new world position of the body's origin (not necessarily
      /// the center of mass).
      /// @param angle the new world rotation angle of the body in radians.
      /// @return false if the movement put a shape outside the world. In this case the
      /// body is automatically frozen.
      function SetXForm(const position: TVector2; angle: Float): Boolean;

      /// Set the linear velocity of the center of mass.
      /// @param v the new linear velocity of the center of mass.
      procedure SetLinearVelocity(const v: TVector2); {$IFDEF INLINE_AVAIL}inline;{$ENDIF}

      /// Set the angular velocity.
      /// @param omega the new angular velocity in radians/second.
      procedure SetAngularVelocity(omega: Float); {$IFDEF INLINE_AVAIL}inline;{$ENDIF}

      /// Apply a force at a world point. If the force is not
      /// applied at the center of mass, it will generate a torque and
      /// affect the angular velocity. This wakes up the body.
      /// @param force the world force vector, usually in Newtons (N).
      /// @param point the world position of the point of application.
      procedure ApplyForce(const force, point: TVector2);

      /// Apply a torque. This affects the angular velocity
      /// without affecting the linear velocity of the center of mass.
      /// This wakes up the body.
      /// @param torque about the z-axis (out of the screen), usually in N-m.
      procedure ApplyTorque(torque: Float);

      /// Apply an impulse at a point. This immediately modifies the velocity.
      /// It also modifies the angular velocity if the point of application
      /// is not at the center of mass. This wakes up the body.
      /// @param impulse the world impulse vector, usually in N-seconds or kg-m/s.
      /// @param point the world position of the point of application.
      procedure ApplyImpulse(const impulse, point: TVector2);

      /// Get the world coordinates of a point given the local coordinates.
      /// @param localPoint a point on the body measured relative the the body's origin.
      /// @return the same point expressed in world coordinates.
      function GetWorldPoint(const localPoint: TVector2): TVector2; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}

      /// Get the world coordinates of a vector given the local coordinates.
      /// @param localVector a vector fixed in the body.
      /// @return the same vector expressed in world coordinates.
      function GetWorldVector(const localVector: TVector2): TVector2; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}

      /// Gets a local point relative to the body's origin given a world point.
      /// @param a point in world coordinates.
      /// @return the corresponding local point relative to the body's origin.
      function GetLocalPoint(const worldPoint: TVector2): TVector2; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}

      /// Gets a local vector given a world vector.
      /// @param a vector in world coordinates.
      /// @return the corresponding local vector.
      function GetLocalVector(const worldVector: TVector2): TVector2; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}

      /// Get the world linear velocity of a world point attached to this body.
      /// @param a point in world coordinates.
      /// @return the world velocity of a point.
      function GetLinearVelocityFromWorldPoint(const worldPoint: TVector2): TVector2; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}

      /// Get the world velocity of a local point.
      /// @param a point in local coordinates.
      /// @return the world velocity of a point.
      function GetLinearVelocityFromLocalPoint(const localPoint: TVector2): TVector2; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}

      /// Is this body treated like a bullet for continuous collision detection?
      function IsBullet: Boolean; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}

      /// Should this body be treated like a bullet for continuous collision detection?
      procedure SetBullet(flag: Boolean);

      /// Is this body static (immovable)?
      function IsStatic: Boolean; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}

      /// Is this body dynamic (movable)?
      function IsDynamic: Boolean; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}

      /// Is this body frozen?
      function IsFrozen: Boolean; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}

      /// Is this body sleeping (not simulating).
      function IsSleeping: Boolean; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}

      /// You can disable sleeping on this body.
      procedure AllowSleeping(flag: Boolean);

      /// Wake up this body so it will begin simulating.
      procedure WakeUp;

      /// Put this body to sleep so it will stop simulating.
      /// This also sets the velocity to zero.
      procedure PutToSleep;

      ////////////////////////////////////
      property GetAngle: Float read m_sweep.a;
      property GetPosition: TVector2 read m_xf.position;
      property GetWorldCenter: TVector2 read m_sweep.c;
      property GetLocalCenter: TVector2 read m_sweep.localCenter;
      property GetLinearVelocity: TVector2 read m_linearVelocity;
      property GetAngularVelocity: Float read m_angularVelocity;
      property GetMass: Float read m_mass;
      property GetInertia: Float read m_I;
      property GetShapeList: Tb2Shape read m_shapeList;
      property GetJointList: Pb2JointEdge read m_jointList;
      property GetNext: Tb2Body read m_next;
      property UserData: Pointer read m_userData write m_userData;
      property GetWorld: Tb2World read m_world;
   end;

   ///////////////////////////////////////////////
   // Specific implementations

   Tb2CircleDef = class(Tb2ShapeDef)
   public
      localPosition: TVector2;
      radius: Float;

      constructor Create;
   end;

   Tb2CircleShape = class(Tb2Shape)
   public
      m_radius: Float;
      m_localPosition: TVector2; // Local position in parent body

      constructor Create(def: Tb2ShapeDef);
      procedure UpdateSweepRadius(const center: TVector2); override;

      function TestPoint(const transform: Tb2XForm; const p: TVector2): Boolean; override;
      function TestSegment(const xf: Tb2XForm; var lambda: Float; var normal: TVector2;
         const segment: Tb2Segment; maxLambda: Float): Boolean; override;
      procedure ComputeAABB(var aabb: Tb2AABB; const xf: Tb2XForm); override;
      procedure ComputeSweptAABB(var aabb: Tb2AABB; const xf1, xf2: Tb2XForm); override;
      procedure ComputeMass(var massData: Tb2MassData); override;

      /// Get the radius of this circle.
      property GetRadius: Float read m_radius;
   end;

   /// Convex polygon. The vertices must be in CCW order for a right-handed
   /// coordinate system with the z-axis coming out of the screen.
   Tb2PolygonDef = class(Tb2ShapeDef)
   public
      /// The polygon vertices in local coordinates.
      vertices: Tb2PolyVertices;
      vertexCount: Int32;

      constructor Create;

      /// Build vertices to represent an axis-aligned box.
      /// @param hx the half-width.
      /// @param hy the half-height.
      procedure SetAsBox(hx, hy: Float); overload;

      /// Build vertices to represent an oriented box.
      /// @param hx the half-width.
      /// @param hy the half-height.
      /// @param center the center of the box in local coordinates.
      /// @param angle the rotation of the box in local coordinates.
      procedure SetAsBox(hx, hy: Float; const center: TVector2; angle: Float); overload;
   end;

   /// A convex polygon.
   Tb2PolygonShape = class(Tb2Shape)
   private
      pm_vertices, pm_normals, pm_coreVertices: PVector2;
   public
      m_centroid: TVector2; // Local position of the polygon centroid.
      m_obb: Tb2OBB; // The oriented bounding box relative to the parent body.

      m_vertices: Tb2PolyVertices;
      m_normals: Tb2PolyVertices;
      m_coreVertices: Tb2PolyVertices;
      m_vertexCount: Int32;

      constructor Create(const def: Tb2ShapeDef);
	    procedure UpdateSweepRadius(const center: TVector2); override;

      function TestPoint(const transform: Tb2XForm; const p: TVector2): Boolean; override;
      function TestSegment(const xf: Tb2XForm; var lambda: Float; var normal: TVector2;
         const segment: Tb2Segment; maxLambda: Float): Boolean; override;
      procedure ComputeAABB(var aabb: Tb2AABB; const xf: Tb2XForm); override;
      procedure ComputeSweptAABB(var aabb: Tb2AABB; const xf1, xf2: Tb2XForm); override;
      procedure ComputeMass(var massData: Tb2MassData); override;

      /// Get the first vertex and apply the supplied transform.
      function GetFirstVertex(const xf: Tb2XForm): TVector2;

      /// Get the centroid and apply the supplied transform.
      function Centroid(const xf: Tb2XForm): TVector2;

      /// Get the support point in the given world direction.
      /// Use the supplied transform.
      function Support(const xf: Tb2XForm; const d: TVector2): TVector2;

      property GetVertices: PVector2 read pm_vertices;

      /// Get the core vertices in local coordinates. These vertices
      /// represent a smaller polygon that is used for time of impact computations.
      property GetCoreVertices: PVector2 read pm_coreVertices;

      /// Get the edge normal vectors. There is one for each vertex.
      property GetNormals: PVector2 read pm_normals;

      property GetVertexCount: Integer read m_vertexCount;
   end;

   ////////////////////////////////////////////////////////////
   Tb2CircleContact = class(Tb2Contact)
   public
      m_manifold: Tb2Manifold;

      constructor Create(shape1, shape2: Tb2Shape); override;

      procedure Evaluate(listener: Tb2ContactListener); override;
      function GetManifolds: Pb2Manifold; override;
   end;

   Tb2PolyAndCircleContact = class(Tb2Contact)
   public
      m_manifold: Tb2Manifold;

      constructor Create(shape1, shape2: Tb2Shape); override;

      procedure Evaluate(listener: Tb2ContactListener); override;
      function GetManifolds: Pb2Manifold; override;
   end;

   Tb2PolygonContact = class(Tb2Contact)
   public
      m_manifold: Tb2Manifold;

      constructor Create(shape1, shape2: Tb2Shape); override;

      procedure Evaluate(listener: Tb2ContactListener); override;
      function GetManifolds: Pb2Manifold; override;
   end;

   ////////////////////////////////////////////////////////////

   /// Distance joint definition. This requires defining an
   /// anchor point on both bodies and the non-zero length of the
   /// distance joint. The definition uses local anchor points
   /// so that the initial configuration can violate the constraint
   /// slightly. This helps when saving and loading a game.
   /// @warning Do not use a zero or short length.
   Tb2DistanceJointDef = class(Tb2JointDef)
   public
      localAnchor1: TVector2; /// The local anchor point relative to body1's origin.
      localAnchor2: TVector2; /// The local anchor point relative to body2's origin.

      length : Float; /// The equilibrium length between the anchor points.
      frequencyHz: Float; /// The response speed.
      dampingRatio: Float; /// The damping ratio. 0 = no damping, 1 = critical damping.

      constructor Create;
      procedure Initialize(body1, body2: Tb2Body; const anchor1, anchor2: TVector2);
   end;

   /// A distance joint constrains two points on two bodies
   /// to remain at a fixed distance from each other. You can view
   /// this as a massless, rigid rod.
   Tb2DistanceJoint = class(Tb2Joint)
   public
      m_localAnchor1, m_localAnchor2, m_u: TVector2;
      m_frequencyHz, m_dampingRatio: Float;
      m_gamma, m_bias, m_impulse,
      m_mass, // effective mass for the constraint.
      m_length: Float;

      constructor Create(def: Tb2DistanceJointDef);

      function GetAnchor1: TVector2; override;
      function GetAnchor2: TVector2; override;

      function GetReactionForce: TVector2; override;
      function GetReactionTorque: Float; override;

      procedure InitVelocityConstraints(const step: Tb2TimeStep); override;
      procedure SolveVelocityConstraints(const step: Tb2TimeStep); override;
      function SolvePositionConstraints: Boolean; override;
   end;

   /// Prismatic joint definition. This requires defining a line of
   /// motion using an axis and an anchor point. The definition uses local
   /// anchor points and a local axis so that the initial configuration
   /// can violate the constraint slightly. The joint translation is zero
   /// when the local anchor points coincide in world space. Using local
   /// anchors and a local axis helps when saving and loading a game.
   Tb2PrismaticJointDef = class(Tb2JointDef)
   public
      localAnchor1: TVector2;
      localAnchor2: TVector2;

      localAxis1: TVector2; /// The local translation axis in body1.
      referenceAngle: Float; /// The constrained angle between the bodies: body2_angle - body1_angle.

      enableLimit: Boolean; /// Enable/disable the joint limit.

      lowerTranslation: Float; /// The lower translation limit, usually in meters.
      upperTranslation: Float; /// The upper translation limit, usually in meters.

      enableMotor: Boolean; /// Enable/disable the joint motor.
      maxMotorForce: Float; /// The maximum motor torque, usually in N-m.
      motorSpeed: Float; /// The desired motor speed in radians per second.

      constructor Create;
      procedure Initialize(body1, body2: Tb2Body; const anchor, axis: TVector2); // world anchor and world axis
   end;

   /// A prismatic joint. This joint provides one degree of freedom: translation
   /// along an axis fixed in body1. Relative rotation is prevented. You can
   /// use a joint limit to restrict the range of motion and a joint motor to
   /// drive the motion or to model joint friction.
   Tb2PrismaticJoint = class(Tb2Joint)
   public
      m_localAnchor1, m_localAnchor2, m_localXAxis1, m_localYAxis1: TVector2;
      m_refAngle: Float;
      m_linearJacobian: Tb2Jacobian;

      m_linearMass: Float; // effective mass for point-to-line constraint.
      m_force: Float;
      m_angularMass: Float; // effective mass for angular constraint.
      m_torque: Float;

      m_motorJacobian: Tb2Jacobian;
      m_motorMass, // effective mass for motor/limit translational constraint.
      m_motorForce,
      m_limitForce,
      m_limitPositionImpulse: Float;

      m_lowerTranslation, m_upperTranslation: Float;
      m_maxMotorForce, m_motorSpeed: Float;

      m_enableLimit: Boolean;
      m_enableMotor: Boolean;

      m_limitState: Tb2LimitState;

      constructor Create(def: Tb2PrismaticJointDef);

      function GetAnchor1: TVector2; override;
      function GetAnchor2: TVector2; override;

      function GetReactionForce: TVector2; override;
      function GetReactionTorque: Float; override;

      procedure InitVelocityConstraints(const step: Tb2TimeStep); override;
      procedure SolveVelocityConstraints(const step: Tb2TimeStep); override;
      function SolvePositionConstraints: Boolean; override;

      /// Get the current joint translation, usually in meters.
      function GetJointTranslation: Float;

      /// Get the current joint translation speed, usually in meters per second.
      function GetJointSpeed: Float;

      /// Set the joint limits, usually in meters.
      procedure SetLimits(lower, upper: Float);

      property GetMotorSpeed: Float read m_motorSpeed; // usually in meters per second.
      property GetMotorForce: Float read m_motorForce; // usually in N.
      property LimitEnabled: Boolean read m_enableLimit write m_enableLimit;
      property GetLowerLimit: Float read m_lowerTranslation;
      property GetUpperLimit: Float read m_upperTranslation;
      property MotorEnabled: Boolean read m_enableMotor write m_enableMotor;
      property MotorSpeed: Float read m_motorSpeed write m_motorSpeed;
      property MaxMotorForce: Float read m_maxMotorForce write m_maxMotorForce;
   end;

   /// Mouse joint definition. This requires a world target point, tuning parameters, and the time step.
   Tb2MouseJointDef = class(Tb2JointDef)
   public
      /// The initial world target point. This is assumed to coincide with the body anchor initially.
      target: TVector2;

      /// The maximum constraint force that can be exerted
      /// to move the candidate body. Usually you will express
      /// as some multiple of the weight (multiplier * mass * gravity).
      maxForce: Float;

      frequencyHz: Float; /// The response speed.
      dampingRatio: Float; /// The damping ratio. 0 = no damping, 1 = critical damping.
      timeStep: Float; /// The time step used in the simulation.

      constructor Create;
   end;

   /// A mouse joint is used to make a point on a body track a
   /// specified world point. This a soft constraint with a maximum
   /// force. This allows the constraint to stretch and without
   /// applying huge forces.
   Tb2MouseJoint = class(Tb2Joint)
   public
      m_localAnchor, m_target, m_impulse: TVector2;
      m_mass: TMatrix22; // effective mass for point-to-point constraint.
      m_C: TVector2; // position error
      m_maxForce: Float;
      m_beta: Float; // bias factor
      m_gamma: Float; // softness

      constructor Create(def: Tb2MouseJointDef);

      function GetAnchor1: TVector2; override;
      function GetAnchor2: TVector2; override;

      function GetReactionForce: TVector2; override;
      function GetReactionTorque: Float; override;

      procedure InitVelocityConstraints(const step: Tb2TimeStep); override;
      procedure SolveVelocityConstraints(const step: Tb2TimeStep); override;
      function SolvePositionConstraints: Boolean; override;

	    procedure SetTarget(const target: TVector2); /// Use this to update the target point.
   end;

   /// Pulley joint definition. This requires two ground anchors,
   /// two dynamic body anchor points, max lengths for each side,
   /// and a pulley ratio.
   Tb2PulleyJointDef = class(Tb2JointDef)
   public
      groundAnchor1: TVector2; /// The first ground anchor in world coordinates. This point never moves.
      groundAnchor2: TVector2; /// The second ground anchor in world coordinates. This point never moves.
      localAnchor1: TVector2; /// The local anchor point relative to body1's origin.
      localAnchor2: TVector2; /// The local anchor point relative to body2's origin.

      length1: Float; /// The a reference length for the segment attached to body1.
      maxLength1: Float; /// The maximum length of the segment attached to body1.

      length2: Float; /// The a reference length for the segment attached to body2.
      maxLength2: Float; /// The maximum length of the segment attached to body2.

      ratio: Float; /// The pulley ratio, used to simulate a block-and-tackle.

      constructor Create;
      /// Initialize the bodies, anchors, lengths, max lengths, and ratio using the world anchors.
      procedure Initialize(body1, body2: Tb2Body; const groundAnchor1, groundAnchor2,
        anchor1, anchor2: TVector2; ratio: Float);
   end;

   /// The pulley joint is connected to two bodies and two fixed ground points.
   /// The pulley supports a ratio such that:
   /// length1 + ratio * length2 <= constant
   /// Yes, the force transmitted is scaled by the ratio.
   /// The pulley also enforces a maximum length limit on both sides. This is
   /// useful to prevent one side of the pulley hitting the top.
   Tb2PulleyJoint = class(Tb2Joint)
   public
      m_ground: Tb2Body;
      m_groundAnchor1, m_groundAnchor2, m_localAnchor1, m_localAnchor2: TVector2;
      m_u1, m_u2: TVector2;

      m_constant, m_ratio: Float;
      m_maxLength1, m_maxLength2: Float;

      m_pulleyMass, m_limitMass1, m_limitMass2: Float; // Effective masses
      m_force, m_limitForce1, m_limitForce2: Float; // Impulses for accumulation/warm starting.

      // Position impulses for accumulation.
      m_positionImpulse, m_limitPositionImpulse1, m_limitPositionImpulse2: Float;

      m_state, m_limitState1, m_limitState2: Tb2LimitState;

      constructor Create(def: Tb2PulleyJointDef);

      function GetAnchor1: TVector2; override;
      function GetAnchor2: TVector2; override;

      function GetReactionForce: TVector2; override;
      function GetReactionTorque: Float; override;

      procedure InitVelocityConstraints(const step: Tb2TimeStep); override;
      procedure SolveVelocityConstraints(const step: Tb2TimeStep); override;
      function SolvePositionConstraints: Boolean; override;

      function GetLength1: Float;
      function GetLength2: Float;

      function GetGroundAnchor1: TVector2;
      function GetGroundAnchor2: TVector2;

      property GetRatio: Float read m_ratio;
   end;

   /// Revolute joint definition. This requires defining an
   /// anchor point where the bodies are joined. The definition
   /// uses local anchor points so that the initial configuration
   /// can violate the constraint slightly. You also need to
   /// specify the initial relative angle for joint limits. This
   /// helps when saving and loading a game.
   /// The local anchor points are measured from the body's origin
   /// rather than the center of mass because:
   /// 1. you might not know where the center of mass will be.
   /// 2. if you add/remove shapes from a body and recompute the mass,
   ///    the joints will be broken.
   Tb2RevoluteJointDef = class(Tb2JointDef)
   public
      localAnchor1: TVector2; /// The local anchor point relative to body1's origin.
      localAnchor2: TVector2; /// The local anchor point relative to body2's origin.
      referenceAngle: Float; /// The body2 angle minus body1 angle in the reference state (radians).

      enableLimit: Boolean; /// A flag to enable joint limits.
      lowerAngle, upperAngle: Float; /// The lower(upper) angle for the joint limit (radians).

      enableMotor: Boolean; /// A flag to enable the joint motor.
      motorSpeed: Float; /// The desired motor speed. Usually in radians per second.
      maxMotorTorque: Float; /// The maximum motor torque used to achieve the desired motor speed. Usually in N-m.

      constructor Create;
    	/// Initialize the bodies, anchors, and reference angle using the world anchor.
    	procedure Initialize( body1, body2: Tb2Body; const anchor: TVector2);
   end;

   /// A revolute joint constrains to bodies to share a common point while they
   /// are free to rotate about the point. The relative rotation about the shared
   /// point is the joint angle. You can limit the relative rotation with
   /// a joint limit that specifies a lower and upper angle. You can use a motor
   /// to drive the relative rotation about the shared point. A maximum motor torque
   /// is provided so that infinite forces are not generated.
   Tb2RevoluteJoint = class(Tb2Joint)
   public
      m_localAnchor1, m_localAnchor2: TVector2; // relative

      m_pivotForce: TVector2;
      m_motorForce, m_limitForce, m_limitPositionImpulse: Float;

      m_pivotMass: TMatrix22;	// effective mass for point-to-point constraint.
      m_motorMass: Float;	// effective mass for motor/limit angular constraint.

      m_enableMotor: Boolean;
      m_maxMotorTorque, m_motorSpeed: Float;

      m_enableLimit: Boolean;
      m_referenceAngle, m_lowerAngle, m_upperAngle: Float;

      m_limitState: Tb2LimitState;

      constructor Create(def: Tb2RevoluteJointDef);

      function GetAnchor1: TVector2; override;
      function GetAnchor2: TVector2; override;

      function GetReactionForce: TVector2; override;
      function GetReactionTorque: Float; override;

      procedure InitVelocityConstraints(const step: Tb2TimeStep); override;
      procedure SolveVelocityConstraints(const step: Tb2TimeStep); override;
      function SolvePositionConstraints: Boolean; override;

      function GetJointAngle: Float;
      function GetJointSpeed: Float; /// Get the current joint angle speed in radians per second.

      property LimitEnabled: Boolean read m_enableLimit write m_enableLimit;
      property LowerLimit: Float read m_lowerAngle write m_lowerAngle;
      property UpperLimit: Float read m_upperAngle write m_upperAngle;
      property MotorEnabled: Boolean read m_enableMotor write m_enableMotor;
      property MotorSpeed: Float read m_motorSpeed write m_motorSpeed; /// Get the motor speed in radians per second.
      property MaxMotorTorque: Float read m_maxMotorTorque write m_maxMotorTorque;
      property MotorTorque: Float read m_motorForce write m_motorForce;
   end;

   /// Gear joint definition. This definition requires two existing
   /// revolute or prismatic joints (any combination will work).
   /// The provided joints must attach a dynamic body to a static body.
   Tb2GearJointDef = class(Tb2JointDef)
   public
      joint1: Tb2Joint; /// The first revolute/prismatic joint attached to the gear joint.
      joint2: Tb2Joint; /// The second revolute/prismatic joint attached to the gear joint.
      ratio: Float; /// The gear ratio.

      constructor Create;
   end;

   /// A gear joint is used to connect two joints together. Either joint
   /// can be a revolute or prismatic joint. You specify a gear ratio
   /// to bind the motions together:
   /// coordinate1 + ratio * coordinate2 = constant
   /// The ratio can be negative or positive. If one joint is a revolute joint
   /// and the other joint is a prismatic joint, then the ratio will have units
   /// of length or units of 1/length.
   /// @warning The revolute and prismatic joints must be attached to
   /// fixed bodies (which must be body1 on those joints).
   Tb2GearJoint = class(Tb2Joint)
   public
      m_ground1, m_ground2: Tb2Body;

      // One of these is nil.
      m_revolute1: Tb2RevoluteJoint;
      m_prismatic1: Tb2PrismaticJoint;

      // One of these is nil.
      m_revolute2: Tb2RevoluteJoint;
      m_prismatic2: Tb2PrismaticJoint;

      m_groundAnchor1, m_groundAnchor2 :TVector2;
      m_localAnchor1, m_localAnchor2: TVector2;

      m_J: Tb2Jacobian;

      m_constant, m_ratio: Float;
      m_mass: Float; // Effective mass
      m_force: Float; // Impulse for accumulation/warm starting.

      constructor Create(def: Tb2GearJointDef);

      function GetAnchor1: TVector2; override;
      function GetAnchor2: TVector2; override;

      function GetReactionForce: TVector2; override;
      function GetReactionTorque: Float; override;

      procedure InitVelocityConstraints(const step: Tb2TimeStep); override;
      procedure SolveVelocityConstraints(const step: Tb2TimeStep); override;
      function SolvePositionConstraints: Boolean; override;

      property GetRatio: Float read m_ratio;
   end;

   /// FixedJoint: Attaches two bodies rigidly together
   Tb2FixedJointDef = class(Tb2JointDef)
   public
      constructor Create;
	    procedure Initialize(body1, body2: Tb2Body); /// Initialize the bodies.
   end;

   /// A fixed joint constrains all degrees of freedom between two bodies
   /// Author: Jorrit Rouwe
   /// See: www.jrouwe.nl/fixedjoint/ for more info
   Tb2FixedJoint = class(Tb2Joint)
   private
	    procedure CalculateMC; // Get effective constraint mass
   public
      // Configured state for bodies
      m_dp: TVector2;		//< Distance between body->GetXForm().position between the two bodies at rest in the reference frame of body1
      m_a: Float;		//< Angle between the bodies at rest
      m_R0: TMatrix22;		//< Rotation matrix of m_a

      // State for solving
      m_inv_dt: Float;	//< Stored 1/dt
      m_d: TVector2;			//< Distance between center of masses for this time step (when the shapes of the bodies change, their local centers can change so we derive this from m_dp every frame)
      m_a1: Float;			//< Stored angle of body 1 (a1) to determine if it changed
      m_c, m_s: Float;		//< cos(a1) and sin(a1)
      m_Ax, m_Ay: Float;	//< A = d/dt (R(a1) d)
      m_mc: array[0..2, 0..2] of Float;	//< Effective constraint mass

      // State after solving
      m_lambda: array[0..2] of Float;	//< Accumulated lambdas for warm starting and returning constraint force

      constructor Create(def: Tb2FixedJointDef);

      function GetAnchor1: TVector2; override;
      function GetAnchor2: TVector2; override;

      function GetReactionForce: TVector2; override;
      function GetReactionTorque: Float; override;

      procedure InitVelocityConstraints(const step: Tb2TimeStep); override;
      procedure SolveVelocityConstraints(const step: Tb2TimeStep); override;
      function SolvePositionConstraints: Boolean; override;
   end;


//////////////////////////////////
procedure b2CollideCircles(var manifold: Tb2Manifold;
   circle1, circle2: Tb2CircleShape; const xf1, xf2: Tb2XForm);
procedure b2CollidePolygonAndCircle(var manifold: Tb2Manifold;
   polygon: Tb2PolygonShape; circle: Tb2CircleShape; const xf1, xf2: Tb2XForm);
procedure b2CollidePolygons(var manifold: Tb2Manifold;
   polyA, polyB: Tb2PolygonShape; xfA, xfB: Tb2XForm);

function b2TimeOfImpact(shape1, shape2: Tb2Shape; const sweep1, sweep2: Tb2Sweep): Float;

function b2Distance(var x1, x2: TVector2; circle1, circle2: Tb2CircleShape;
   const xf1, xf2: Tb2XForm): Float; overload;
function b2Distance(var x1, x2: TVector2; poly: Tb2PolygonShape;
   circle: Tb2CircleShape; const xf1, xf2: Tb2XForm): Float; overload;
function b2Distance(var x1, x2: TVector2; circle: Tb2CircleShape;
   poly: Tb2PolygonShape; const xf1, xf2: Tb2XForm): Float; overload; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}
function b2Distance(var x1, x2: TVector2; poly1, poly2: Tb2PolygonShape;
   const xf1, xf2: Tb2XForm): Float; overload;
function b2Distance(var x1, x2: TVector2; shape1, shape2: Tb2Shape;
   const xf1, xf2: Tb2XForm): Float; overload;

/////////////////// Color functions //////
function MakeColor(r, g, b: Single; a: Single = 1.0): RGBA;

var
   g_GJK_Iterations: Int32;

implementation
var
   b2_defaultFilter: Tb2ContactFilter;
{$IFNDEF CLASSVAR_AVAIL}
   b2BroadPhase_s_validate: Boolean;
{$ENDIF}

const
   // Tb2Pair.status
   e_pairBuffered = 1;
   e_pairRemoved = 2;
   e_pairFinal = 4;

   // Tb2Contact.m_flags
   e_nonSolidFlag	= 1;
   e_slowFlag	= 2;
   e_contact_islandFlag	= 4;
   e_toiFlag = 8;

   // Tb2Body.m_flags
   e_frozenFlag	= $0002;
   e_body_islandFlag	= $0004;
   e_sleepFlag = $0008;
   e_allowSleepFlag	= $0010;
   e_bulletFlag	= $0020;
   e_fixedRotationFlag = $0040;

function MakeColor(r, g, b: Single; a: Single = 1.0): RGBA;
begin
   Result[0] := r;
   Result[1] := g;
   Result[2] := b;
   Result[3] := a;
end;

//////////// Implements <b2Contact.cpp> InitializeRegisters and AddType
type
   TContactCreateRecord = record
      ClassType: Tb2ContactClass;
      Primary: Boolean;
   end;

const
   ContactCreateRecords: array[e_circleShape..e_polygonShape,
      e_circleShape..e_polygonShape] of TContactCreateRecord = (
      ((ClassType: Tb2CircleContact; Primary: True),
       (ClassType: Tb2PolyAndCircleContact; Primary: False)),
      ((ClassType: Tb2PolyAndCircleContact; Primary: True),
       (ClassType: Tb2PolygonContact; Primary: True)));

{$IFNDEF OP_OVERLOAD}
// Record methods

function TestSegment(const Self: Tb2Segment; var lambda: Float; 
   var normal: TVector2; const segment: Tb2Segment; maxLambda: Float): Boolean;
const
   k_slop = 100.0 * FLT_EPSILON;
var
   s, r, d, n, b: TVector2;
   denom, a, mu2: Float;
begin
   with Self do
   begin
      s := segment.p1;
      r := Subtract(segment.p2, s);
      d := Subtract(p2, p1);
      n := b2Cross(d, 1.0);

      denom := -b2Dot(r, n);

      // Cull back facing collision and ignore parallel segments.
      if denom > k_slop then
      begin
         // Does the segment intersect the infinite line associated with this segment?
         b := Subtract(s, p1);
         a := b2Dot(b, n);

         if (0.0 <= a) and (a <= maxLambda * denom) then
         begin
            mu2 := r.y * b.x - r.x * b.y;

            // Does the segment intersect this segment?
            if (-k_slop * denom <= mu2) and (mu2 <= denom * (1.0 + k_slop)) then
            begin
               Normalize(n);
               lambda := a / denom;
               normal := n;
               Result := True;
               Exit;
            end;
         end;
      end;
   end;
   Result := False;
end;

function IsValid(const AABB: Tb2AABB): Boolean; overload; 
var
   d: TVector2;
begin
   with AABB do
   begin
      d := Subtract(upperBound, lowerBound);
      Result := (d.x >= 0.0) and (d.y >= 0.0) and 
         UPhysics2DTypes.IsValid(upperBound) and UPhysics2DTypes.IsValid(lowerBound);
   end;
end;

function IsLower(const bound: Tb2Bound): Boolean; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}
begin
   with bound do
      Result := (value and 1) = 0;
end;

function IsUpper(const bound: Tb2Bound): Boolean; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}
begin
   with bound do
      Result := (value and 1) = 1;
end;
            
function GetNext(const proxy: Tb2Proxy): UInt16; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}
begin
   Result := proxy.lowerBounds[0];
end;

procedure SetNext(var proxy: Tb2Proxy; Next: UInt16); {$IFDEF INLINE_AVAIL}inline;{$ENDIF}
begin
   proxy.lowerBounds[0] := Next;
end;

function IsValid(const proxy: Tb2Proxy): Boolean; {$IFDEF INLINE_AVAIL}inline;{$ENDIF} overload;
begin
   Result := proxy.overlapCount <> b2_invalid;
end;

procedure SetBuffered(var pair: Tb2Pair); {$IFDEF INLINE_AVAIL}inline;{$ENDIF}
begin
   with pair do
      status := status or e_pairBuffered;
end;

procedure ClearBuffered(var pair: Tb2Pair); {$IFDEF INLINE_AVAIL}inline;{$ENDIF}
begin
   with pair do
      status := status and (not e_pairBuffered);
end;

function IsBuffered(const pair: Tb2Pair): Boolean; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}
begin
   with pair do
      Result := (status and e_pairBuffered) = e_pairBuffered;
end;

procedure SetRemoved(var pair: Tb2Pair); {$IFDEF INLINE_AVAIL}inline;{$ENDIF}
begin
   with pair do
      status := status or e_pairRemoved;
end;

procedure ClearRemoved(var pair: Tb2Pair); {$IFDEF INLINE_AVAIL}inline;{$ENDIF}
begin
   with pair do
      status := status and (not e_pairRemoved);
end;

function IsRemoved(const pair: Tb2Pair): Boolean; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}
begin
   with pair do
      Result := (status and e_pairRemoved) = e_pairRemoved;
end;

procedure SetFinal(var pair: Tb2Pair); {$IFDEF INLINE_AVAIL}inline;{$ENDIF}
begin
   with pair do
      status := status or e_pairFinal;
end;

function IsFinal(const pair: Tb2Pair): Boolean; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}
begin
   with pair do
      Result := (status and e_pairFinal) = e_pairFinal;
end;

procedure SetZero(var jb: Tb2Jacobian); overload;
begin
   with jb do
   begin
      SetZero(linear1);
      SetZero(linear2);
      angular1 := 0.0;
      angular2 := 0.0;
   end;
end;

procedure SetValue(var jb: Tb2Jacobian; const x1, x2: TVector2; a1, a2: Float); overload;
begin
   with jb do
   begin
      linear1 := x1;
      linear2 := x2;
      angular1 := a1;
      angular2 := a2;
   end;
end;

function Compute(var jb: Tb2Jacobian; const x1, x2: TVector2; a1, a2: Float): Float; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}
begin
   with jb do
      Result := b2Dot(linear1, x1) + angular1 * a1 + b2Dot(linear2, x2) + angular2 * a2;
end;

{$ENDIF}

procedure b2Swap(var a, b: Tb2Bound); {$IFDEF INLINE_AVAIL}inline;{$ENDIF} overload;
var
   tmp: Tb2Bound;
begin
   tmp := a;
   a := b;
   b := tmp;
end;

{ b2BroadPhase.cpp }
function BroadPhase_BinarySearch(var bounds: Tb2AxialBoundsArray; count: Int32; value: UInt16): Int32;
var
   low, high, mid: Int32;
   pbm: Pb2Bound;
begin
   low := 0;
   high := count - 1;
   while (low <= high) do
   begin
      mid := (low + high) shr 1;
      pbm := @bounds[mid];
      if pbm^.value > value then
         high := mid - 1
      else if pbm^.value < value then
         low := mid + 1
      else
      begin
         Result := mid;
         Exit;
      end;
   end;
   Result := low;
end;

{ b2PolygonShape. cpp}

function ComputeCentroid(const vs: Tb2PolyVertices; count: Int32): TVector2;
const
   inv3 = 1 / 3;
var
   i: Integer;
   pRef, p1, p2, p3, e1, e2: TVector2;
   area, triangleArea: Float;
begin
   //b2Assert(count >= 3);    

   area := 0.0;
   // pRef is the reference point for forming triangles.
   // It's location doesn't change the result (except for rounding error).   
   SetZero(Result);
   SetZero(pRef);

  (* // This code would put the reference point inside the polygon.
   for (int32 i = 0; i < count; ++i)
   {
     pRef += vs[i];
   }
   pRef *= 1.0f / count; *)

   for i := 0 to count - 1 do
   begin
      // Triangle vertices.
      p1 := pRef;
      p2 := vs[i];
      if i + 1 < count then
         p3 := vs[i + 1]
      else
         p3 := vs[0];   

      {$IFDEF OP_OVERLOAD}         
      e1 := p2 - p1;
      e2 := p3 - p1;
      {$ELSE}      
      e1 := Subtract(p2, p1);
      e2 := Subtract(p3, p1);
      {$ENDIF}            

      triangleArea := 0.5 * b2Cross(e1, e2);
      area := area + triangleArea;

      // Area weighted centroid    

      {$IFDEF OP_OVERLOAD}        
      Result := Result + triangleArea * inv3 * (p1 + p2 + p3);
      {$ELSE}
      AddBy(p1, p2);
      AddBy(p1, p3);
      MultiplyBy(p1, triangleArea * inv3);
      AddBy(Result, p1); 
      {$ENDIF}  
   end;

   // Centroid
   //b2Assert(area > B2_FLT_EPSILON);
   {$IFDEF OP_OVERLOAD}        
   Result := Result / area;
   {$ELSE}   
   DivideBy(Result, area);
   {$ENDIF}  
end;

// http://www.geometrictools.com/Documentation/MinimumAreaRectangle.pdf
procedure ComputeOBB(var obb: Tb2OBB; const vs: Tb2PolyVertices; count: Int32);
var
   i, j: Integer;
   p: array[0..b2_maxPolygonVertices] of TVector2;
   minArea, length, area: Float;
   root, ux, uy, lower, upper, d, r, center: TVector2;
begin
   //b2Assert(count <= b2_maxPolygonVertices);

   Move(vs, p, SizeOf(vs));
   p[count] := p[0];

   minArea := FLT_MAX;

   for i := 1 to Count - 1 do
   begin
      root := p[i - 1];        
      {$IFDEF OP_OVERLOAD}    
      ux := p[i] - root;  
      length := ux.Normalize;
      //b2Assert(length > B2_FLT_EPSILON);
      uy.SetValue(-ux.y, ux.x); 
      lower.SetValue(FLT_MAX, FLT_MAX);
      upper.SetValue(-FLT_MAX, -FLT_MAX);
      {$ELSE}
      ux := Subtract(p[i], root);
      length := Normalize(ux);
      //b2Assert(length > B2_FLT_EPSILON);
      SetValue(uy, -ux.y, ux.x);
      SetValue(lower, FLT_MAX, FLT_MAX);
      SetValue(upper, -FLT_MAX, -FLT_MAX);        
      {$ENDIF}      

      for j := 0 to Count - 1 do
      begin     
         {$IFDEF OP_OVERLOAD}       
         d := p[j] - root;
         {$ELSE}
         d := Subtract(p[j], root);
         {$ENDIF}
         r.x := b2Dot(ux, d);
         r.y := b2Dot(uy, d);
         lower := b2Min(lower, r);
         upper := b2Max(upper, r);
      end;

      area := (upper.x - lower.x) * (upper.y - lower.y);
      if area < 0.95 * minArea then
      begin
         minArea := area;
         obb.R.col1 := ux;
         obb.R.col2 := uy;

         {$IFDEF OP_OVERLOAD}         
         center := 0.5 * (lower + upper);
         obb.center := root + b2Mul(obb.R, center);
         obb.extents := 0.5 * (upper - lower);
         {$ELSE}
         center := Add(lower, upper);
         MultiplyBy(center, 0.5);

         obb.center := Add(root, b2Mul(obb.R, center)); 
         obb.extents := Subtract(upper, lower);
         MultiplyBy(obb.extents, 0.5);         
         {$ENDIF}         
      end;
   end;

   //b2Assert(minArea < B2_FLT_MAX);
end;

{ b2Distance.cpp }

// GJK using Voronoi regions (Christer Ericson) and region selection
// optimizations (Casey Muratori).

// The origin is either in the region of points[1] or in the edge region. The origin is
// not in region of points[0] because that is the old point.
type
   TCalcDistanceVectors = array[0..2] of TVector2;

function Distance_ProcessTwo(var x1, x2: TVector2; var p1s, p2s,
   points: TCalcDistanceVectors): Int32;
var
   r, d: TVector2;
   length, lambda: Float;
begin
   {$IFDEF OP_OVERLOAD}
   r := -points[1];
   d := points[0] - points[1];
   length := d.Normalize;
   {$ELSE}
   r := Negative(points[1]);
   d := Subtract(points[0], points[1]);
   length := Normalize(d);   
   {$ENDIF}
   lambda := b2Dot(r, d);
   if (lambda <= 0.0) or (length < FLT_EPSILON) then
   begin
      // The simplex is reduced to a point.
      x1 := p1s[1];
      x2 := p2s[1];
      p1s[0] := p1s[1];
      p2s[0] := p2s[1];
      points[0] := points[1];
      Result := 1;
      Exit;
   end;

   // Else in edge region
   lambda := lambda / length;

   {$IFDEF OP_OVERLOAD}   
   x1 := p1s[1] + lambda * (p1s[0] - p1s[1]);
   x2 := p2s[1] + lambda * (p2s[0] - p2s[1]);
   {$ELSE}
   x1 := Subtract(p1s[0], p1s[1]);
   MultiplyBy(x1, lambda);
   AddBy(x1, p1s[1]);   
   x2 := Subtract(p2s[0], p2s[1]);
   MultiplyBy(x2, lambda);
   AddBy(x2, p2s[1]);          
   {$ENDIF}
   Result := 2;
end;

// Possible regions:
// - points[2]
// - edge points[0]-points[2]
// - edge points[1]-points[2]
// - inside the triangle
function Distance_ProcessThree(var x1, x2: TVector2; var p1s, p2s,
   points: TCalcDistanceVectors): Int32;
var
   a, b, c, ab, ac, bc: TVector2;
   tn, td, un, ud, n, vc, va, lambda, vb, denom: Float;
begin
   a := points[0];
   b := points[1];
   c := points[2];

   {$IFDEF OP_OVERLOAD}   
   ab := b - a;
   ac := c - a;
   bc := c - b;
   {$ELSE}
   ab := Subtract(b, a);
   ac := Subtract(c, a);
   bc := Subtract(c, b);
   {$ENDIF}   

   //float32 sn := -b2Dot(a, ab), sd := b2Dot(b, ab);
   tn := -b2Dot(a, ac);
   td := b2Dot(c, ac);
   un := -b2Dot(b, bc);
   ud := b2Dot(c, bc);

   // In vertex c region?
   if (td <= 0.0) and (ud <= 0.0) then
   begin
      // Single point
      x1 := p1s[2];
      x2 := p2s[2];
      p1s[0] := p1s[2];
      p2s[0] := p2s[2];
      points[0] := points[2];
      Result := 1;
      Exit;
   end;

   // Should not be in vertex a or b region.
   //B2_NOT_USED(sd);
   //B2_NOT_USED(sn);
   //b2Assert(sn > 0.0 || tn > 0.0);
   //b2Assert(sd > 0.0 || un > 0.0);

   n := b2Cross(ab, ac);

   // Should not be in edge ab region.
   vc := n * b2Cross(a, b);
   //b2Assert(vc > 0.0 || sn > 0.0 || sd > 0.0);

   // In edge bc region?
   va := n * b2Cross(b, c);
   if (va <= 0.0) and (un >= 0.0) and (ud >= 0.0) and (un + ud > 0.0) then
   begin
      //b2Assert(un + ud > 0.0);
      lambda := un / (un + ud);
      {$IFDEF OP_OVERLOAD}      
      x1 := p1s[1] + lambda * (p1s[2] - p1s[1]);
      x2 := p2s[1] + lambda * (p2s[2] - p2s[1]);
      {$ELSE}     
      x1 := Subtract(p1s[2], p1s[1]);
      MultiplyBy(x1, lambda);
      AddBy(x1, p1s[1]);

      x2 := Subtract(p2s[2], p2s[1]);
      MultiplyBy(x2, lambda);
      AddBy(x2, p2s[1]);      
      {$ENDIF}          
      p1s[0] := p1s[2];
      p2s[0] := p2s[2];
      points[0] := points[2];
      Result := 2;
      Exit;
   end;

   // In edge ac region?
   vb := n * b2Cross(c, a);
   if (vb <= 0.0) and (tn >= 0.0) and (td >= 0.0) and (tn + td > 0.0) then
   begin
      //b2Assert(tn + td > 0.0);
      lambda := tn / (tn + td);
      {$IFDEF OP_OVERLOAD}      
      x1 := p1s[0] + lambda * (p1s[2] - p1s[0]);
      x2 := p2s[0] + lambda * (p2s[2] - p2s[0]);
      {$ELSE}      
      x1 := Subtract(p1s[2], p1s[0]);
      MultiplyBy(x1, lambda);
      AddBy(x1, p1s[0]);

      x2 := Subtract(p2s[2], p2s[0]);
      MultiplyBy(x2, lambda);
      AddBy(x2, p2s[0]);     
      {$ENDIF}            
      p1s[1] := p1s[2];
      p2s[1] := p2s[2];
      points[1] := points[2];
      Result := 2;
      Exit;
   end;

   // Inside the triangle, compute barycentric coordinates
   denom := va + vb + vc;
   //b2Assert(denom > 0.0);
   denom := 1.0 / denom;

   tn := va * denom;
   td := vb * denom;
   un := 1.0 - tn - td;
   {$IFDEF OP_OVERLOAD}   
   x1 := tn * p1s[0] + td * p1s[1] + un * p1s[2];
   x2 := tn * p2s[0] + td * p2s[1] + un * p2s[2];
   {$ELSE}   
   MultiplyBy(p1s[0], tn);
   MultiplyBy(p1s[1], td);
   MultiplyBy(p1s[2], un);
   x1 := Add(p1s[0], p1s[1], p1s[2]);

   MultiplyBy(p2s[0], tn);
   MultiplyBy(p2s[1], td);
   MultiplyBy(p2s[2], un);
   x2 := Add(p2s[0], p2s[1], p2s[2]);       
   {$ENDIF}      
   Result := 3;
end;

function Distance_InPoints(const w: TVector2; var points: TCalcDistanceVectors;
   pointCount: Int32): Boolean;
const
   k_tolerance = 100.0 * FLT_EPSILON;
var
   i: Integer;
   d, m: TVector2;
begin
   for i := 0 to pointCount - 1 do
   begin
      {$IFDEF OP_OVERLOAD}   
      d := b2Abs(w - points[i]);
      {$ELSE}      
      d := b2Abs(Subtract(w, points[i]));
      {$ENDIF}            
      m := b2Max(b2Abs(w), b2Abs(points[i]));

      if (d.x < k_tolerance * (m.x + 1.0)) and (d.y < k_tolerance * (m.y + 1.0)) then
      begin
         Result := True;
         Exit;
      end;
   end;
   Result := False;
end;

{ b2TimeOfImpact.cpp }
// This algorithm uses conservative advancement to compute the time of
// impact (TOI) of two shapes.
// Refs: Bullet, Young Kim
function b2TimeOfImpact(shape1, shape2: Tb2Shape; const sweep1, sweep2: Tb2Sweep): Float;
var
   t0: Float;
   omega1, omega2, alpha: Float;
   v1, v2, p1, p2, normal: TVector2;
   k_maxIterations, iter: Int32;
   distance, targetDistance, t, approachVelocityBound, newAlpha: Float;
   xf1, xf2: Tb2XForm;
begin
   //b2Assert(sweep1.t0 == sweep2.t0);
   //b2Assert(1.0 - sweep1.t0 > B2_FLT_EPSILON);

   t0 := sweep1.t0;
   {$IFDEF OP_OVERLOAD}   
   v1 := sweep1.c - sweep1.c0;
   v2 := sweep2.c - sweep2.c0;
   {$ELSE}   
   v1 := Subtract(sweep1.c, sweep1.c0);
   v2 := Subtract(sweep2.c, sweep2.c0);
   {$ENDIF}      
   omega1 := sweep1.a - sweep1.a0;
   omega2 := sweep2.a - sweep2.a0;

   alpha := 0.0;

   k_maxIterations := 20;	// TODO_ERIN b2Settings
   iter := 0;
   normal := b2Vec2_zero;
   distance := 0.0;
   targetDistance := 0.0;
   while True do
   begin
      t := (1.0 - alpha) * t0 + alpha;
      {$IFDEF OP_OVERLOAD}      
      sweep1.GetXForm(xf1, t);
      sweep2.GetXForm(xf2, t);
      {$ELSE}      
      GetXForm(sweep1, xf1, t);
      GetXForm(sweep2, xf2, t);
      {$ENDIF}            

      // Get the distance between shapes.
      distance := b2Distance(p1, p2, shape1, shape2, xf1, xf2);

      if iter = 0 then
      begin
         // Compute a reasonable target distance to give some breathing room
         // for conservative advancement.
         if distance > 2.0 * b2_toiSlop then
            targetDistance := 1.5 * b2_toiSlop
         else
            targetDistance := b2Max(0.05 * b2_toiSlop, distance - 0.5 * b2_toiSlop);
      end;

      if (distance - targetDistance < 0.05 * b2_toiSlop) or (iter = k_maxIterations) then
         Break;

      {$IFDEF OP_OVERLOAD}         
      normal := p2 - p1;
      normal.Normalize;
      {$ELSE}      
      normal := Subtract(p2, p1);
      Normalize(normal);
      {$ENDIF}            

      // Compute upper bound on remaining movement.
      {$IFDEF OP_OVERLOAD}      
      approachVelocityBound := b2Dot(normal, v1 - v2) + Abs(omega1) *
         shape1.GetSweepRadius + Abs(omega2) * shape2.GetSweepRadius;
      {$ELSE}         
      approachVelocityBound := b2Dot(normal, Subtract(v1, v2)) + Abs(omega1) *
         shape1.GetSweepRadius + Abs(omega2) * shape2.GetSweepRadius;
      {$ENDIF}            
      if Abs(approachVelocityBound) < FLT_EPSILON then
      begin
         alpha := 1.0;
         Break;
      end;

      // Get the conservative time increment. Don't advance all the way.
      newAlpha := alpha + (distance - targetDistance) / approachVelocityBound;

      // The shapes may be moving apart or a safe distance apart.
      if (newAlpha < 0.0) or (1.0 < newAlpha) then
      begin
         alpha := 1.0;
         Break;
      end;

      // Ensure significant advancement.
      if newAlpha < (1.0 + 100.0 * FLT_EPSILON) * alpha then
         Break;

      alpha := newAlpha;
      Inc(iter);
   end;

   Result := alpha;
end;

{ b2Distance.cpp }
// Circle to circle
function b2Distance(var x1, x2: TVector2; circle1, circle2: Tb2CircleShape;
   const xf1, xf2: Tb2XForm): Float; overload;
var
   p1, p2, d: TVector2;
   dsqr, r1, r2, r, dLen: Float;
begin
   p1 := b2Mul(xf1, circle1.m_localPosition);
   p2 := b2Mul(xf2, circle2.m_localPosition);

   {$IFDEF OP_OVERLOAD}
   d := p2 - p1;
   {$ELSE}
   d := Subtract(p2, p1);
   {$ENDIF}      
   dSqr := b2Dot(d, d);
   r1 := circle1.GetRadius - b2_toiSlop;
   r2 := circle2.GetRadius - b2_toiSlop;
   r := r1 + r2;
   if dSqr > r * r then
   begin
      {$IFDEF OP_OVERLOAD}   
      dLen := d.Normalize;
      x1 := p1 + r1 * d;
      x2 := p2 - r2 * d;
      {$ELSE}      
      dLen := Normalize(d);  
      x1 := Multiply(d, r1);   
      AddBy(x1, p1);
      x2 := Multiply(d, -r2);   
      AddBy(x2, p2);      
      {$ENDIF}       
      Result := dLen - r;     
      Exit;
   end
   else if (dSqr > FLT_EPSILON * FLT_EPSILON) then
   begin
      {$IFDEF OP_OVERLOAD}   
      d.Normalize;
      x1 := p1 + r1 * d;        
      {$ELSE}      
      Normalize(d);
      x1 := Multiply(d, r1);
      AddBy(x1, p1);
      {$ENDIF}  
      x2 := x1;          
      Result := 0.0;
      Exit;
   end;

   x1 := p1;
   x2 := x1;
   Result := 0.0;
end;

// Polygon to circle
const
   maxIterations = 20;

function Distance_Generic_Simulate(var x1, x2: TVector2;
   poly1: Tb2PolygonShape; poly2: Pointer; const xf1, xf2: Tb2XForm;
   poly2AsVector: Boolean): Float;
var
   i, iter: Integer;
   p1s, p2s, points: TCalcDistanceVectors;
   pointCount: Int32;
   vSqr, vw, maxSqr: Float;
   v, w1, w2, w: TVector2;
begin
   pointCount := 0;
   x1 := poly1.GetFirstVertex(xf1);
   if poly2AsVector then
      x2 := PVector2(poly2)^
   else
      x2 := Tb2PolygonShape(poly2).GetFirstVertex(xf2);

   vSqr := 0.0;
   for iter := 0 to maxIterations - 1 do
   begin
      {$IFDEF OP_OVERLOAD}   
      v := x2 - x1;
      {$ELSE}      
      v := Subtract(x2, x1);
      {$ENDIF}            
      w1 := poly1.Support(xf1, v);
      if poly2AsVector then
         w2 := PVector2(poly2)^ // Ths same
      else
         {$IFDEF OP_OVERLOAD}      
         w2 := Tb2PolygonShape(poly2).Support(xf2, -v);
         {$ELSE}         
         w2 := Tb2PolygonShape(poly2).Support(xf2, Negative(v));
         {$ENDIF}                  

      vSqr := b2Dot(v, v);
      {$IFDEF OP_OVERLOAD}      
      w := w2 - w1;
      {$ELSE}      
      w := Subtract(w2, w1);
      {$ENDIF}
      vw := b2Dot(v, w);
      if (vSqr - vw <= 0.01 * vSqr) or Distance_InPoints(w, points, pointCount) then // or w in points
      begin
         if pointCount = 0 then
         begin
            x1 := w1;
            x2 := w2;
         end;
         g_GJK_Iterations := iter;
         Result := Sqrt(vSqr);
         Exit;
      end;

      case pointCount of
         0: begin
            p1s[0] := w1;
            p2s[0] := w2;
            points[0] := w;
            x1 := p1s[0];
            x2 := p2s[0];
            Inc(pointCount);
         end;
         1: begin
            p1s[1] := w1;
            p2s[1] := w2;
            points[1] := w;
            pointCount := Distance_ProcessTwo(x1, x2, p1s, p2s, points);
         end;
         2: begin
            p1s[2] := w1;
            p2s[2] := w2;
            points[2] := w;
            pointCount := Distance_ProcessThree(x1, x2, p1s, p2s, points);
         end;
      end;

      // If we have three points, then the origin is in the corresponding triangle.
      if pointCount = 3 then
      begin
         g_GJK_Iterations := iter;
         Result := 0.0;
         Exit;
      end;

      maxSqr := -FLT_MAX;
      for i := 0 to pointCount - 1 do
         maxSqr := b2Max(maxSqr, b2Dot(points[i], points[i]));

      if (pointCount = 3) or (vSqr <= 100.0 * FLT_EPSILON * maxSqr) then
      begin
         g_GJK_Iterations := iter;
         {$IFDEF OP_OVERLOAD}
         v := x2 - x1;
         {$ELSE}         
         v := Subtract(x2, x1);
         {$ENDIF}                  
         vSqr := b2Dot(v, v);
         Result := Sqrt(vSqr);
         Exit;
      end;
   end;

   g_GJK_Iterations := maxIterations;
   Result := Sqrt(vSqr);
end;

function b2Distance(var x1, x2: TVector2; poly: Tb2PolygonShape;
   circle: Tb2CircleShape; const xf1, xf2: Tb2XForm): Float; overload;
var
   p: TVector2;
   r: Float;
   d: TVector2;
begin
   p := b2Mul(xf2, circle.m_localPosition);

   Result := Distance_Generic_Simulate(x1, x2, poly, Pointer(@p),
      xf1, b2XForm_identity, True);
   r := circle.GetRadius - b2_toiSlop;

   if Result > r then
   begin
      Result := Result - r;
      {$IFDEF OP_OVERLOAD}      
      d := x2 - x1;
      d.Normalize;
      x2 := x2 - r * d;
      {$ELSE}      
      d := Subtract(x2, x1);
      Normalize(d);
      SubtractBy(x2, Multiply(d, r));
      {$ENDIF}            
   end
   else
   begin
      Result := 0.0;
      x2 := x1;
   end;
end;

function b2Distance(var x1, x2: TVector2; circle: Tb2CircleShape;
   poly: Tb2PolygonShape; const xf1, xf2: Tb2XForm): Float; overload; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}
begin
   Result := b2Distance(x2, x1, poly, circle, xf2, xf1);
end;

function b2Distance(var x1, x2: TVector2; poly1, poly2: Tb2PolygonShape;
   const xf1, xf2: Tb2XForm): Float; overload;
begin
   Result := Distance_Generic_Simulate(x1, x2, poly1, poly2, xf1, xf2, False);
end;

function b2Distance(var x1, x2: TVector2; shape1, shape2: Tb2Shape;
   const xf1, xf2: Tb2XForm): Float; overload;
begin
   case shape1.m_type of
      e_circleShape:
         case shape2.m_type of
            e_circleShape: Result := b2Distance(x1, x2, Tb2CircleShape(shape1),
               Tb2CircleShape(shape2), xf1, xf2);
            e_polygonShape: Result := b2Distance(x2, x1, Tb2PolygonShape(shape2),
               Tb2CircleShape(shape1), xf2, xf1);
         else
            Result := 0.0;
         end;
      e_polygonShape:
         case shape2.m_type of
            e_circleShape: Result := b2Distance(x1, x2, Tb2PolygonShape(shape1),
               Tb2CircleShape(shape2), xf1, xf2);
            e_polygonShape: Result := Distance_Generic_Simulate(x1, x2,
               Tb2PolygonShape(shape1), shape2, xf1, xf2, False);
         else
            Result := 0.0;
         end;
   else
      Result := 0.0;
   end;
end;

{ b2PairManager.cpp }
// Thomas Wang's hash, see: http://www.concentric.net/~Ttwang/tech/inthash.htm
// This assumes proxyId1 and proxyId2 are 16-bit.
function PairManager_Hash(proxyId1, proxyId2: UInt32): UInt32;
begin
   Result := (proxyId2 shl 16) or proxyId1;
   Result := (not Result) + (Result shl 15);
   Result := Result xor (Result shr 12);
   Result := Result + (Result shl 2);
   Result := Result xor (Result shr 4);
   Result := Result * 2057;
   Result := Result xor (Result shr 16);
end;

function PairManager_Equals(const pair: Tb2Pair; proxyId1,
   proxyId2: Int32): Boolean; overload; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}
begin
	 Result := (pair.proxyId1 = proxyId1) and (pair.proxyId2 = proxyId2);
end;

function PairManager_Equals(const pair1, pair2: Tb2BufferedPair): Boolean; overload; {$IFDEF INLINE_AVAIL}inline;{$ENDIF}
begin
	 Result := (pair1.proxyId1 = pair2.proxyId1) and (pair1.proxyId2 = pair2.proxyId2);
end;

{ Tb2Segment }

// Collision Detection in Interactive 3D Environments by Gino van den Bergen
// From Section 3.4.1
// x = mu1 * p1 + mu2 * p2
// mu1 + mu2 = 1 && mu1 >= 0 && mu2 >= 0
// mu1 = 1 - mu2;
// x = (1 - mu2) * p1 + mu2 * p2
//   = p1 + mu2 * (p2 - p1)
// x = s + a * r (s := start, r := end - start)
// s + a * r = p1 + mu2 * d (d := p2 - p1)
// -a * r + mu2 * d = b (b := s - p1)
// [-r d] * [a; mu2] = b
// Cramer's rule:
// denom = det[-r d]
// a = det[b d] / denom
// mu2 = det[-r b] / denom
{$IFDEF OP_OVERLOAD}
function Tb2Segment.TestSegment(var lambda: Float; var normal: TVector2;
   const segment: Tb2Segment; maxLambda: Float): Boolean;
const
   k_slop = 100.0 * FLT_EPSILON;
var
   s, r, d, n, b: TVector2;
   denom, a, mu2: Float;
begin
   s := segment.p1;
   r := segment.p2 - s;
   d := p2 - p1;
   n := b2Cross(d, 1.0);

   denom := -b2Dot(r, n);

   // Cull back facing collision and ignore parallel segments.
   if denom > k_slop then
   begin
      // Does the segment intersect the infinite line associated with this segment?
      b := s - p1;
      a := b2Dot(b, n);

      if (0.0 <= a) and (a <= maxLambda * denom) then
      begin
         mu2 := r.y * b.x - r.x * b.y;

         // Does the segment intersect this segment?
         if (-k_slop * denom <= mu2) and (mu2 <= denom * (1.0 + k_slop)) then
         begin
            n.Normalize;
            lambda := a / denom;
            normal := n;
            Result := True;
            Exit;
         end;
      end;
   end;
   Result := False;
end;
{$ENDIF}

{ Tb2AABB }
{$IFDEF OP_OVERLOAD}
function Tb2AABB.IsValid: Boolean;
var
   d: TVector2;
begin
   d := upperBound - lowerBound;
   Result := (d.x >= 0.0) and (d.y >= 0.0) and upperBound.IsValid and lowerBound.IsValid;
end;
{$ENDIF}

{ Tb2DebugDraw }

constructor Tb2DebugDraw.Create;
begin
   m_drawFlags := [];
   m_shapeColor_Static := MakeColor(0.5, 0.9, 0.5);
   m_shapeColor_Sleeping := MakeColor(0.5, 0.5, 0.9);
   m_shapeColor_Normal := MakeColor(0.9, 0.9, 0.9);
   m_pairColor := MakeColor(0.9, 0.9, 0.3);
   m_aabbColor := MakeColor(0.9, 0.3, 0.9);
   m_obbColor := MakeColor(0.5, 0.3, 0.5);
   m_world_aabbColor := MakeColor(0.3, 0.9, 0.9);
   m_coreColor := MakeColor(0.9, 0.6, 0.6);
   m_jointLineColor := MakeColor(0.5, 0.8, 0.8);

   (* !!! *)
   m_shapeColor_Sensor := MakeColor(0.1, 0.1, 0.5);
end;

//////////////////////////////////////////////////////////////
// World

constructor Tb2World.Create(const worldAABB: Tb2AABB;
   const gravity: TVector2; doSleep: Boolean);
var
   bd: Tb2BodyDef;
begin
   m_destructionListener := nil;
   m_boundaryListener := nil;
   m_contactFilter := b2_defaultFilter;
   m_contactListener := nil;
   m_debugDraw := nil;

   m_bodyList := nil;
   m_contactList := nil;
   m_jointList := nil;

   m_bodyCount := 0;
   m_contactCount := 0;
   m_jointCount := 0;

   m_positionCorrection := True;
   m_warmStarting := True;
   m_continuousPhysics := True;

   m_allowSleep := doSleep;
   m_gravity := gravity;

   m_lock := False;
   m_inv_dt0 := 0.0;

   m_contactManager := Tb2ContactManager.Create(Self);
   m_broadPhase := Tb2BroadPhase.Create(worldAABB, m_contactManager);

   bd := Tb2BodyDef.Create;
	 m_groundBody := CreateBody(bd);   
end;

destructor Tb2World.Destroy;
var
   p: Tb2Body;
begin
   // Free all shapes
   while Assigned(m_bodyList) do
   begin
      p := m_bodyList.m_next;
      DestroyBody(m_bodyList);
      m_bodyList := p;
   end;

   m_contactManager.Free;
	 m_broadPhase.Free;
end;

procedure Tb2World.Solve(const step: Tb2TimeStep);
var
   i: Integer;
   island: Tb2Island;
   b, seed, other: Tb2Body;
   c: Tb2Contact;
   j: Tb2Joint;
   stackCount: Int32;
   stack: TList;
   cn: Pb2ContactEdge;
   jn: Pb2JointEdge;
begin
   m_positionIterationCount := 0;

   // Size the island for the worst case.
   island := Tb2Island.Create(m_bodyCount, m_contactCount, m_jointCount, m_contactListener);

   // Clear all the island flags.
   b := m_bodyList;
   while Assigned(b) do
   begin
      b.m_flags := b.m_flags and (not e_body_islandFlag);
      b := b.m_next;
   end;

   c := m_contactList;
   while Assigned(c) do
   begin
      c.m_flags := c.m_flags and (not e_contact_islandFlag);
      c := c.m_next;
   end;

   j := m_jointList;
   while Assigned(j) do
   begin
      j.m_islandFlag := False;
      j := j.m_next;
   end;

   // Build and simulate all awake islands.
   stack := TList.Create;
   stack.Count := m_bodyCount;

   seed := m_bodyList;
   while Assigned(seed) do
   begin
      if (seed.m_flags and (e_body_islandFlag or e_sleepFlag or e_frozenFlag)) <> 0 then
      begin
         seed := seed.m_next;
         Continue;
      end;

      if seed.IsStatic() then
      begin
         seed := seed.m_next;
         Continue;
      end;

      // Reset island and stack.
      island.Clear;
      stackCount := 0;
      stack[stackCount] := seed;
      Inc(stackCount);
      seed.m_flags := seed.m_flags or e_body_islandFlag;

      // Perform a depth first search (DFS) on the constraint graph.
      while (stackCount > 0) do
      begin
         // Grab the next body off the stack and add it to the island.
         Dec(stackCount);
         b := Tb2Body(stack[stackCount]);
         island.Add(b);

         // Make sure the body is awake.
         b.m_flags := b.m_flags and (not e_sleepFlag);

         // To keep islands as small as possible, we don't
         // propagate islands across static bodies.
         if b.IsStatic() then
            Continue;

         // Search all contacts connected to this body.
         cn := b.m_contactList;
         while Assigned(cn) do
         begin
            // Has this contact already been added to an island?
            if (cn^.contact.m_flags and (e_contact_islandFlag or e_nonSolidFlag)) <> 0 then
            begin
               cn := cn^.next;
               Continue;
            end;

            // Is this contact touching?
            if cn.contact.GetManifoldCount = 0 then
            begin
               cn := cn^.next;
               Continue;
            end;

            island.Add(cn.contact);
            cn.contact.m_flags := cn.contact.m_flags or e_contact_islandFlag;

            other := cn.other;

            // Was the other body already added to this island?
            if (other.m_flags and e_body_islandFlag) <> 0 then
            begin
               cn := cn^.next;
               Continue;
            end;

            //b2Assert(stackCount < stackSize);
            stack[stackCount] := other;
            Inc(stackCount);

            other.m_flags := other.m_flags or e_body_islandFlag;
            cn := cn^.next;
         end;

         // Search all joints connect to this body.
         jn := b.m_jointList;
         while Assigned(jn) do
         begin
            if jn^.joint.m_islandFlag then
            begin
               jn := jn^.next;
               Continue;
            end;

            island.Add(jn^.joint);
            jn^.joint.m_islandFlag := True;

            other := jn^.other;
            if (other.m_flags and e_body_islandFlag) <> 0 then
            begin
               jn := jn^.next;
               Continue;
            end;

            //b2Assert(stackCount < stackSize);
            stack[stackCount] := other;
            Inc(stackCount);
            other.m_flags := other.m_flags or e_body_islandFlag;
            jn := jn^.next;
         end;
      end;

      island.Solve(step, m_gravity, m_positionCorrection, m_allowSleep);
      m_positionIterationCount := b2Max(m_positionIterationCount, island.m_positionIterationCount);

      // Post solve cleanup.
      for i := 0 to island.m_bodyCount - 1 do
      begin
         // Allow static bodies to participate in other islands.
         b := Tb2Body(island.m_bodies[i]);
         if b.IsStatic() then
            b.m_flags := b.m_flags and (not e_body_islandFlag);
      end;
      seed := seed.m_next;
   end;
   stack.Free;

   // Synchronize shapes, check for out of range bodies.
   b := m_bodyList;
   while Assigned(b) do
   begin
      if (b.m_flags and (e_sleepFlag or e_frozenFlag)) <> 0 then
      begin
         b := b.GetNext;
         Continue;
      end;

      if b.IsStatic() then
      begin
         b := b.GetNext;
         Continue;
      end;

      // Update shapes (for broad-phase). If the shapes go out of
      // the world AABB then shapes and contacts may be destroyed,
      // including contacts that are
      // Did the body's shapes leave the world?
      if (not b.SynchronizeShapes()) and Assigned(m_boundaryListener) then
         m_boundaryListener.Violation(b);
      b := b.GetNext;
   end;

   // Commit shape proxy movements to the broad-phase so that new contacts are created.
   // Also, some contacts can be destroyed.
   m_broadPhase.Commit;
   island.Free;
end;

procedure Tb2World.SolveTOI(const step: Tb2TimeStep);
var
   i: Integer;
   island: Tb2Island;
   stackCount: Int32;
   stack: TList;
   b, b1, b2, seed, other: Tb2Body;
   c: Tb2Contact;
   minContact: Tb2Contact;
   cn: Pb2ContactEdge;
   minTOI, toi, t0: Float;
   subStep: Tb2TimeStep;
   inRange: Boolean;
begin
   // Reserve an island and a stack for TOI island solution.
   island := Tb2Island.Create(m_bodyCount, b2_maxTOIContactsPerIsland, 0, m_contactListener);
   stack := TList.Create;
   stack.Count := m_bodyCount;

   b := m_bodyList;
   while Assigned(b) do
   begin
      b.m_flags := b.m_flags and (not e_body_islandFlag);
      b.m_sweep.t0 := 0.0;
      b := b.m_next;
   end;

   c := m_contactList;
   while Assigned(c) do
   begin
      // Invalidate TOI
      c.m_flags := c.m_flags and (not (e_toiFlag or e_contact_islandFlag));
      c := c.m_next;
   end;

   // Find TOI events and solve them.
   while True do
   begin
      // Find the first TOI.
      minContact := nil;
      minTOI := 1.0;

      c := m_contactList;
      while Assigned(c) do
      begin
         if (c.m_flags and (e_slowFlag or e_nonSolidFlag)) <> 0 then
         begin
            c := c.m_next;
            Continue;     
         end;
         // TODO_ERIN keep a counter on the contact, only respond to M TOIs per contact.

         toi := 1.0;
         if (c.m_flags and e_toiFlag) <> 0 then
            toi := c.m_toi // This contact has a valid cached TOI.
         else
         begin
            // Compute the TOI for this contact.
            b1 := c.GetShape1.GetBody;
            b2 := c.GetShape2.GetBody;

            if (b1.IsStatic() or b1.IsSleeping()) and (b2.IsStatic() or b2.IsSleeping()) then
            begin
               c := c.m_next;
               Continue;     
            end;

            // Put the sweeps onto the same time interval.
            t0 := b1.m_sweep.t0;

            if b1.m_sweep.t0 < b2.m_sweep.t0 then
            begin
               t0 := b2.m_sweep.t0;
               {$IFDEF OP_OVERLOAD}
               b1.m_sweep.Advance(t0);
               {$ELSE}
               Advance(b1.m_sweep, t0);
               {$ENDIF}
            end
            else if b2.m_sweep.t0 < b1.m_sweep.t0 then
            begin
               t0 := b1.m_sweep.t0;
               {$IFDEF OP_OVERLOAD}
               b2.m_sweep.Advance(t0);
               {$ELSE}
               Advance(b2.m_sweep, t0);
               {$ENDIF}
            end;

            //b2Assert(t0 < 1.0);

            // Compute the time of impact.
            toi := b2TimeOfImpact(c.m_shape1, c.m_shape2, b1.m_sweep, b2.m_sweep);

            //b2Assert(0.0 <= toi && toi <= 1.0);
            if (toi > 0.0) and (toi < 1.0) then
               toi := b2Min((1.0 - toi) * t0 + toi, 1.0);

            c.m_toi := toi;
            c.m_flags := c.m_flags or e_toiFlag;
         end;

         if (FLT_EPSILON < toi) and (toi < minTOI) then
         begin
            // This is the minimum TOI found so far.
            minContact := c;
            minTOI := toi;
         end;
         c := c.m_next;
      end;

      if (not Assigned(minContact)) or (1.0 - 100.0 * FLT_EPSILON < minTOI) then // No more TOI events. Done!
         Break;

      // Advance the bodies to the TOI.
      b1 := minContact.GetShape1.GetBody;
      b2 := minContact.GetShape2.GetBody;
      b1.Advance(minTOI);
      b2.Advance(minTOI);

      // The TOI contact likely has some new contact points.
      minContact.Update(m_contactListener);
      minContact.m_flags := minContact.m_flags and (not e_toiFlag);

      if minContact.GetManifoldCount = 0 then // This shouldn't happen. Numerical error?
         Continue;

      // Build the TOI island. We need a dynamic seed.
      seed := b1;
      if seed.IsStatic() then
         seed := b2;

      // Reset island and stack.
      island.Clear;
      stackCount := 0;
      stack[stackCount] := seed;
      Inc(stackCount);
      seed.m_flags := seed.m_flags or e_body_islandFlag;

      // Perform a depth first search (DFS) on the contact graph.
      while (stackCount > 0) do
      begin
         // Grab the next body off the stack and add it to the island.
         Dec(stackCount);
         b := Tb2Body(stack[stackCount]);
         island.Add(b);

         // Make sure the body is awake.
         b.m_flags := b.m_flags and (not e_sleepFlag);

         // To keep islands as small as possible, we don't
         // propagate islands across static bodies.
         if b.IsStatic() then
            Continue;

         // Search all contacts connected to this body.
         cn := b.m_contactList;
         while Assigned(cn) do
         begin
            // Does the TOI island still have space for contacts?
            if island.m_contactCount = island.m_contactCapacity then
            begin
               cn := cn.next;
               Continue;
            end;

            // Has this contact already been added to an island? Skip slow or non-solid contacts.
            if (cn.contact.m_flags and (e_contact_islandFlag or
               e_slowFlag or e_nonSolidFlag)) <> 0 then
            begin
               cn := cn.next;
               Continue;
            end;

            // Is this contact touching? For performance we are not updating this contact.
            if cn.contact.GetManifoldCount = 0 then
            begin
               cn := cn.next;
               Continue;
            end;

            island.Add(cn.contact);
            cn.contact.m_flags := cn.contact.m_flags or e_contact_islandFlag;

            // Update other body.
            other := cn.other;

            // Was the other body already added to this island?
            if (other.m_flags and e_body_islandFlag) <> 0 then
            begin
               cn := cn.next;
               Continue;
            end;

            // March forward, this can do no harm since this is the min TOI.
            if not other.IsStatic() then
            begin
               other.Advance(minTOI);
               other.WakeUp;
            end;

            //b2Assert(stackCount < stackSize);
            stack[stackCount] := other;
            Inc(stackCount);
            other.m_flags := other.m_flags or e_body_islandFlag;
            cn := cn.next;
         end;
      end;

      subStep.dt := (1.0 - minTOI) * step.dt;
      //b2Assert(subStep.dt > B2_FLT_EPSILON);
      subStep.inv_dt := 1.0 / subStep.dt;
      subStep.maxIterations := step.maxIterations;

      island.SolveTOI(subStep);

      // Post solve cleanup.
      for i := 0 to island.m_bodyCount - 1 do
      begin
         // Allow bodies to participate in future TOI islands.
         b := Tb2Body(island.m_bodies[i]);
         b.m_flags := b.m_flags and (not e_body_islandFlag);

         if (b.m_flags and (e_sleepFlag or e_frozenFlag)) <> 0 then
            Continue;

         if b.IsStatic() then
            Continue;

         // Update shapes (for broad-phase). If the shapes go out of
         // the world AABB then shapes and contacts may be destroyed,
         // including contacts that are
         inRange := b.SynchronizeShapes;

         // Did the body's shapes leave the world?
         if (not inRange) and Assigned(m_boundaryListener) then
            m_boundaryListener.Violation(b);

         // Invalidate all contact TOIs associated with this body. Some of these
         // may not be in the island because they were not touching.
         cn := b.m_contactList;
         while Assigned(cn) do
         begin
            cn.contact.m_flags := cn.contact.m_flags and (not e_toiFlag);
            cn := cn.next;
         end;
      end;

      for i := 0 to island.m_contactCount - 1 do
         with Tb2Contact(island.m_contacts[i]) do
         begin
            // Allow contacts to participate in future TOI islands.
            m_flags := m_flags and (not (e_toiFlag or e_contact_islandFlag));
         end;

      // Commit shape proxy movements to the broad-phase so that new contacts are created.
      // Also, some contacts can be destroyed.
      m_broadPhase.Commit;
   end;

   stack.Free;
   island.Free;
end;

procedure Tb2World.DrawDebugData;
var
   core: Boolean;
   xf: Tb2XForm;
   b: Tb2Body;
   s: Tb2Shape;
   j: Tb2Joint;
   invQ, h: TVector2;
   i: Integer;
   b1, b2: Tb2AABB;
   index: UInt16;
   vs: TVectorArray4;
begin
   if not Assigned(m_debugDraw) then
      Exit;

   with m_debugDraw do
   begin
      if e_shapeBit in m_drawFlags then
      begin
         core := e_coreShapeBit in m_drawFlags;

         b := m_bodyList;
         while Assigned(b) do
         begin
            xf := b.m_xf;
            s := b.GetShapeList;
            while Assigned(s) do
            begin
               (* !!! *)
               if b.IsStatic then
               begin
                if s.IsSensor then
                  DrawShape(s, xf, m_shapeColor_Sensor, core)
                else                
                  DrawShape(s, xf, m_shapeColor_Static, core)
               end
               else if b.IsSleeping then
                  DrawShape(s, xf, m_shapeColor_Sleeping, core)
               else
                  DrawShape(s, xf, m_shapeColor_Normal, core);
               s := s.m_next;
            end;

            b := b.GetNext;
         end;
      end;

      if e_jointBit in m_drawFlags then
      begin
         j := m_jointList;
         while Assigned(j) do
         begin
            if j.m_type <> e_mouseJoint then
              DrawJoint(j);
            j := j.m_next;
         end;
      end;

      with m_broadPhase do
      begin
         if e_pairBit in m_drawFlags then
         begin
            {$IFDEF OP_OVERLOAD}
            invQ.SetValue(1.0 / m_quantizationFactor.x, 1.0 / m_quantizationFactor.y);
            {$ELSE}
            SetValue(invQ, 1.0 / m_quantizationFactor.x, 1.0 / m_quantizationFactor.y);
            {$ENDIF}

            for i := 0 to b2_tableCapacity - 1 do
            begin
               index := m_pairManager.m_hashTable[i];
               while (index <> b2_nullPair) do
                  with m_pairManager.m_pairs[index] do
                  begin
                     with m_proxyPool[proxyId1] do
                     begin
                        b1.lowerBound.x := m_worldAABB.lowerBound.x + invQ.x * m_bounds[0][lowerBounds[0]].value;
                        b1.lowerBound.y := m_worldAABB.lowerBound.y + invQ.y * m_bounds[1][lowerBounds[1]].value;
                        b1.upperBound.x := m_worldAABB.lowerBound.x + invQ.x * m_bounds[0][upperBounds[0]].value;
                        b1.upperBound.y := m_worldAABB.lowerBound.y + invQ.y * m_bounds[1][upperBounds[1]].value;
                     end;

                     with m_proxyPool[proxyId2] do
                     begin
                        b2.lowerBound.x := m_worldAABB.lowerBound.x + invQ.x * m_bounds[0][lowerBounds[0]].value;
                        b2.lowerBound.y := m_worldAABB.lowerBound.y + invQ.y * m_bounds[1][lowerBounds[1]].value;
                        b2.upperBound.x := m_worldAABB.lowerBound.x + invQ.x * m_bounds[0][upperBounds[0]].value;
                        b2.upperBound.y := m_worldAABB.lowerBound.y + invQ.y * m_bounds[1][upperBounds[1]].value;
                     end;

                     {$IFDEF OP_OVERLOAD}
                     DrawSegment(0.5 * (b1.lowerBound + b1.upperBound),
                        0.5 * (b2.lowerBound + b2.upperBound), m_pairColor);
                     {$ELSE}
                     DrawSegment(Multiply(Add(b1.lowerBound, b1.upperBound), 0.5),
                        Multiply(Add(b2.lowerBound, b2.upperBound), 0.5), m_pairColor);
                     {$ENDIF}

                     index := next;
                  end;
            end;
         end;

         if e_aabbBit in m_drawFlags then
         begin
            {$IFDEF OP_OVERLOAD}
            invQ.SetValue(1.0 / m_quantizationFactor.x, 1.0 / m_quantizationFactor.y);
            {$ELSE}
            SetValue(invQ, 1.0 / m_quantizationFactor.x, 1.0 / m_quantizationFactor.y);
            {$ENDIF}
            for i := 0 to b2_maxProxies - 1 do
            begin
               with m_proxyPool[i] do
               begin
                  {$IFDEF OP_OVERLOAD}
                  if not IsValid then
                  {$ELSE}
                  if not IsValid(m_proxyPool[i]) then
                  {$ENDIF}
                     Continue;

                  b1.lowerBound.x := m_worldAABB.lowerBound.x + invQ.x * m_bounds[0][lowerBounds[0]].value;
                  b1.lowerBound.y := m_worldAABB.lowerBound.y + invQ.y * m_bounds[1][lowerBounds[1]].value;
                  b1.upperBound.x := m_worldAABB.lowerBound.x + invQ.x * m_bounds[0][upperBounds[0]].value;
                  b1.upperBound.y := m_worldAABB.lowerBound.y + invQ.y * m_bounds[1][upperBounds[1]].value;
               end;

               vs[0] := b1.lowerBound;
               vs[2] := b1.upperBound;
               {$IFDEF OP_OVERLOAD}
               vs[1].SetValue(b1.upperBound.x, b1.lowerBound.y);
               vs[3].SetValue(b1.lowerBound.x, b1.upperBound.y);
               {$ELSE}
               SetValue(vs[1], b1.upperBound.x, b1.lowerBound.y);
               SetValue(vs[3], b1.lowerBound.x, b1.upperBound.y);
               {$ENDIF}

               DrawPolygon4(vs, 4, m_aabbColor);
            end;

            vs[0] := m_worldAABB.lowerBound;
            vs[2] := m_worldAABB.upperBound;
            {$IFDEF OP_OVERLOAD}
            vs[1].SetValue(m_worldAABB.upperBound.x, m_worldAABB.lowerBound.y);
            vs[3].SetValue(m_worldAABB.lowerBound.x, m_worldAABB.upperBound.y);
            {$ELSE}
            SetValue(vs[1], m_worldAABB.upperBound.x, m_worldAABB.lowerBound.y);
            SetValue(vs[3], m_worldAABB.lowerBound.x, m_worldAABB.upperBound.y);
            {$ENDIF}
            DrawPolygon4(vs, 4, m_world_aabbColor);
         end;
      end;

      if e_obbBit in m_drawFlags then
      begin
         b := m_bodyList;
         while Assigned(b) do
         begin
            xf := b.m_xf;
            s := b.GetShapeList;
            while Assigned(s) do
            begin
               if s.GetType <> e_polygonShape then
               begin
                  s := s.m_next;
                  Continue;
               end;

               h := Tb2PolygonShape(s).m_obb.extents;

               {$IFDEF OP_OVERLOAD}
               vs[0].SetValue(-h.x, -h.y);
               vs[1].SetValue(h.x, -h.y);
               vs[3].SetValue(-h.x,  h.y);
               {$ELSE}
               SetValue(vs[0], -h.x, -h.y);
               SetValue(vs[1], h.x, -h.y);
               SetValue(vs[3], -h.x,  h.y);
               {$ENDIF}
               vs[2] := h;

               for i := 0 to 3 do
                  with Tb2PolygonShape(s).m_obb do
                  begin
                     {$IFDEF OP_OVERLOAD}
                     vs[i] := center + b2Mul(R, vs[i]);
                     {$ELSE}
                     vs[i] := Add(center, b2Mul(R, vs[i]));
                     {$ENDIF}
                     vs[i] := b2Mul(xf, vs[i]);
                  end;

               DrawPolygon4(vs, 4, m_obbColor);
               s := s.m_next;
            end;

            b := b.GetNext;
         end;
      end;

      if e_centerOfMassBit in m_drawFlags then
      begin
         b := m_bodyList;
         while Assigned(b) do
         begin
            xf := b.m_xf;
            xf.position := b.GetWorldCenter;
            DrawXForm(xf);
            b := b.GetNext;
         end;
      end;
   end;
end;

procedure Tb2World.DrawShape(shape: Tb2Shape; const xf: Tb2XForm;
   const color: RGBA; core: Boolean);
var
   i: Integer;
   center: TVector2;
   vertices: Tb2PolyVertices;
begin
   with m_debugDraw do
      case shape.GetType of
         e_circleShape:
            with Tb2CircleShape(shape) do
            begin
               center := b2Mul(xf, m_localPosition);
               DrawSolidCircle(center, xf.R.col1, GetRadius, color);

               if core then
                  DrawCircle(center, GetRadius - b2_toiSlop, m_coreColor);
            end;
         e_polygonShape:
            with Tb2PolygonShape(shape) do
            begin
               //b2Assert(vertexCount <= b2_maxPolygonVertices);
               for i := 0 to GetVertexCount - 1 do
                  vertices[i] := b2Mul(xf, m_vertices[i]);

               DrawSolidPolygon(vertices, GetVertexCount, color);

               if core then
               begin
                  for i := 0 to GetVertexCount - 1 do
                     vertices[i] := b2Mul(xf, m_coreVertices[i]);
                  DrawPolygon(vertices, GetVertexCount, m_coreColor);
               end;
            end;
      end;
end;

procedure Tb2World.DrawJoint(joint: Tb2Joint);
var
   p1, p2, s1, s2: TVector2;
begin
   p1 := joint.GetAnchor1;
   p2 := joint.GetAnchor2;

   with m_debugDraw do
   begin
      case joint.GetType of
         e_distanceJoint: DrawSegment(p1, p2, m_jointLineColor);
         e_pulleyJoint:
            with Tb2PulleyJoint(joint) do
            begin
               s1 := GetGroundAnchor1;
               s2 := GetGroundAnchor2;
               DrawSegment(s1, p1, m_jointLineColor);
               DrawSegment(s2, p2, m_jointLineColor);
               DrawSegment(s1, s2, m_jointLineColor);
            end;
         e_mouseJoint: ;
         e_fixedJoint: DrawSegment(p1, p2, m_jointLineColor);
      else
         DrawSegment(joint.GetBody1.m_xf.position, p1, m_jointLineColor);
         DrawSegment(p1, p2, m_jointLineColor);
         DrawSegment(joint.GetBody2.m_xf.position, p2, m_jointLineColor);
      end;
   end;
end;

function Tb2World.CreateBody(def: Tb2BodyDef; AutoFreeBodyDef: Boolean = True): Tb2Body;
begin
   //b2Assert(m_lock == False);
   if m_lock then
   begin
      Result := nil;
      Exit;
   end;

   Result := Tb2Body.Create(def, Self);

   // Add to world doubly linked list.
   Result.m_prev := nil;
   Result.m_next := m_bodyList;
   if Assigned(m_bodyList) then
      m_bodyList.m_prev := Result;
   m_bodyList := Result;
   Inc(m_bodyCount);

   if AutoFreeBodyDef then
      def.Free;
end;

procedure Tb2World.DestroyBody(body: Tb2Body; DoFree: Boolean = True);
var
   jn, jn0: Pb2JointEdge;
   s, s0: Tb2Shape;
begin
   //b2Assert(m_bodyCount > 0);
   //b2Assert(m_lock == False);
   if m_lock then
      Exit;

   // Delete the attached joints.
   jn := body.m_jointList;

   if Assigned(m_destructionListener) then
   begin
      while Assigned(jn) do
      begin
         jn0 := jn;
         jn := jn^.next;
         m_destructionListener.SayGoodbye(jn0^.joint);
         DestroyJoint(jn0^.joint);
      end;
   end
   else
   begin
      while Assigned(jn) do
      begin
         jn0 := jn;
         jn := jn^.next;
         DestroyJoint(jn0^.joint);
      end;
   end;

   // Delete the attached shapes. This destroys broad-phase
   // proxies and pairs, leading to the destruction of contacts.
   s := body.m_shapeList;
   if Assigned(m_destructionListener) then
   begin
      while Assigned(s) do
      begin
         s0 := s;
         s := s.m_next;
         m_destructionListener.SayGoodbye(s0);
         s0.DestroyProxy(m_broadPhase);
         s0.Free;
      end;
   end
   else
   begin
      while Assigned(s) do
      begin
         s0 := s;
         s := s.m_next;
         s0.DestroyProxy(m_broadPhase);
         s0.Free;
      end;
   end;

   // Remove world body list.
   if Assigned(body.m_prev) then
      body.m_prev.m_next := body.m_next;

   if Assigned(body.m_next) then
      body.m_next.m_prev := body.m_prev;

   if body = m_bodyList then
      m_bodyList := body.m_next;

   Dec(m_bodyCount);
   if DoFree then
      body.Destroy2;
end;

function Tb2World.CreateJoint(def: Tb2JointDef; AutoFreeJointDef: Boolean = True): Tb2Joint;
var
   j: Tb2Joint;
   b: Tb2Body;
   s: Tb2Shape;
begin
   //b2Assert(m_lock == False);
   Result := nil;
   case def.JointType of
      e_unknownJoint: Exit;
      e_revoluteJoint: j := Tb2RevoluteJoint.Create(Tb2RevoluteJointDef(def));
      e_prismaticJoint: j := Tb2PrismaticJoint.Create(Tb2PrismaticJointDef(def));
      e_distanceJoint: j := Tb2DistanceJoint.Create(Tb2DistanceJointDef(def));
      e_pulleyJoint: j := Tb2PulleyJoint.Create(Tb2PulleyJointDef(def));
      e_mouseJoint: j := Tb2MouseJoint.Create(Tb2MouseJointDef(def));
      e_gearJoint: j := Tb2GearJoint.Create(Tb2GearJointDef(def));
      e_fixedJoint: j := Tb2FixedJoint.Create(Tb2FixedJointDef(def));
   end;

   // Connect to the world list.
   j.m_prev := nil;
   j.m_next := m_jointList;
   if Assigned(m_jointList) then
      m_jointList.m_prev := j;

   m_jointList := j;
   Inc(m_jointCount);

   // Connect to the bodies' doubly linked lists.
   j.m_node1.joint := j;
   j.m_node1.other := j.m_body2;
   j.m_node1.prev := nil;
   j.m_node1.next := j.m_body1.m_jointList;
   if Assigned(j.m_body1.m_jointList) then
      j.m_body1.m_jointList.prev := @j.m_node1;
   j.m_body1.m_jointList := @j.m_node1;

   j.m_node2.joint := j;
   j.m_node2.other := j.m_body1;
   j.m_node2.prev := nil;
   j.m_node2.next := j.m_body2.m_jointList;
   if Assigned(j.m_body2.m_jointList) then
      j.m_body2.m_jointList.prev := @j.m_node2;
   j.m_body2.m_jointList := @j.m_node2;

   // If the joint prevents collisions, then reset collision filtering.
   if not def.collideConnected then
   begin
     // Reset the proxies on the body with the minimum number of shapes.
     if def.body1.m_shapeCount < def.body2.m_shapeCount then
        b := def.body1
     else
        b := def.body2;

     s := b.m_shapeList;
     while Assigned(s) do
     begin
        s.RefilterProxy(m_broadPhase, b.m_xf);
        s := s.m_next;
     end;
   end;

   if AutoFreeJointDef then
      def.Free;

   Result := j;
end;

procedure Tb2World.DestroyJoint(j: Tb2Joint);
var
   collideConnected: Boolean;
   body1, body2, b: Tb2Body;
   s: Tb2Shape;
begin
   //b2Assert(m_lock == False);

   collideConnected := j.m_collideConnected;

   // Remove from the doubly linked list.
   if Assigned(j.m_prev) then
      j.m_prev.m_next := j.m_next;

   if Assigned(j.m_next) then
      j.m_next.m_prev := j.m_prev;

   if j = m_jointList then
      m_jointList := j.m_next;

   // Disconnect from island graph.
   body1 := j.m_body1;
   body2 := j.m_body2;

   // Wake up connected bodies.
   body1.WakeUp;
   body2.WakeUp;

   // Remove from body 1.
   if Assigned(j.m_node1.prev) then
      j.m_node1.prev^.next := j.m_node1.next;

   if Assigned(j.m_node1.next) then
      j.m_node1.next^.prev := j.m_node1.prev;

   if (@j.m_node1) = body1.m_jointList then
      body1.m_jointList := j.m_node1.next;

   j.m_node1.prev := nil;
   j.m_node1.next := nil;

   // Remove from body 2
   if Assigned(j.m_node2.prev) then
      j.m_node2.prev^.next := j.m_node2.next;

   if Assigned(j.m_node2.next) then
      j.m_node2.next^.prev := j.m_node2.prev;

   if (@j.m_node2) = body2.m_jointList then
      body2.m_jointList := j.m_node2.next;

   j.m_node2.prev := nil;
   j.m_node2.next := nil;

   j.Free;

   //b2Assert(m_jointCount > 0);
   Dec(m_jointCount);

   // If the joint prevents collisions, then reset collision filtering.
   if not collideConnected then
   begin
      // Reset the proxies on the body with the minimum number of shapes.
      if body1.m_shapeCount < body2.m_shapeCount then
         b := body1
      else
         b := body2;

      s := b.m_shapeList;
      while Assigned(s) do
      begin
         s.RefilterProxy(m_broadPhase, b.m_xf);
         s := s.m_next;
      end;
   end;
end;

procedure Tb2World.Step(timeStep: Float; iterations: Int32; drawThisStep: Boolean = True);
var
   step: Tb2TimeStep;
begin
   m_lock := True;

   step.dt := timeStep;
   step.maxIterations	:= iterations;
   if timeStep > 0.0 then
      step.inv_dt := 1.0 / timeStep
   else
      step.inv_dt := 0.0;

   step.dtRatio := m_inv_dt0 * timeStep;
   step.positionCorrection := m_positionCorrection;
   step.warmStarting := m_warmStarting;

   // Update contacts.
   m_contactManager.Collide;

   if step.dt > 0.0 then
   begin
      Solve(step); // Integrate velocities, solve velocity constraints, and integrate positions.

      if m_continuousPhysics then
         SolveTOI(step); // Handle TOI events.
   end;

   if drawThisStep then
      DrawDebugData;

   m_inv_dt0 := step.inv_dt;
   m_lock := False;
end;

function Tb2World.Query(const aabb: Tb2AABB; shapes: TList; maxCount: Int32): Int32;
begin
   Result := m_broadPhase.Query(aabb, shapes, maxCount);
end;

procedure Tb2World.Refilter(shape: Tb2Shape);
begin
   shape.RefilterProxy(m_broadPhase, shape.GetBody.m_xf);
end;

procedure Tb2World.Validate;
begin
   m_broadPhase.Validate;
end;

function Tb2World.GetProxyCount: Int32;
begin
   Result := m_broadPhase.m_proxyCount;
end;

function Tb2World.GetPairCount: Int32;
begin
   Result := m_broadPhase.m_pairManager.m_pairCount;
end;

procedure Tb2World.SetGravity(const gravity: TVector2);
begin
   m_gravity := gravity;
end;

procedure Tb2World.WakeAllSleepingBodies;
var
   b: Tb2Body;
begin
   b := m_bodyList;
   while Assigned(b) do
   begin
      if b.IsDynamic and b.IsSleeping then
         b.WakeUp;
      b := b.GetNext;
   end;
end;

////////////////////////////////////////////////////
// Contact

{ Tb2ContactFilter }

function Tb2ContactFilter.ShouldCollide(shape1, shape2: Tb2Shape): Boolean;
begin
   with shape1.m_filter do
      if (groupIndex = shape2.m_filter.groupIndex) and (groupIndex <> 0) then
         Result := groupIndex > 0
      else
         Result := ((maskBits and shape2.m_filter.categoryBits) <> 0) and
            ((categoryBits and shape2.m_filter.maskBits) <> 0);
end;

{ Tb2ContactListener }

procedure Tb2ContactListener.Add(var point: Tb2ContactPoint);
begin
end;

procedure Tb2ContactListener.Persist(var point: Tb2ContactPoint);
begin
end;

procedure Tb2ContactListener.Remove(var point: Tb2ContactPoint);
begin
end;

procedure Tb2ContactListener.Result(var point: Tb2ContactResult);
begin
end;

{ Tb2Contact }

constructor Tb2Contact.Create;
begin
   m_flags := 0;
   m_shape1 := nil;
   m_shape2 := nil;
end;

constructor Tb2Contact.Create(shape1, shape2: Tb2Shape);
begin
   Create;
   if shape1.IsSensor or shape2.IsSensor then
      m_flags := m_flags or e_nonSolidFlag;

   m_shape1 := shape1;
   m_shape2 := shape2;

   m_manifoldCount := 0;

   m_friction := b2MixFriction(m_shape1.GetFriction, m_shape2.GetFriction);
   m_restitution := b2MixRestitution(m_shape1.GetRestitution, m_shape2.GetRestitution);
   m_prev := nil;
   m_next := nil;

   m_node1.contact := nil;
   m_node1.prev := nil;
   m_node1.next := nil;
   m_node1.other := nil;

   m_node2.contact := nil;
   m_node2.prev := nil;
   m_node2.next := nil;
   m_node2.other := nil;
end;

class function Tb2Contact.CreateContact(Shape1, Shape2: Tb2Shape): Tb2Contact;
var
   i: Integer;
   p: Pb2Manifold;
begin
   with ContactCreateRecords[Shape1.GetType][Shape2.GetType] do
   begin
      if Primary then
         Result := ClassType.Create(Shape1, Shape2)
      else
      begin
         Result := ClassType.Create(Shape2, Shape1);

         p := Result.GetManifolds;
         for i := 0 to Result.GetManifoldCount - 1 do
            {$IFDEF D2009UP}
            {$IFDEF OP_OVERLOAD}
            (p + i)^.normal.SetNegative;
            {$ELSE}
            SetNegative((p + i)^.normal);
            {$ENDIF}
            {$ELSE}
            {$IFDEF OP_OVERLOAD}
            Pb2Manifold(Integer(p) + i * SizeOf(Tb2Manifold))^.normal.SetNegative;
            {$ELSE}
            SetNegative(Pb2Manifold(Integer(p) + i * SizeOf(Tb2Manifold))^.normal);
            {$ENDIF}
            {$ENDIF}
      end;
   end;
end;

function Tb2Contact.IsSolid: Boolean;
begin
   Result := (m_flags and e_nonSolidFlag) = 0;
end;

procedure Tb2Contact.Update(listener: Tb2ContactListener);
var
   oldCount, newCount: Int32;
   body1, body2: Tb2Body;
begin
   oldCount := m_manifoldCount;
   Evaluate(listener);
   newCount := m_manifoldCount;

   body1 := m_shape1.GetBody;
   body2 := m_shape2.GetBody;

   if (newCount = 0) and (oldCount > 0) then
   begin
      body1.WakeUp;
      body2.WakeUp;
   end;

   // Slow contacts don't generate TOI events.
   if body1.IsStatic or body1.IsBullet or body2.IsStatic or body2.IsBullet then
      m_flags := m_flags and (not e_slowFlag)
   else
      m_flags := m_flags or e_slowFlag;
end;

{ Tb2ContactSolver }

constructor Tb2ContactSolver.Create(const step: Tb2TimeStep;
   contacts: TList; contactCount: Int32);
var
   i, j, k: Integer;
   contact: Tb2Contact;
   b1, b2: Tb2Body;
   manifolds, manifold: Pb2Manifold;
   count: Int32;
   friction, restitution: Float;
   w1, w2: Float;
   v1, v2: TVector2;
   normal, tangent: TVector2;
   c: Pb2ContactConstraint;
   cp: Pb2ManifoldPoint;
   ccp: Pb2ContactConstraintPoint;
   r1Sqr, r2Sqr, rn1, rn2: Float;
   kNormal, kEqualized: Float;
   rt1, rt2, kTangent: Float;
   vRel: Float;
   {$IFNDEF OP_OVERLOAD}tmpV: TVector2;{$ENDIF} 
begin
   //m_step := step;
   m_constraintCount := 0;
   for i := 0 to contactCount - 1 do
   begin
      //b2Assert(contacts[i]->IsSolid());
      m_constraintCount := m_constraintCount + Tb2Contact(contacts[i]).GetManifoldCount;
   end;

   m_constraints := Pb2ContactConstraint(GetMemory(m_constraintCount * SizeOf(Tb2ContactConstraint)));

   count := 0;
   for i := 0 to contactCount - 1 do
   begin
      contact := Tb2Contact(contacts[i]);
      b1 := contact.m_shape1.GetBody;
      b2 := contact.m_shape2.GetBody;
      manifolds := contact.GetManifolds;
      friction := contact.m_friction;
      restitution := contact.m_restitution;

      v1 := b1.m_linearVelocity;
      v2 := b2.m_linearVelocity;
      w1 := b1.m_angularVelocity;
      w2 := b2.m_angularVelocity;

      for j := 0 to contact.GetManifoldCount - 1 do
      begin
         {$IFDEF D2009UP}
         manifold := manifolds + j;
         {$ELSE}
         manifold := Pb2Manifold(Integer(manifolds) + j * SizeOf(Tb2Manifold));
         {$ENDIF}
         //b2Assert(manifold->pointCount > 0);

         normal := manifold^.normal;

         //b2Assert(count < m_constraintCount);
         {$IFDEF D2009UP}
         c := m_constraints + count;
         {$ELSE}
         c := Pb2ContactConstraint(Integer(m_constraints) + count * SizeOf(Tb2ContactConstraint));
         {$ENDIF}

         c.body1 := b1;
         c.body2 := b2;
         c.manifold := manifold;
         c.normal := normal;
         c.pointCount := manifold^.pointCount;
         c.friction := friction;
         c.restitution := restitution;

         for k := 0 to c.pointCount - 1 do
         begin
            cp := @manifold^.points[k];
            ccp := @c^.points[k];

            with ccp^ do
            begin
               normalImpulse := cp^.normalImpulse;
               tangentImpulse := cp^.tangentImpulse;
               separation := cp^.separation;
               positionImpulse := 0.0;

               localAnchor1 := cp^.localPoint1;
               localAnchor2 := cp^.localPoint2;
               {$IFDEF OP_OVERLOAD}               
               r1 := b2Mul(b1.m_xf.R, cp^.localPoint1 - b1.GetLocalCenter);
               r2 := b2Mul(b2.m_xf.R, cp^.localPoint2 - b2.GetLocalCenter);
               {$ELSE}               
               r1 := b2Mul(b1.m_xf.R, Subtract(cp^.localPoint1, b1.GetLocalCenter));
               r2 := b2Mul(b2.m_xf.R, Subtract(cp^.localPoint2, b2.GetLocalCenter));
               {$ENDIF}                              

               r1Sqr := b2Dot(r1, r1);
               r2Sqr := b2Dot(r2, r2);
               rn1 := b2Dot(r1, normal);
               rn2 := b2Dot(r2, normal);

               kNormal := b1.m_invMass + b2.m_invMass + b1.m_invI * (r1Sqr -
                  rn1 * rn1) + b2.m_invI * (r2Sqr - rn2 * rn2);

               //b2Assert(kNormal > B2_FLT_EPSILON);
               normalMass := 1.0 / kNormal;

               kEqualized := b1.m_mass * b1.m_invMass + b2.m_mass * b2.m_invMass +
                  b1.m_mass * b1.m_invI * (r1Sqr - rn1 * rn1) + b2.m_mass *
                  b2.m_invI * (r2Sqr - rn2 * rn2);

               //b2Assert(kEqualized > B2_FLT_EPSILON);
               equalizedMass := 1.0 / kEqualized;

               tangent := b2Cross(normal, 1.0);

               rt1 := b2Dot(r1, tangent);
               rt2 := b2Dot(r2, tangent);
               kTangent := b1.m_invMass + b2.m_invMass + b1.m_invI * (r1Sqr -
                  rt1 * rt1) + b2.m_invI * (r2Sqr - rt2 * rt2);

               //b2Assert(kTangent > B2_FLT_EPSILON);
               tangentMass := 1.0 / kTangent;

               // Setup a velocity bias for restitution.
               velocityBias := 0.0;
               if separation > 0.0 then
                  velocityBias := -60.0 * separation; // TODO_ERIN b2TimeStep

               {$IFDEF OP_OVERLOAD}
               vRel := b2Dot(c.normal, v2 + b2Cross(w2, r2) - v1 - b2Cross(w1, ccp.r1));
               {$ELSE}
               tmpV := Subtract(v2, v1);
               AddBy(tmpV, b2Cross(w2, r2));
               SubtractBy(tmpV, b2Cross(w1, ccp.r1));
               vRel := b2Dot(c.normal, tmpV);
               {$ENDIF}
               if vRel < -b2_velocityThreshold then
                  velocityBias := velocityBias - c.restitution * vRel;
            end;
         end;
         Inc(count);
      end;
   end;
   //b2Assert(count == m_constraintCount);
end;

destructor Tb2ContactSolver.Destroy;
begin
   FreeMemory(m_constraints);
end;

procedure Tb2ContactSolver.InitVelocityConstraints(const step: Tb2TimeStep);
var
   i, j: Integer;
   b1, b2: Tb2Body;
   c: Pb2ContactConstraint;
   invMass1, invI1, invMass2, invI2: Float;
   normal, tangent: TVector2;
   P: TVector2;
begin
   // Warm start.
   for i := 0 to m_constraintCount - 1 do
   begin
      {$IFDEF D2009UP}
      c := m_constraints + i;
      {$ELSE}
      c := Pb2ContactConstraint(Integer(m_constraints) + i * SizeOf(Tb2ContactConstraint));
      {$ENDIF}

      b1 := c.body1;
      b2 := c.body2;
      invMass1 := b1.m_invMass;
      invI1 := b1.m_invI;
      invMass2 := b2.m_invMass;
      invI2 := b2.m_invI;
      normal := c.normal;
      tangent := b2Cross(normal, 1.0);

      if step.warmStarting then
      begin
         for j := 0 to c.pointCount - 1 do
            with c^.points[j] do
            begin
               normalImpulse := normalImpulse * step.dtRatio;
               tangentImpulse := tangentImpulse * step.dtRatio;
               {$IFDEF OP_OVERLOAD}               
               P := normalImpulse * normal + tangentImpulse * tangent;                 
               b1.m_linearVelocity.SubtractBy(invMass1 * P);
               b2.m_linearVelocity.AddBy(invMass2 * P);
               {$ELSE}               
               P := Multiply(normal, normalImpulse);
               AddBy(P, Multiply(tangent, tangentImpulse));
               SubtractBy(b1.m_linearVelocity, Multiply(P, invMass1));
               AddBy(b2.m_linearVelocity, Multiply(P, invMass2));
               {$ENDIF}                              
               
               b1.m_angularVelocity := b1.m_angularVelocity - invI1 * b2Cross(r1, P);
               b2.m_angularVelocity := b2.m_angularVelocity + invI2 * b2Cross(r2, P);                  
            end;
      end
      else
         for j := 0 to c.pointCount - 1 do
            with c^.points[j] do
            begin
               normalImpulse := 0.0;
               tangentImpulse := 0.0;
            end;
   end;
end;

procedure Tb2ContactSolver.SolveVelocityConstraints;
var
   i, j: Integer;
   b1, b2: Tb2Body;
   w1, w2: Float;
   v1, v2: TVector2;
   invMass1, invI1, invMass2, invI2: Float;
   normal, tangent: TVector2;
   friction: Float;
   c: Pb2ContactConstraint;
   dv, P: TVector2;
   vn, vt, lambda, newImpulse, maxFriction: Float;
begin
   for i := 0 to m_constraintCount - 1 do
   begin
      {$IFDEF D2009UP}
      c := m_constraints + i;
      {$ELSE}
      c := Pb2ContactConstraint(Integer(m_constraints) + i * SizeOf(Tb2ContactConstraint));
      {$ENDIF}

      b1 := c^.body1;
      b2 := c^.body2;
      w1 := b1.m_angularVelocity;
      w2 := b2.m_angularVelocity;
      v1 := b1.m_linearVelocity;
      v2 := b2.m_linearVelocity;
      invMass1 := b1.m_invMass;
      invI1 := b1.m_invI;
      invMass2 := b2.m_invMass;
      invI2 := b2.m_invI;
      normal := c^.normal;
      tangent := b2Cross(normal, 1.0);
      friction := c^.friction;

      // Solve normal constraints
      for j := 0 to c^.pointCount - 1 do
         with c^.points[j] do
         begin
            // Relative velocity at contact
            {$IFDEF OP_OVERLOAD}            
            dv := v2 + b2Cross(w2, r2) - v1 - b2Cross(w1, r1);
            {$ELSE}            
            dv := Subtract(v2, v1);
            AddBy(dv, b2Cross(w2, r2));
            SubtractBy(dv, b2Cross(w1, r1));
            {$ENDIF}                        

            // Compute normal impulse
            vn := b2Dot(dv, normal);
            lambda := normalMass * (velocityBias - vn);

            // b2Clamp the accumulated impulse
            newImpulse := b2Max(normalImpulse + lambda, 0.0);
            lambda := newImpulse - normalImpulse;

            // Apply contact impulse
            {$IFDEF OP_OVERLOAD}
            P := lambda * normal;
            v1 := v1 - invMass1 * P;
            v2 := v2 + invMass2 * P;
            {$ELSE}
            P := Multiply(normal, lambda);
            SubtractBy(v1, Multiply(P, invMass1));
            AddBy(v2, Multiply(P, invMass2));
            {$ENDIF}
            w1 := w1 - invI1 * b2Cross(r1, P);
            w2 := w2 + invI2 * b2Cross(r2, P);
            normalImpulse := newImpulse;
         end;

      // Solve tangent constraints
      for j := 0 to c^.pointCount - 1 do
         with c^.points[j] do
         begin
            // Relative velocity at contact
            {$IFDEF OP_OVERLOAD}            
            dv := v2 + b2Cross(w2, r2) - v1 - b2Cross(w1, r1);
            {$ELSE}            
            dv := Subtract(v2, v1);
            AddBy(dv, b2Cross(w2, r2));
            SubtractBy(dv, b2Cross(w1, r1));
            {$ENDIF}                        

            // Compute tangent force
            vt := b2Dot(dv, tangent);
            lambda := tangentMass * (-vt);

            // b2Clamp the accumulated force
            maxFriction := friction * normalImpulse;
            newImpulse := b2Clamp(tangentImpulse + lambda, -maxFriction, maxFriction);
            lambda := newImpulse - tangentImpulse;

            // Apply contact impulse
            {$IFDEF OP_OVERLOAD}
            P := lambda * tangent;
            v1 := v1 - invMass1 * P;
            v2 := v2 + invMass2 * P;
            {$ELSE}            
            P := Multiply(tangent, lambda);
            SubtractBy(v1, Multiply(P, invMass1));
            AddBy(v2, Multiply(P, invMass2));
            {$ENDIF}

            w1 := w1 - invI1 * b2Cross(r1, P);
            w2 := w2 + invI2 * b2Cross(r2, P);

            tangentImpulse := newImpulse;
         end;

      b1.m_linearVelocity := v1;
      b1.m_angularVelocity := w1;
      b2.m_linearVelocity := v2;
      b2.m_angularVelocity := w2;
   end;
end;

procedure Tb2ContactSolver.FinalizeVelocityConstraints;
var
   i, j: Integer;
   m: Pb2Manifold;
   c: Pb2ContactConstraint;
begin
   for i := 0 to m_constraintCount - 1 do
   begin
      {$IFDEF D2009UP}
      c := m_constraints + i;
      {$ELSE}
      c := Pb2ContactConstraint(Integer(m_constraints) + i * SizeOf(Tb2ContactConstraint));
      {$ENDIF}

      m := c^.manifold;

      for j := 0 to c^.pointCount - 1 do
      begin
         m^.points[j].normalImpulse := c^.points[j].normalImpulse;
         m^.points[j].tangentImpulse := c^.points[j].tangentImpulse;
      end;
   end;
end;

function Tb2ContactSolver.SolvePositionConstraints(baumgarte: Float): Boolean;
var
   i, j: Integer;
   b1, b2: Tb2Body;
   c: Pb2ContactConstraint;
   minSeparation: Float;
   invMass1, invI1, invMass2, invI2: Float;
   normal: TVector2;
   _r1, _r2: TVector2;
   p1, p2, dp: TVector2;
   _separation: Float;
   dImpulse, impulse0: Float;
   impulse: TVector2;
begin
   minSeparation := 0.0;

   for i := 0 to m_constraintCount - 1 do
   begin
      {$IFDEF D2009UP}
      c := m_constraints + i;
      {$ELSE}
      c := Pb2ContactConstraint(Integer(m_constraints) + i * SizeOf(Tb2ContactConstraint));
      {$ENDIF}

      b1 := c.body1;
      b2 := c.body2;
      invMass1 := b1.m_mass * b1.m_invMass;
      invI1 := b1.m_mass * b1.m_invI;
      invMass2 := b2.m_mass * b2.m_invMass;
      invI2 := b2.m_mass * b2.m_invI;

      normal := c^.normal;

      // Solver normal constraints
      for j := 0 to c^.pointCount - 1 do
         with c^.points[j] do
         begin
            {$IFDEF OP_OVERLOAD}         
            _r1 := b2Mul(b1.m_xf.R, localAnchor1 - b1.GetLocalCenter);
            _r2 := b2Mul(b2.m_xf.R, localAnchor2 - b2.GetLocalCenter);  
            p1 := b1.m_sweep.c + _r1;
            p2 := b2.m_sweep.c + _r2;
            dp := p2 - p1;
            {$ELSE}            
            _r1 := b2Mul(b1.m_xf.R, Subtract(localAnchor1, b1.GetLocalCenter));
            _r2 := b2Mul(b2.m_xf.R, Subtract(localAnchor2, b2.GetLocalCenter));  
            p1 := Add(b1.m_sweep.c, _r1);
            p2 := Add(b2.m_sweep.c, _r2);
            dp := Subtract(p2, p1);
            {$ENDIF}                        

            // Approximate the current separation.
            _separation := b2Dot(dp, normal) + separation;

            // Track max constraint error.
            minSeparation := b2Min(minSeparation, _separation);

            // Compute normal impulse
            dImpulse := -equalizedMass * baumgarte * b2Clamp(_separation +
               b2_linearSlop, -b2_maxLinearCorrection, 0.0);

            // b2Clamp the accumulated impulse
            impulse0 := positionImpulse;
            positionImpulse := b2Max(impulse0 + dImpulse, 0.0);
            dImpulse := positionImpulse - impulse0;

            {$IFDEF OP_OVERLOAD}            
            impulse := dImpulse * normal;           
            b1.m_sweep.c.SubtractBy(invMass1 * impulse);
            {$ELSE}            
            impulse := Multiply(normal, dImpulse);           
            SubtractBy(b1.m_sweep.c, Multiply(impulse, invMass1));
            {$ENDIF}                        
            b1.m_sweep.a := b1.m_sweep.a - invI1 * b2Cross(_r1, impulse);
            b1.SynchronizeTransform;

            {$IFDEF OP_OVERLOAD}            
            b2.m_sweep.c.AddBy(invMass2 * impulse);
            {$ELSE}            
            AddBy(b2.m_sweep.c, Multiply(impulse, invMass2));
            {$ENDIF}                        
            b2.m_sweep.a := b2.m_sweep.a + invI2 * b2Cross(_r2, impulse);
            b2.SynchronizeTransform;
         end;
   end;

   // We can't expect minSpeparation >= -b2_linearSlop because we don't
   // push the separation above -b2_linearSlop.
   Result := minSeparation >= -1.5 * b2_linearSlop;
end;

{ Tb2NullContact }

procedure Tb2NullContact.Evaluate(listener: Tb2ContactListener);
begin
end;

function Tb2NullContact.GetManifolds: Pb2Manifold;
begin
   Result := nil;
end;

{ Tb2ContactManager }

constructor Tb2ContactManager.Create(world: Tb2World);
begin         
   m_world := world;
   m_destroyImmediate := False;
end;

function Tb2ContactManager.PairAdded(proxyUserData1,
   proxyUserData2: Pointer): Pointer;
var
   shape1, shape2: Tb2Shape;
   body1, body2: Tb2Body;
   c: Tb2Contact;
begin
   shape1 := Tb2Shape(proxyUserData1);
   shape2 := Tb2Shape(proxyUserData2);

   body1 := shape1.GetBody;
   body2 := shape2.GetBody;

   if body1.IsStatic() and body2.IsStatic() then
   begin
      Result := @m_nullContact;
      Exit;
   end;

   if shape1.GetBody = shape2.GetBody then
   begin
      Result := @m_nullContact;
      Exit;
   end;

   if body2.IsConnected(body1) then
   begin
      Result := @m_nullContact;
      Exit;
   end;

   if Assigned(m_world.m_contactFilter) and
      (not m_world.m_contactFilter.ShouldCollide(shape1, shape2)) then
   begin
      Result := @m_nullContact;
      Exit;
   end;

   // Call the factory.
   c := Tb2Contact.CreateContact(shape1, shape2);

   // Contact creation may swap shapes.
   shape1 := c.GetShape1;
   shape2 := c.GetShape2;
   body1 := shape1.GetBody;
   body2 := shape2.GetBody;

   // Insert into the world.
   c.m_prev := nil;
   c.m_next := m_world.m_contactList;
   if Assigned(m_world.m_contactList) then
      m_world.m_contactList.m_prev := c;
   m_world.m_contactList := c;

   // Connect to island graph.
   // Connect to body 1
   c.m_node1.contact := c;
   c.m_node1.other := body2;

   c.m_node1.prev := nil;
   c.m_node1.next := body1.m_contactList;
   if Assigned(body1.m_contactList) then
      body1.m_contactList.prev := @c.m_node1;
   body1.m_contactList := @c.m_node1;

   // Connect to body 2
   c.m_node2.contact := c;
   c.m_node2.other := body1;

   c.m_node2.prev := nil;
   c.m_node2.next := body2.m_contactList;
   if Assigned(body2.m_contactList) then
      body2.m_contactList.prev := @c.m_node2;
   body2.m_contactList := @c.m_node2;

   Inc(m_world.m_contactCount);
   Result := c;
end;

procedure Tb2ContactManager.PairRemoved(proxyUserData1, proxyUserData2,
   pairUserData: Pointer);
var
   c: Tb2Contact;
begin
	 //B2_NOT_USED(proxyUserData1);
	 //B2_NOT_USED(proxyUserData2);

   if not Assigned(pairUserData) then
      Exit;

   c := Tb2Contact(pairUserData);
   if c = @m_nullContact then
      Exit;

   // An attached body is being destroyed, we must destroy this contact
   // immediately to avoid orphaned shape pointers.
   Destroy(c);
end;

procedure Tb2ContactManager.Destroy(c: Tb2Contact);
var
   manifoldCount: Int32;
   manifolds, manifold: Pb2Manifold;
   cp: Tb2ContactPoint;
   b1, b2: Tb2Body;
   i, j: Integer;
   v1, v2: TVector2;
begin
   b1 := c.GetShape1.GetBody;
   b2 := c.GetShape2.GetBody;
   // Inform the user that this contact is ending.
   manifoldCount := c.GetManifoldCount;
   if (manifoldCount > 0) and Assigned(m_world.m_contactListener) then
   begin
      manifolds := c.GetManifolds;
      cp.shape1 := c.GetShape1;
      cp.shape2 := c.GetShape2;
      cp.friction := c.m_friction;
      cp.restitution := c.m_restitution;

      for i := 0 to manifoldCount - 1 do
      begin
         {$IFDEF D2009UP}
         manifold := @manifolds[i];
         {$ELSE}
         manifold := Pb2Manifold(Integer(manifolds) + i * SizeOf(Tb2Manifold));
         {$ENDIF}

         cp.normal := manifold^.normal;

         for j := 0 to manifold^.pointCount - 1 do
            with manifold^.points[j] do
            begin
               cp.position := b1.GetWorldPoint(localPoint1);
               v1 := b1.GetLinearVelocityFromLocalPoint(localPoint1);
               v2 := b2.GetLinearVelocityFromLocalPoint(localPoint2);
               {$IFDEF OP_OVERLOAD}               
               cp.velocity := v2 - v1;
               {$ELSE}               
               cp.velocity := Subtract(v2, v1);
               {$ENDIF}                              
               cp.separation := separation;
               cp.id := id;
               m_world.m_contactListener.Remove(cp);
            end;
      end;
   end;

   // Remove from the world.
   if Assigned(c.m_prev) then
      c.m_prev.m_next := c.m_next;

   if Assigned(c.m_next) then
      c.m_next.m_prev := c.m_prev;

   if (c = m_world.m_contactList) then
      m_world.m_contactList := c.m_next;

   // Remove from body 1
   if Assigned(c.m_node1.prev) then
      c.m_node1.prev.next := c.m_node1.next;

   if Assigned(c.m_node1.next) then
      c.m_node1.next.prev := c.m_node1.prev;

   if @c.m_node1 = b1.m_contactList then
      b1.m_contactList := c.m_node1.next;

   // Remove from body 2
   if Assigned(c.m_node2.prev) then
      c.m_node2.prev.next := c.m_node2.next;

   if Assigned(c.m_node2.next) then
      c.m_node2.next.prev := c.m_node2.prev;

   if @c.m_node2 = b2.m_contactList then
      b2.m_contactList := c.m_node2.next;

   c.Free;
   Dec(m_world.m_contactCount);
end;

procedure Tb2ContactManager.Collide;
var
   c: Tb2Contact;
begin
   // Update awake contacts.
   c := m_world.m_contactList;
   while Assigned(c) do
   begin
      if c.GetShape1.GetBody.IsSleeping and
         c.GetShape2.GetBody.IsSleeping then
      begin
         c := c.GetNext;
         Continue;
      end;

      c.Update(m_world.m_contactListener);
      c := c.GetNext;
   end;
end;

////////////////////////////////////////////////////
// Island
{ Tb2Island }

(*
Position Correction Notes
=========================
I tried the several algorithms for position correction of the 2D revolute joint.
I looked at these systems:
- simple pendulum (1m diameter sphere on massless 5m stick) with initial angular velocity of 100 rad/s.
- suspension bridge with 30 1m long planks of length 1m.
- multi-link chain with 30 1m long links.

Here are the algorithms:

Baumgarte - A fraction of the position error is added to the velocity error. There is no
separate position solver.

Pseudo Velocities - After the velocity solver and position integration,
the position error, Jacobian, and effective mass are recomputed. Then
the velocity constraints are solved with pseudo velocities and a fraction
of the position error is added to the pseudo velocity error. The pseudo
velocities are initialized to zero and there is no warm-starting. After
the position solver, the pseudo velocities are added to the positions.
This is also called the First Order World method or the Position LCP method.

Modified Nonlinear Gauss-Seidel (NGS) - Like Pseudo Velocities except the
position error is re-computed for each constraint and the positions are updated
after the constraint is solved. The radius vectors (aka Jacobians) are
re-computed too (otherwise the algorithm has horrible instability). The pseudo
velocity states are not needed because they are effectively zero at the beginning
of each iteration. Since we have the current position error, we allow the
iterations to terminate early if the error becomes smaller than b2_linearSlop.

Full NGS or just NGS - Like Modified NGS except the effective mass are re-computed
each time a constraint is solved.

Here are the results:
Baumgarte - this is the cheapest algorithm but it has some stability problems,
especially with the bridge. The chain links separate easily close to the root
and they jitter as they struggle to pull together. This is one of the most common
methods in the field. The big drawback is that the position correction artificially
affects the momentum, thus leading to instabilities and False bounce. I used a
bias factor of 0.2. A larger bias factor makes the bridge less stable, a smaller
factor makes joints and contacts more spongy.

Pseudo Velocities - the is more stable than the Baumgarte method. The bridge is
stable. However, joints still separate with large angular velocities. Drag the
simple pendulum in a circle quickly and the joint will separate. The chain separates
easily and does not recover. I used a bias factor of 0.2. A larger value lead to
the bridge collapsing when a heavy cube drops on it.

Modified NGS - this algorithm is better in some ways than Baumgarte and Pseudo
Velocities, but in other ways it is worse. The bridge and chain are much more
stable, but the simple pendulum goes unstable at high angular velocities.

Full NGS - stable in all tests. The joints display good stiffness. The bridge
still sags, but this is better than infinite forces.

Recommendations
Pseudo Velocities are not really worthwhile because the bridge and chain cannot
recover from joint separation. In other cases the benefit over Baumgarte is small.

Modified NGS is not a robust method for the revolute joint due to the violent
instability seen in the simple pendulum. Perhaps it is viable with other constraint
types, especially scalar constraints where the effectivprocedure Tb2Island.Solve(const step: Tb2TimeStep; const gravity: TVector2;
  correctPositions, allowSleep: Boolean);
begin

end;

procedure Tb2Island.SolveTOI(const subStep: Tb2TimeStep);
begin

end;
e mass is a scalar.

This leaves Baumgarte and Full NGS. Baumgarte has small, but manageable instabilities
and is very fast. I don't think we can escape Baumgarte, especially in highly
demanding cases where high constraint fidelity is not needed.

Full NGS is robust and easy on the eyes. I recommend this as an option for
higher fidelity simulation and certainly for suspension bridges and long chains.
Full NGS might be a good choice for ragdolls, especially motorized ragdolls where
joint separation can be problematic. The number of NGS iterations can be reduced
for better performance without harming robustness much.

Each joint in a can be handled differently in the position solver. So I recommend
a system where the user can select the algorithm on a per joint basis. I would
probably default to the slower Full NGS and let the user select the faster
Baumgarte method in performance critical scenarios.
*)

constructor Tb2Island.Create(bodyCapacity, contactCapacity, jointCapacity: Int32;
   listener: Tb2ContactListener);
begin
   m_bodyCapacity := bodyCapacity;
   m_contactCapacity := contactCapacity;
   m_jointCapacity	 := jointCapacity;
   m_bodyCount := 0;
   m_contactCount := 0;
   m_jointCount := 0;

   m_listener := listener;

   m_bodies := TList.Create;
   m_bodies.Count := bodyCapacity;
   m_contacts := TList.Create;
   m_contacts.Count := contactCapacity;
   m_joints := TList.Create;
   m_joints.Count := jointCapacity;
   m_positionIterationCount := 0;
end;

destructor Tb2Island.Destroy;
begin
   // Warning: the order should reverse the constructor order.
   m_joints.Free;
   m_contacts.Free;
   m_bodies.Free;
   inherited;
end;

procedure Tb2Island.Clear;
begin
   m_bodyCount := 0;
   m_contactCount := 0;
   m_jointCount := 0;
end;

procedure Tb2Island.Solve(const step: Tb2TimeStep; const gravity: TVector2;
   correctPositions, allowSleep: Boolean);
const
   linTolSqr = b2_linearSleepTolerance * b2_linearSleepTolerance;
   angTolSqr = b2_angularSleepTolerance * b2_angularSleepTolerance;
var
   i, j: Integer;
   b: Tb2Body;
   contactSolver: Tb2ContactSolver;
   contactsOkay, jointsOkay: Boolean;
   minSleepTime: Float;
begin
   // Integrate velocities and apply damping.
   for i := 0 to m_bodyCount - 1 do
   begin
      b := Tb2Body(m_bodies[i]);

      with b do
      begin
         if IsStatic() then
            Continue;

         // Integrate velocities.
         {$IFDEF OP_OVERLOAD}         
         m_linearVelocity.AddBy(step.dt * (gravity + m_invMass * m_force));
         {$ELSE}
         AddBy(m_linearVelocity, Multiply(UPhysics2DTypes.Add(gravity, 
            Multiply(m_force, m_invMass)), step.dt));
         {$ENDIF}                  
         m_angularVelocity := m_angularVelocity + step.dt * m_invI * m_torque;

         // Reset forces.
         SetZero(m_force);
         m_torque := 0.0;

         // Apply damping.
         // ODE: dv/dt + c * v = 0
         // Solution: v(t) = v0 * exp(-c * t)
         // Time step: v(t + dt) = v0 * exp(-c * (t + dt)) = v0 * exp(-c * t) * exp(-c * dt) = v * exp(-c * dt)
         // v2 = exp(-c * dt) * v1
         // Taylor expansion:
         // v2 = (1.0f - c * dt) * v1
         {$IFDEF OP_OVERLOAD}         
         m_linearVelocity.MultiplyBy(b2Clamp(1.0 - step.dt * m_linearDamping, 0.0, 1.0));
         {$ELSE}         
         MultiplyBy(m_linearVelocity, b2Clamp(1.0 - step.dt * m_linearDamping, 0.0, 1.0));
         {$ENDIF}                  
         m_angularVelocity := m_angularVelocity * b2Clamp(1.0 - step.dt *
            b.m_angularDamping, 0.0, 1.0);

         // Check for large velocities.
         if (b2Dot(m_linearVelocity, m_linearVelocity) > b2_maxLinearVelocitySquared) then
         begin
            {$IFDEF OP_OVERLOAD}         
            m_linearVelocity.Normalize;
            m_linearVelocity.MultiplyBy(b2_maxLinearVelocity);
            {$ELSE}            
            Normalize(m_linearVelocity);
            MultiplyBy(m_linearVelocity, b2_maxLinearVelocity);
            {$ENDIF}                        
         end;

         if m_angularVelocity * m_angularVelocity > b2_maxAngularVelocitySquared then
            if m_angularVelocity < 0.0 then
               m_angularVelocity := -b2_maxAngularVelocity
            else
               m_angularVelocity := b2_maxAngularVelocity;
      end;
   end;

   contactSolver := Tb2ContactSolver.Create(step, m_contacts, m_contactCount);

   try
      // Initialize velocity constraints.
      contactSolver.InitVelocityConstraints(step);

      for i := 0 to m_jointCount - 1 do
         Tb2Joint(m_joints[i]).InitVelocityConstraints(step);

      // Solve velocity constraints.
      if (contactSolver.m_constraintCount > 0) or (m_jointCount > 0) then
         for i := 0 to step.maxIterations - 1 do
         begin
            contactSolver.SolveVelocityConstraints;
            for j := 0 to m_jointCount - 1 do
               Tb2Joint(m_joints[j]).SolveVelocityConstraints(step);
         end;

      // Post-solve (store impulses for warm starting).
      contactSolver.FinalizeVelocityConstraints;

      // Integrate positions.
      for i := 0 to m_bodyCount - 1 do
      begin
         b := Tb2Body(m_bodies[i]);

         with b do
         begin
            if IsStatic() then
               Continue;

            // Store positions for continuous collision.
            m_sweep.c0 := m_sweep.c;
            m_sweep.a0 := m_sweep.a;

            // Integrate
            {$IFDEF OP_OVERLOAD}            
            m_sweep.c.AddBy(step.dt * m_linearVelocity);
            {$ELSE}            
            AddBy(m_sweep.c, Multiply(m_linearVelocity, step.dt));
            {$ENDIF}                        
            m_sweep.a := m_sweep.a + step.dt * m_angularVelocity;

            // Compute new transform
            SynchronizeTransform;
            // Note: shapes are synchronized later.
         end;
      end;

      if correctPositions then
      begin
         // Initialize position constraints.
         // Contacts don't need initialization.
         for i := 0 to m_jointCount - 1 do
            Tb2Joint(m_joints[i]).InitPositionConstraints;

         // Iterate over constraints.
         m_positionIterationCount := 0;
         while (m_positionIterationCount < step.maxIterations) do
         begin
            contactsOkay := contactSolver.SolvePositionConstraints(b2_contactBaumgarte);
            jointsOkay := True;

            for i := 0 to m_jointCount - 1 do
               jointsOkay := Tb2Joint(m_joints[i]).SolvePositionConstraints and jointsOkay;

            if contactsOkay and jointsOkay then
               Break;

            Inc(m_positionIterationCount);
         end;
      end;

      Report(contactSolver.m_constraints);

      if allowSleep then
      begin
         minSleepTime := FLT_MAX;

         for i := 0 to m_bodyCount - 1 do
         begin
            b := Tb2Body(m_bodies[i]);

            with b do
            begin
               if m_invMass = 0.0 then
                  Continue;

               if (m_flags and e_allowSleepFlag) = 0 then
               begin
                  m_sleepTime := 0.0;
                  minSleepTime := 0.0;
               end;

               if ((m_flags and e_allowSleepFlag) = 0) or
                  (m_angularVelocity * m_angularVelocity > angTolSqr) or
                  (b2Dot(m_linearVelocity, m_linearVelocity) > linTolSqr) then
               begin
                  m_sleepTime := 0.0;
                  minSleepTime := 0.0;
               end
               else
               begin
                  m_sleepTime := m_sleepTime + step.dt;
                  minSleepTime := b2Min(minSleepTime, m_sleepTime);
               end;
            end;
         end;

         if minSleepTime >= b2_timeToSleep then
            for i := 0 to m_bodyCount - 1 do
            begin
               b := Tb2Body(m_bodies[i]);
               with b do
               begin
                  m_flags := m_flags or e_sleepFlag;
                  m_linearVelocity := b2Vec2_zero;
                  m_angularVelocity := 0.0;
               end;
            end;
      end;
   finally
      contactSolver.Free;
   end;
end;

procedure Tb2Island.SolveTOI(const subStep: Tb2TimeStep);
var
   b: Tb2Body;
   i: Integer;
   contactSolver: Tb2ContactSolver;
begin
   contactSolver := Tb2ContactSolver.Create(subStep, m_contacts, m_contactCount);

   try
      // No warm starting needed for TOI events.
      // Solve velocity constraints.
      for i := 0 to subStep.maxIterations - 1 do
         contactSolver.SolveVelocityConstraints;

      // Don't store the TOI contact forces for warm starting
      // because they can be quite large.

      // Integrate positions.
      for i := 0 to m_bodyCount - 1 do
      begin
         b := Tb2Body(m_bodies[i]);

         with b do
         begin
            if IsStatic() then
              Continue;

            // Store positions for continuous collision.
            m_sweep.c0 := m_sweep.c;
            m_sweep.a0 := m_sweep.a;

            // Integrate
            {$IFDEF OP_OVERLOAD}
            m_sweep.c.AddBy(subStep.dt * m_linearVelocity);
            {$ELSE}            
            AddBy(m_sweep.c, Multiply(m_linearVelocity, subStep.dt));
            {$ENDIF}                        
            m_sweep.a := m_sweep.a + subStep.dt * m_angularVelocity;

            // Compute new transform
            SynchronizeTransform;
            // Note: shapes are synchronized later.
         end;
      end;

      // Solve position constraints.
      for i := 0 to subStep.maxIterations - 1 do
         if contactSolver.SolvePositionConstraints(0.75) then
            Break;

      Report(contactSolver.m_constraints);
   finally
      contactSolver.Free;
   end;
end;

procedure Tb2Island.Add(body: Tb2Body);
begin
	 //b2Assert(m_bodyCount < m_bodyCapacity);
   m_bodies[m_bodyCount] := body;
   Inc(m_bodyCount);
end;

procedure Tb2Island.Add(contact: Tb2Contact);
begin
	 //b2Assert(m_contactCount < m_contactCapacity);
   m_contacts[m_contactCount] := contact;
   Inc(m_contactCount);
end;

procedure Tb2Island.Add(joint: Tb2Joint);
begin
	 //b2Assert(m_jointCount < m_jointCapacity);
   m_joints[m_jointCount] := joint;
   Inc(m_jointCount);
end;

procedure Tb2Island.Report(constraints: Pb2ContactConstraint);
var
   i, j, k: Integer;
   c: Tb2Contact;
   cc: Pb2ContactConstraint;
   cr: Tb2ContactResult;
   b1: Tb2Body;
   manifoldCount: Int32;
   manifolds, manifold: Pb2Manifold;
   point: Pb2ManifoldPoint;
begin
   if not Assigned(m_listener) then
      Exit;

   for i := 0 to m_contactCount - 1 do
   begin
      c := Tb2Contact(m_contacts[i]);
      {$IFDEF D2009UP}
      cc := constraints + i;
      {$ELSE}
      cc := Pb2ContactConstraint(Integer(constraints) + i * SizeOf(Tb2ContactConstraint));
      {$ENDIF}

      cr.shape1 := c.GetShape1;
      cr.shape2 := c.GetShape2;
      b1 := cr.shape1.GetBody;
      manifoldCount := c.GetManifoldCount;
      manifolds := c.GetManifolds;
      for j := 0 to manifoldCount - 1 do
      begin
         {$IFDEF D2009UP}
         manifold := manifolds + j;
         {$ELSE}
         manifold := Pb2Manifold(Integer(manifolds) + j * SizeOf(Tb2Manifold));
         {$ENDIF}
         cr.normal := manifold.normal;

         for k := 0 to manifold^.pointCount - 1 do
         begin
            point := @manifold.points[k];
            with cc.points[k] do
            begin
               cr.position := b1.GetWorldPoint(point^.localPoint1);

               // TOI constraint results are not stored, so get
               // the result from the constraint.
               cr.normalImpulse := normalImpulse;
               cr.tangentImpulse := tangentImpulse;
               cr.id := point^.id;
               m_listener.Result(cr);
            end;
         end;
      end;
   end;
end;

////////////////////////////////////////////////////
//

{$IFDEF OP_OVERLOAD}
function Tb2Bound.IsLower: Boolean;
begin
   Result := (value and 1) = 0;
end;

function Tb2Bound.IsUpper: Boolean;
begin
   Result := (value and 1) = 1;
end;
{$ENDIF}

{$IFDEF OP_OVERLOAD}
function Tb2Proxy.GetNext: UInt16;
begin
   Result := lowerBounds[0];
end;

procedure Tb2Proxy.SetNext(Next: UInt16);
begin
   lowerBounds[0] := Next;
end;

function Tb2Proxy.IsValid: Boolean;
begin
   Result := overlapCount <> b2_invalid;
end;
{$ENDIF}

{ Tb2BroadPhase }

constructor Tb2BroadPhase.Create(const worldAABB: Tb2AABB; callback: Tb2PairCallback);
var
   d: TVector2;
   i: Integer;
begin
   m_pairManager := Tb2PairManager.Create;
   m_pairManager.Initialize(Self, callback);

   //b2Assert(worldAABB.IsValid());
   m_worldAABB := worldAABB;
   m_proxyCount := 0;

   {$IFDEF OP_OVERLOAD}   
   d := worldAABB.upperBound - worldAABB.lowerBound;
   {$ELSE}   
   d := Subtract(worldAABB.upperBound, worldAABB.lowerBound);
   {$ENDIF}      
   m_quantizationFactor.x := B2BROADPHASE_MAX / d.x;
   m_quantizationFactor.y := B2BROADPHASE_MAX / d.y;

   for i := 0 to b2_maxProxies - 2 do
      with m_proxyPool[i] do
      begin
         {$IFDEF OP_OVERLOAD}      
         SetNext(i + 1);
         {$ELSE}         
         //SetNext(m_proxyPool[i], i + 1);
         lowerBounds[0] := i + 1;
         {$ENDIF}                  
         timeStamp := 0;
         overlapCount := b2_invalid;
         userData := nil;
      end;

   with m_proxyPool[b2_maxProxies - 1] do
   begin
      {$IFDEF OP_OVERLOAD}   
      SetNext(b2_nullProxy);
      {$ELSE}      
      //SetNext(m_proxyPool[b2_maxProxies - 1], b2_nullProxy);
      lowerBounds[0] := b2_nullProxy;
      {$ENDIF}            
      timeStamp := 0;
      overlapCount := b2_invalid;
      userData := nil;
   end;
   m_freeProxy := 0;

   m_timeStamp := 1;
   m_queryResultCount := 0;
end;

destructor Tb2BroadPhase.Destroy;
begin
   m_pairManager.Free;
   inherited;
end;

procedure Tb2BroadPhase.ComputeBounds(var lowerValues, upperValues: Tb2BoundValuesArray;
   const aabb: Tb2AABB);
var
   minVertex, maxVertex: TVector2;
begin
   //b2Assert(aabb.upperBound.x > aabb.lowerBound.x);
   //b2Assert(aabb.upperBound.y > aabb.lowerBound.y);

   minVertex := b2Clamp(aabb.lowerBound, m_worldAABB.lowerBound, m_worldAABB.upperBound);
   maxVertex := b2Clamp(aabb.upperBound, m_worldAABB.lowerBound, m_worldAABB.upperBound);

   // Bump lower bounds downs and upper bounds up. This ensures correct sorting of
   // lower/upper bounds that would have equal values.
   // TODO_ERIN implement fast float to uint16 conversion.
   lowerValues[0] := UInt16(Trunc((m_quantizationFactor.x * (minVertex.x -
      m_worldAABB.lowerBound.x)))) and (B2BROADPHASE_MAX - 1);
   upperValues[0] := UInt16(Trunc((m_quantizationFactor.x * (maxVertex.x -
      m_worldAABB.lowerBound.x)))) or 1;

   lowerValues[1] := UInt16(Trunc((m_quantizationFactor.y * (minVertex.y -
      m_worldAABB.lowerBound.y)))) and (B2BROADPHASE_MAX - 1);
   upperValues[1] := UInt16(Trunc((m_quantizationFactor.y * (maxVertex.y -
      m_worldAABB.lowerBound.y)))) or 1;
end;

function Tb2BroadPhase.TestOverlap(var p1, p2: Tb2Proxy): Boolean;
var
   axis: Integer;
   bounds: Pb2AxialBoundsArray;
begin
   for axis := 0 to 1 do
   begin
      bounds := @m_bounds[axis];
      //b2Assert(p1->lowerBounds[axis] < 2 * m_proxyCount);
      //b2Assert(p1->upperBounds[axis] < 2 * m_proxyCount);
      //b2Assert(p2->lowerBounds[axis] < 2 * m_proxyCount);
      //b2Assert(p2->upperBounds[axis] < 2 * m_proxyCount);

      if bounds^[p1.lowerBounds[axis]].value > bounds^[p2.upperBounds[axis]].value then
      begin
         Result := False;
         Exit;
      end;

      if bounds^[p1.upperBounds[axis]].value < bounds^[p2.lowerBounds[axis]].value then
      begin
         Result := False;
         Exit;
      end;
   end;
   Result := True;
end;

function Tb2BroadPhase.TestOverlap(const b: Tb2BoundValues; var p: Tb2Proxy): Boolean;
var
   axis: Integer;
   bounds: Pb2AxialBoundsArray;
begin
   for axis := 0 to 1 do
   begin
      bounds := @m_bounds[axis];

      //b2Assert(p->lowerBounds[axis] < 2 * m_proxyCount);
      //b2Assert(p->upperBounds[axis] < 2 * m_proxyCount);

      if b.lowerValues[axis] > bounds^[p.upperBounds[axis]].value then
      begin
         Result := False;
         Exit;
      end;

      if b.upperValues[axis] < bounds^[p.lowerBounds[axis]].value then
      begin
         Result := False;
         Exit;
      end;
   end;
   Result := True;
end;

procedure Tb2BroadPhase.Query(var lowerIndex, upperIndex: Int32; lowerValue,
   upperValue: UInt16; var bounds: Tb2AxialBoundsArray; boundCount, axis: Int32);
var
   i, s: Integer;
   lowerQuery, upperQuery: Int32;
begin
   lowerQuery := BroadPhase_BinarySearch(bounds, boundCount, lowerValue);
   upperQuery := BroadPhase_BinarySearch(bounds, boundCount, upperValue);

   // Easy case: lowerQuery <= lowerIndex(i) < upperQuery
   // Solution: search query range for min bounds.
   for i := lowerQuery to upperQuery - 1 do
   begin
      {$IFDEF OP_OVERLOAD}
      if bounds[i].IsLower() then
      {$ELSE}      
      if IsLower(bounds[i]) then
      {$ENDIF}            
        IncrementOverlapCount(bounds[i].proxyId);
   end;

   // Hard case: lowerIndex(i) < lowerQuery < upperIndex(i)
   // Solution: use the stabbing count to search down the bound array.
   if lowerQuery > 0 then
   begin
      i := lowerQuery - 1;
      s := bounds[i].stabbingCount;

      // Find the s overlaps.
      while (s > 0) do  //b2Assert(i >= 0);
         with bounds[i] do
         begin
            {$IFDEF OP_OVERLOAD}
            if IsLower() then
            {$ELSE}            
            if IsLower(bounds[i]) then
            {$ENDIF}                        
               if lowerQuery <= m_proxyPool[proxyId].upperBounds[axis] then
               begin
                  IncrementOverlapCount(proxyId);
                  Dec(s);
               end;
            Dec(i);
         end;
   end;
   lowerIndex := lowerQuery;
   upperIndex := upperQuery;
end;

procedure Tb2BroadPhase.IncrementOverlapCount(proxyId: Int32);
begin
   with m_proxyPool[proxyId] do
      if timeStamp < m_timeStamp then
      begin
         timeStamp := m_timeStamp;
         overlapCount := 1;
      end
      else
      begin
         overlapCount := 2;
         //b2Assert(m_queryResultCount < b2_maxProxies);
         m_queryResults[m_queryResultCount] := proxyId;
         Inc(m_queryResultCount);
      end;
end;

procedure Tb2BroadPhase.IncrementTimeStamp;
var
   i: Integer;
begin
   if m_timeStamp = B2BROADPHASE_MAX then
   begin
      for i := 0 to b2_maxProxies - 1 do
         m_proxyPool[i].timeStamp := 0;
      m_timeStamp := 1;
   end
   else
      Inc(m_timeStamp);
end;

function Tb2BroadPhase.InRange(const aabb: Tb2AABB): Boolean;
var
   d: TVector2;
begin
   {$IFDEF OP_OVERLOAD}
   d := b2Max(aabb.lowerBound - m_worldAABB.upperBound,
      m_worldAABB.lowerBound - aabb.upperBound);
   {$ELSE}      
   d := b2Max(Subtract(aabb.lowerBound, m_worldAABB.upperBound),
      Subtract(m_worldAABB.lowerBound, aabb.upperBound));
   {$ENDIF}      
   Result := b2Max(d.x, d.y) < 0.0;
end;

function Tb2BroadPhase.CreateProxy(const aabb: Tb2AABB; userData: Pointer): UInt16;
var
   AproxyId: UInt16;
   proxy: Pb2Proxy;
   index, boundCount, axis, lowerIndex, upperIndex: Int32;
   lowerValues, upperValues: Tb2BoundValuesArray;
   bounds: Pb2AxialBoundsArray;
begin
   //b2Assert(m_proxyCount < b2_maxProxies);
   //b2Assert(m_freeProxy != b2_nullProxy);

   AproxyId := m_freeProxy;
   proxy := @(m_proxyPool[AproxyId]);
   {$IFDEF OP_OVERLOAD}   
   m_freeProxy := proxy^.GetNext;
   {$ELSE}   
   m_freeProxy := GetNext(proxy^);
   {$ENDIF}      

   proxy^.overlapCount := 0;
   proxy^.userData := userData;

   boundCount := 2 * m_proxyCount;
   ComputeBounds(lowerValues, upperValues, aabb);

   for axis := 0 to 1 do
   begin
      bounds := @m_bounds[axis];
      Query(lowerIndex, upperIndex, lowerValues[axis], upperValues[axis],
         bounds^, boundCount, axis);

      Move(bounds^[upperIndex], bounds^[upperIndex + 2],
         (boundCount - upperIndex) * SizeOf(Tb2Bound));
      Move(bounds^[lowerIndex], bounds^[lowerIndex + 1],
         (upperIndex - lowerIndex) * SizeOf(Tb2Bound));

      // The upper index has increased because of the lower bound insertion.
      Inc(upperIndex);

      // Copy in the new bounds.
      with bounds^[lowerIndex] do
      begin
         value := lowerValues[axis];
         proxyId := AproxyId;
         if lowerIndex = 0 then
            stabbingCount := 0
         else
            stabbingCount := bounds^[lowerIndex - 1].stabbingCount;
      end;

      with bounds^[upperIndex] do
      begin
         value := upperValues[axis];
         proxyId := AproxyId;
         stabbingCount := bounds^[upperIndex - 1].stabbingCount;
      end;

      // Adjust the stabbing count between the new bounds.
      for index := lowerIndex to upperIndex - 1 do
         Inc(bounds^[index].stabbingCount);

      // Adjust the all the affected bound indices.
      for index := lowerIndex to boundCount + 1 do
      begin
         proxy := @m_proxyPool[bounds^[index].proxyId];
         {$IFDEF OP_OVERLOAD}
         if bounds^[index].IsLower() then
         {$ELSE}         
         if IsLower(bounds^[index]) then
         {$ENDIF}                  
            proxy^.lowerBounds[axis] := index
         else
            proxy^.upperBounds[axis] := index;
      end;
   end;

   Inc(m_proxyCount);

   //b2Assert(m_queryResultCount < b2_maxProxies);

   // Create pairs if the AABB is in range.
   for index := 0 to m_queryResultCount - 1 do
   begin
      //b2Assert(m_queryResults[i] < b2_maxProxies);
      //b2Assert(m_proxyPool[m_queryResults[i]].IsValid());

      m_pairManager.AddBufferedPair(AproxyId, m_queryResults[index]);
   end;

   m_pairManager.Commit;

   {$IFDEF CLASSVAR_AVAIL}
   if s_validate then
   {$ELSE}
   if b2BroadPhase_s_validate then
   {$ENDIF}
      Validate;

   // Prepare for next query.
   m_queryResultCount := 0;
   IncrementTimeStamp;

   Result := AproxyId;
end;

procedure Tb2BroadPhase.DestroyProxy(proxyId: Int32);
var
   proxy, proxy2: Pb2Proxy;
   i, boundCount: Int32;
   axis: Integer;
   bounds: Pb2AxialBoundsArray;
   lowerIndex, upperIndex: Int32;
   lowerValue, upperValue: UInt16;
begin
   //b2Assert(0 < m_proxyCount && m_proxyCount <= b2_maxProxies);
   proxy := @m_proxyPool[proxyId];
   //b2Assert(proxy->IsValid());

   boundCount := 2 * m_proxyCount;

   for axis := 0 to 1 do
   begin
      bounds := @m_bounds[axis];
      lowerIndex := proxy^.lowerBounds[axis];
      upperIndex := proxy^.upperBounds[axis];
      lowerValue := bounds^[lowerIndex].value;
      upperValue := bounds^[upperIndex].value;

      Move(bounds^[lowerIndex + 1], bounds^[lowerIndex],
         (upperIndex - lowerIndex - 1) * SizeOf(Tb2Bound));

      Move(bounds^[upperIndex + 1], bounds^[upperIndex - 1],
         (boundCount - upperIndex - 1) * SizeOf(Tb2Bound));

      // Fix bound indices.
      for i := lowerIndex to boundCount - 3 do
      begin
         proxy2 := @m_proxyPool[bounds^[i].proxyId];
         {$IFDEF OP_OVERLOAD}
         if bounds^[i].IsLower() then
         {$ELSE}         
         if IsLower(bounds^[i]) then
         {$ENDIF}                  
            proxy2^.lowerBounds[axis] := i
         else
            proxy2^.upperBounds[axis] := i;
      end;

      // Fix stabbing count.
      for i := lowerIndex to upperIndex - 2 do
         Dec(bounds^[i].stabbingCount);

      // Query for pairs to be removed. lowerIndex and upperIndex are not needed.
      Query(lowerIndex, upperIndex, lowerValue, upperValue,
         bounds^, boundCount - 2, axis);
   end;

   //b2Assert(m_queryResultCount < b2_maxProxies);

   for i := 0 to m_queryResultCount - 1 do
   begin
      //b2Assert(m_proxyPool[m_queryResults[i]].IsValid());
      m_pairManager.RemoveBufferedPair(proxyId, m_queryResults[i]);
   end;

   m_pairManager.Commit;

   // Prepare for next query.
   m_queryResultCount := 0;
   IncrementTimeStamp;

   // Return the proxy to the pool.
   with proxy^ do
   begin
      userData := nil;
      overlapCount := b2_invalid;
      lowerBounds[0] := b2_invalid;
      lowerBounds[1] := b2_invalid;
      upperBounds[0] := b2_invalid;
      upperBounds[1] := b2_invalid;
   end;

   {$IFDEF OP_OVERLOAD}   
   proxy^.SetNext(m_freeProxy);
   {$ELSE}   
   SetNext(proxy^, m_freeProxy);
   {$ENDIF}      
   m_freeProxy := proxyId;
   Dec(m_proxyCount);

   {$IFDEF CLASSVAR_AVAIL}
   if s_validate then
   {$ELSE}
   if b2BroadPhase_s_validate then
   {$ENDIF}
      Validate;
end;

procedure Tb2BroadPhase.MoveProxy(proxyId: Int32; const aabb: Tb2AABB);
var
   axis: Integer;
   index, boundCount, prevProxyId, nextProxyId: Int32;
   lowerIndex, upperIndex, deltaLower, deltaUpper: Int32;
   lowerValue, upperValue: UInt16;   
   proxy, prevProxy, nextProxy: Pb2Proxy;
   newValues, oldValues: Tb2BoundValues;
   bounds: Pb2AxialBoundsArray;
   bound, prevBound, nextBound: Pb2Bound;
begin
   if (proxyId = b2_nullProxy) or (b2_maxProxies <= proxyId) then
   begin
      //b2Assert(False);
      Exit;
   end;

   {$IFDEF OP_OVERLOAD}   
   if not aabb.IsValid() then
   {$ELSE}   
   if not IsValid(aabb) then
   {$ENDIF}      
   begin
      //b2Assert(False);
      Exit;;
   end;

   boundCount := 2 * m_proxyCount;
   proxy := @m_proxyPool[proxyId];

   // Get new bound values
   ComputeBounds(newValues.lowerValues, newValues.upperValues, aabb);

   // Get old bound values
   for axis := 0 to 1 do
   begin
      oldValues.lowerValues[axis] := m_bounds[axis][proxy^.lowerBounds[axis]].value;
      oldValues.upperValues[axis] := m_bounds[axis][proxy^.upperBounds[axis]].value;
   end;

   for axis := 0 to 1 do
   begin
      bounds := @m_bounds[axis];

      lowerIndex := proxy^.lowerBounds[axis];
      upperIndex := proxy^.upperBounds[axis];

      lowerValue := newValues.lowerValues[axis];
      upperValue := newValues.upperValues[axis];

      deltaLower := lowerValue - bounds^[lowerIndex].value;
      deltaUpper := upperValue - bounds^[upperIndex].value;
      bounds^[lowerIndex].value := lowerValue;
      bounds^[upperIndex].value := upperValue;

      // Expanding adds overlaps
      // Should we move the lower bound down?
      if deltaLower < 0 then
      begin
         index := lowerIndex;
         while ((index > 0) and (lowerValue < bounds^[index - 1].value)) do
         begin
            bound := @bounds^[index];
            prevBound := @bounds^[index - 1];
            prevProxyId := prevBound^.proxyId;
            prevProxy := @m_proxyPool[prevProxyId];

            Inc(prevBound^.stabbingCount);

            {$IFDEF OP_OVERLOAD}            
            if prevBound^.IsUpper() then
            {$ELSE}            
            if IsUpper(prevBound^) then
            {$ENDIF}                        
            begin
              if TestOverlap(newValues, prevProxy^) then
                 m_pairManager.AddBufferedPair(proxyId, prevProxyId);

               Inc(prevProxy^.upperBounds[axis]);
               Inc(bound^.stabbingCount);
            end
            else
            begin
               Inc(prevProxy^.lowerBounds[axis]);
               Dec(bound^.stabbingCount);
            end;

            Dec(proxy^.lowerBounds[axis]);
            b2Swap(bound^, prevBound^);
            Dec(index);
         end;
      end;

      // Should we move the upper bound up?
      if deltaUpper > 0 then
      begin
         index := upperIndex;
         while ((index < boundCount - 1) and (bounds^[index + 1].value <= upperValue)) do
         begin
            bound := @bounds^[index];
            nextBound := @bounds^[index + 1];
            nextProxyId := nextBound^.proxyId;
            nextProxy := @m_proxyPool[nextProxyId];
            Inc(nextBound^.stabbingCount);

            {$IFDEF OP_OVERLOAD}
            if nextBound^.IsLower() then
            {$ELSE}            
            if IsLower(nextBound^) then
            {$ENDIF}                        
            begin
               if TestOverlap(newValues, nextProxy^) then
                  m_pairManager.AddBufferedPair(proxyId, nextProxyId);

               Dec(nextProxy^.lowerBounds[axis]);
               Inc(bound^.stabbingCount);
            end
            else
            begin
               Dec(nextProxy^.upperBounds[axis]);
               Dec(bound^.stabbingCount);
            end;

            Inc(proxy^.upperBounds[axis]);
            b2Swap(bound^, nextBound^);
            Inc(index);
         end;
      end;

      // Shrinking removes overlaps
      // Should we move the lower bound up?
      if deltaLower > 0 then
      begin
         index := lowerIndex;
         while (index < boundCount - 1) and (bounds^[index + 1].value <= lowerValue) do
         begin
            bound := @bounds^[index];
            nextBound := @bounds^[index + 1];
            nextProxyId := nextBound^.proxyId;
            nextProxy := @m_proxyPool[nextProxyId];

            Dec(nextBound^.stabbingCount);

            {$IFDEF OP_OVERLOAD}            
            if nextBound^.IsUpper() then
            {$ELSE}            
            if IsUpper(nextBound^) then
            {$ENDIF}                        
            begin
               if TestOverlap(oldValues, nextProxy^) then
                  m_pairManager.RemoveBufferedPair(proxyId, nextProxyId);

               Dec(nextProxy^.upperBounds[axis]);
               Dec(bound^.stabbingCount);
            end
            else
            begin
               Dec(nextProxy^.lowerBounds[axis]);
               Inc(bound^.stabbingCount);
            end;

            Inc(proxy^.lowerBounds[axis]);
            b2Swap(bound^, nextBound^);
            Inc(index);
         end;
      end;

      // Should we move the upper bound down?
      if deltaUpper < 0 then
      begin
         index := upperIndex;
         while (index > 0) and (upperValue < bounds^[index - 1].value) do
         begin
            bound := @bounds^[index];
            prevBound := @bounds^[index - 1];
            prevProxyId := prevBound^.proxyId;
            prevProxy := @m_proxyPool[prevProxyId];
            Dec(prevBound^.stabbingCount);

            {$IFDEF OP_OVERLOAD}
            if prevBound^.IsLower() then
            {$ELSE}            
            if IsLower(prevBound^) then
            {$ENDIF}                        
            begin
               if TestOverlap(oldValues, prevProxy^) then
                  m_pairManager.RemoveBufferedPair(proxyId, prevProxyId);

               Inc(prevProxy^.lowerBounds[axis]);
               Dec(bound^.stabbingCount);
            end
            else
            begin
               Inc(prevProxy^.upperBounds[axis]);
               Inc(bound^.stabbingCount);
            end;

            Dec(proxy^.upperBounds[axis]);
            b2Swap(bound^, prevBound^);
            Dec(index);
         end;
      end;
   end;

   {$IFDEF CLASSVAR_AVAIL}
   if s_validate then
   {$ELSE}
   if b2BroadPhase_s_validate then
   {$ENDIF}
     Validate;
end;

procedure Tb2BroadPhase.Commit;
begin
   m_pairManager.Commit;
end;

function Tb2BroadPhase.GetProxy(proxyId: Int32): Pb2Proxy;
begin
   {$IFDEF OP_OVERLOAD}
   if (proxyId = b2_nullProxy) or (not m_proxyPool[proxyId].IsValid) then
   {$ELSE}   
   if (proxyId = b2_nullProxy) or (not IsValid(m_proxyPool[proxyId])) then
   {$ENDIF}      
      Result := nil
   else
      Result := @(m_proxyPool[proxyId]);
end;

function Tb2BroadPhase.Query(const aabb: Tb2AABB; userData: TList;
   maxCount: Int32): Int32;
var
   lowerValues, upperValues: Tb2BoundValuesArray;
   lowerIndex, upperIndex: Int32;
   count: Int32;
   i: Integer;
   proxy: Pb2Proxy;
begin
   ComputeBounds(lowerValues, upperValues, aabb);

   Query(lowerIndex, upperIndex, lowerValues[0], upperValues[0],
      m_bounds[0], 2 * m_proxyCount, 0);
   Query(lowerIndex, upperIndex, lowerValues[1], upperValues[1],
      m_bounds[1], 2 * m_proxyCount, 1);

   //b2Assert(m_queryResultCount < b2_maxProxies);

   count := 0;
   i := 0;
   while (i < m_queryResultCount) and (count < maxCount) do
   begin
      {//b2Assert(m_queryResults[i] < b2_maxProxies);
      proxy := @m_proxyPool[m_queryResults[i]];
      //b2Assert(proxy->IsValid());
      userData[i] := proxy^.userData;    }
      userData.Add(m_proxyPool[m_queryResults[i]].userData);
      Inc(i);
      Inc(count);
   end;

   // Prepare for next query.
   m_queryResultCount := 0;
   IncrementTimeStamp;

   Result := count;
end;

procedure Tb2BroadPhase.Validate;
var
   axis: Integer;
   bounds: Pb2AxialBoundsArray;
   bound: Pb2Bound;
   boundCount: Int32;
   stabbingCount: UInt16;
   i: Integer;
begin
   for axis := 0 to 1 do
   begin
      bounds := @m_bounds[axis];

      boundCount := 2 * m_proxyCount;
      stabbingCount := 0;

      for i := 0 to boundCount - 1 do
      begin
         bound := @bounds^[i];
         //b2Assert(i == 0 || bounds[i-1].value <= bound->value);
         //b2Assert(bound->proxyId != b2_nullProxy);
         //b2Assert(m_proxyPool[bound->proxyId].IsValid());

         {$IFDEF OP_OVERLOAD}
         if bound^.IsLower() then
         {$ELSE}         
         if IsLower(bound^) then
         {$ENDIF}                  
         begin
            //b2Assert(m_proxyPool[bound^.proxyId].lowerBounds[axis] == i);
            Inc(stabbingCount);
         end
         else
         begin
            //b2Assert(m_proxyPool[bound^.proxyId].upperBounds[axis] == i);
            Dec(stabbingCount);
         end;

         //b2Assert(bound->stabbingCount == stabbingCount);
      end;
   end;
end;

procedure Tb2BroadPhase.ValidatePairs;
begin

end;

////////////////////////////////////////////////////
// Pair

{$IFDEF OP_OVERLOAD}

procedure Tb2Pair.SetBuffered;
begin
   status := status or e_pairBuffered;
end;

procedure Tb2Pair.ClearBuffered;
begin
   status := status and (not e_pairBuffered);
end;

function Tb2Pair.IsBuffered: Boolean;
begin
   Result := (status and e_pairBuffered) = e_pairBuffered;
end;

procedure Tb2Pair.SetRemoved;
begin
   status := status or e_pairRemoved;
end;

procedure Tb2Pair.ClearRemoved;
begin
   status := status and (not e_pairRemoved);
end;

function Tb2Pair.IsRemoved: Boolean;
begin
   Result := (status and e_pairRemoved) = e_pairRemoved;
end;

procedure Tb2Pair.SetFinal;
begin
   status := status or e_pairFinal;
end;

function Tb2Pair.IsFinal: Boolean;
begin
   Result := (status and e_pairFinal) = e_pairFinal;
end;
{$ENDIF}

{ Tb2PairManager }

constructor Tb2PairManager.Create;
var
   i: Integer;
begin
   for i := 0 to b2_tableCapacity - 1 do
      m_hashTable[i] := b2_nullPair;

   m_freePair := 0;
   for i := 0 to b2_maxPairs - 1 do
   begin
      m_pairs[i].proxyId1 := b2_nullProxy;
      m_pairs[i].proxyId2 := b2_nullProxy;
      m_pairs[i].userData := nil;
      m_pairs[i].status := 0;
      m_pairs[i].next := i + 1;
   end;
   m_pairs[b2_maxPairs - 1].next := b2_nullPair;
   m_pairCount := 0;
   m_pairBufferCount := 0;
end;

function Tb2PairManager.Find(proxyId1, proxyId2: Int32): Pb2Pair;
var
   hash: Int32;
begin
   if proxyId1 > proxyId2 then
      b2Swap(proxyId1, proxyId2);
   hash := PairManager_Hash(proxyId1, proxyId2) and b2_tableMask;
   Result := Find(proxyId1, proxyId2, hash);
end;

function Tb2PairManager.Find(proxyId1, proxyId2: Int32; hashValue: UInt32): Pb2Pair;
var
   index: Int32;
begin
	 index := m_hashTable[hashValue];
   while (index <> b2_nullPair) and (not PairManager_Equals(m_pairs[index], proxyId1, proxyId2)) do
		  index := m_pairs[index].next;

	 if index = b2_nullPair then
      Result := nil
   else
   begin
      //b2Assert(index < b2_maxPairs);
      Result := Pb2Pair(@m_pairs[index]);
   end;
end;

function Tb2PairManager.AddPair(proxyId1, proxyId2: Int32): Pb2Pair;
var
   hash: Int32;
   pairIndex: Int16;
begin
   if proxyId1 > proxyId2 then
      b2Swap(proxyId1, proxyId2);

   hash := PairManager_Hash(proxyId1, proxyId2) and b2_tableMask;
   Result := Find(proxyId1, proxyId2, hash);
   if Assigned(Result) then
      Exit;

   //b2Assert(m_pairCount < b2_maxPairs && m_freePair != b2_nullPair);

   pairIndex := m_freePair;
   Result := @m_pairs[pairIndex];
   m_freePair := Result^.next;

   Result^.proxyId1 := proxyId1;
   Result^.proxyId2 := proxyId2;
   Result^.status := 0;
   Result^.userData := nil;
   Result^.next := m_hashTable[hash];

   m_hashTable[hash] := pairIndex;
   Inc(m_pairCount);
end;

function Tb2PairManager.RemovePair(proxyId1, proxyId2: Int32): Pointer;
var
   hash: Int32;
   index: UInt16;
   pair: Pb2Pair;
   node: ^UInt16;
   userData: Pointer;
begin
   //b2Assert(m_pairCount > 0);
   if proxyId1 > proxyId2 then
      b2Swap(proxyId1, proxyId2);

   hash := PairManager_Hash(proxyId1, proxyId2) and b2_tableMask;
   node := @m_hashTable[hash];
   while (node^ <> b2_nullPair) do
   begin
      if PairManager_Equals(m_pairs[node^], proxyId1, proxyId2) then
      begin
         index := node^;
         node^ := m_pairs[node^].next;

         pair := @(m_pairs[index]);
         userData := pair^.userData;

         // Scrub
         pair^.next := m_freePair;
         pair^.proxyId1 := b2_nullProxy;
         pair^.proxyId2 := b2_nullProxy;
         pair^.userData := nil;
         pair^.status := 0;

         m_freePair := index;
         Dec(m_pairCount);
         Result := userData;
         Exit;
      end
      else
         node := @m_pairs[node^].next;
   end;
	 //b2Assert(False);
   Result := nil;
end;

procedure Tb2PairManager.ValidateBuffer;
begin
(*
#ifdef _DEBUG
	b2Assert(m_pairBufferCount <= m_pairCount);

	std::sort(m_pairBuffer, m_pairBuffer + m_pairBufferCount);

	for (int32 i = 0; i < m_pairBufferCount; ++i)
	{
		if (i > 0)
		{
			b2Assert(Equals(m_pairBuffer[i], m_pairBuffer[i-1]) == False);
		}

		b2Pair* pair = Find(m_pairBuffer[i].proxyId1, m_pairBuffer[i].proxyId2);
		b2Assert(pair^.IsBuffered());

		b2Assert(pair^.proxyId1 != pair^.proxyId2);
		b2Assert(pair^.proxyId1 < b2_maxProxies);
		b2Assert(pair^.proxyId2 < b2_maxProxies);

		b2Proxy* proxy1 = m_broadPhase^.m_proxyPool + pair^.proxyId1;
		b2Proxy* proxy2 = m_broadPhase^.m_proxyPool + pair^.proxyId2;

		b2Assert(proxy1^.IsValid() == True);
		b2Assert(proxy2^.IsValid() == True);
	}
#endif
*)
end;

procedure Tb2PairManager.ValidateTable;
begin
(*
#ifdef _DEBUG
	for (int32 i = 0; i < b2_tableCapacity; ++i)
	{
		uint16 index = m_hashTable[i];
		while (index != b2_nullPair)
		{
			b2Pair* pair = m_pairs + index;
			b2Assert(pair^.IsBuffered() == False);
			b2Assert(pair^.IsFinal() == True);
			b2Assert(pair^.IsRemoved() == False);

			b2Assert(pair^.proxyId1 != pair^.proxyId2);
			b2Assert(pair^.proxyId1 < b2_maxProxies);
			b2Assert(pair^.proxyId2 < b2_maxProxies);

			b2Proxy* proxy1 = m_broadPhase^.m_proxyPool + pair^.proxyId1;
			b2Proxy* proxy2 = m_broadPhase^.m_proxyPool + pair^.proxyId2;

			b2Assert(proxy1^.IsValid() == True);
			b2Assert(proxy2^.IsValid() == True);

			b2Assert(m_broadPhase^.TestOverlap(proxy1, proxy2) == True);

			index = pair^.next;
		}
	}
#endif
*)
end;

procedure Tb2PairManager.Initialize(broadPhase: Tb2BroadPhase;
   callback: Tb2PairCallback);
begin
	 m_broadPhase := broadPhase;
	 m_callback := callback;
end;

{
As proxies are created and moved, many pairs are created and destroyed. Even worse, the same
pair may be added and removed multiple times in a single time step of the physics engine. To reduce
traffic in the pair manager, we try to avoid destroying pairs in the pair manager until the
end of the physics step. This is done by buffering all the RemovePair requests. AddPair
requests are processed immediately because we need the hash table entry for quick lookup.

All user user callbacks are delayed until the buffered pairs are confirmed in Commit.
This is very important because the user callbacks may be very expensive and client logic
may be harmed if pairs are added and removed within the same time step.

Buffer a pair for addition.
We may add a pair that is not in the pair manager or pair buffer.
We may add a pair that is already in the pair manager and pair buffer.
If the added pair is not a new pair, then it must be in the pair buffer (because RemovePair was called).
}
procedure Tb2PairManager.AddBufferedPair(proxyId1, proxyId2: Int32);
var
   pair: Pb2Pair;
begin
   //b2Assert(id1 != b2_nullProxy && id2 != b2_nullProxy);
   //b2Assert(m_pairBufferCount < b2_maxPairs);

   pair := AddPair(proxyId1, proxyId2);

   // If this pair is not in the pair buffer ...
   {$IFDEF OP_OVERLOAD}
   if not pair^.IsBuffered() then
   {$ELSE}
   if not IsBuffered(pair^) then
   {$ENDIF}
   begin
      // This must be a newly added pair.
      //b2Assert(pair^.IsFinal() == False);

      // Add it to the pair buffer.
      {$IFDEF OP_OVERLOAD}
      pair^.SetBuffered;
      {$ELSE}
      SetBuffered(pair^);
      {$ENDIF}
      m_pairBuffer[m_pairBufferCount].proxyId1 := pair^.proxyId1;
      m_pairBuffer[m_pairBufferCount].proxyId2 := pair^.proxyId2;
      Inc(m_pairBufferCount);

      //b2Assert(m_pairBufferCount <= m_pairCount);
   end;

   // Confirm this pair for the subsequent call to Commit.
   {$IFDEF OP_OVERLOAD}
   pair^.ClearRemoved;
   {$ELSE}
   ClearRemoved(pair^);
   {$ENDIF}

   {$IFDEF CLASSVAR_AVAIL}
   if Tb2BroadPhase.s_validate then
   {$ELSE}
   if b2BroadPhase_s_validate then
   {$ENDIF}
      ValidateBuffer;
end;

procedure Tb2PairManager.RemoveBufferedPair(proxyId1, proxyId2: Int32);
var
   pair: Pb2Pair;
begin
   //b2Assert(id1 != b2_nullProxy && id2 != b2_nullProxy);
   //b2Assert(m_pairBufferCount < b2_maxPairs);

   pair := Find(proxyId1, proxyId2);

   if not Assigned(pair) then
      // The pair never existed. This is legal (due to collision filtering).
      Exit;

   // If this pair is not in the pair buffer ...
   {$IFDEF OP_OVERLOAD}
   if (not pair^.IsBuffered()) then
   {$ELSE}
   if (not IsBuffered(pair^)) then
   {$ENDIF}
   begin
     // This must be an old pair.
     //b2Assert(pair^.IsFinal() == True);

     {$IFDEF OP_OVERLOAD}
     pair^.SetBuffered;
     {$ELSE}
     SetBuffered(pair^);
     {$ENDIF}
     m_pairBuffer[m_pairBufferCount].proxyId1 := pair^.proxyId1;
     m_pairBuffer[m_pairBufferCount].proxyId2 := pair^.proxyId2;
     Inc(m_pairBufferCount)

     //Assert(m_pairBufferCount <= m_pairCount);
   end;
   {$IFDEF OP_OVERLOAD}
   pair^.SetRemoved;
   {$ELSE}
   SetRemoved(pair^);
   {$ENDIF}

   {$IFDEF CLASSVAR_AVAIL}
   if Tb2BroadPhase.s_validate then
   {$ELSE}
   if b2BroadPhase_s_validate then
   {$ENDIF}
      ValidateBuffer;
end;

procedure Tb2PairManager.Commit;
var
   i: Integer;
   removeCount: Int32;
   proxy1, proxy2: Pb2Proxy;
   pair: Pb2Pair;
begin
   removeCount := 0;

   for i := 0 to m_pairBufferCount - 1 do
   begin
      pair := Find(m_pairBuffer[i].proxyId1, m_pairBuffer[i].proxyId2);
      //b2Assert(pair^.IsBuffered());
      {$IFDEF OP_OVERLOAD}
      pair^.ClearBuffered;
      {$ELSE}
      ClearBuffered(pair^);
      {$ENDIF}

      //b2Assert(pair^.proxyId1 < b2_maxProxies && pair^.proxyId2 < b2_maxProxies);

      proxy1 := @m_broadPhase.m_proxyPool[pair^.proxyId1];
      proxy2 := @m_broadPhase.m_proxyPool[pair^.proxyId2];

      //b2Assert(proxy1^.IsValid());
      //b2Assert(proxy2^.IsValid());

      {$IFDEF OP_OVERLOAD}
      if pair^.IsRemoved() then
      {$ELSE}
      if IsRemoved(pair^) then
      {$ENDIF}
      begin
         // It is possible a pair was added then removed before a commit. Therefore,
         // we should be careful not to tell the user the pair was removed when the
         // the user didn't receive a matching add.
         {$IFDEF OP_OVERLOAD}
         if pair^.IsFinal() then
         {$ELSE}
         if IsFinal(pair^) then
         {$ENDIF}
            m_callback.PairRemoved(proxy1^.userData, proxy2^.userData, pair^.userData);

         // Store the ids so we can actually remove the pair below.
         m_pairBuffer[removeCount].proxyId1 := pair^.proxyId1;
         m_pairBuffer[removeCount].proxyId2 := pair^.proxyId2;
         Inc(removeCount);
      end
      else
      begin
         //b2Assert(m_broadPhase^.TestOverlap(proxy1, proxy2) == True);
         {$IFDEF OP_OVERLOAD}
         if (not pair^.IsFinal()) then
         {$ELSE}
         if (not IsFinal(pair^)) then
         {$ENDIF}
         begin
            pair^.userData := m_callback.PairAdded(proxy1^.userData, proxy2^.userData);
            {$IFDEF OP_OVERLOAD}
            pair^.SetFinal;
            {$ELSE}
            SetFinal(pair^);
            {$ENDIF}
         end;
      end;
   end;

   for i := 0 to removeCount - 1 do
      RemovePair(m_pairBuffer[i].proxyId1, m_pairBuffer[i].proxyId2);

   m_pairBufferCount := 0;
   {$IFDEF CLASSVAR_AVAIL}
   if Tb2BroadPhase.s_validate then
   {$ELSE}
   if b2BroadPhase_s_validate then
   {$ENDIF}
      ValidateBuffer;
end;

{ Tb2ShapeDef }

constructor Tb2ShapeDef.Create;
begin
   ShapeType := e_unknownShape;
   userData := nil;
   friction := 0.2;
   restitution := 0.0;
   density := 0.0;
   filter.categoryBits := $0001;
   filter.maskBits := $FFFF;
   filter.groupIndex := 0;
   isSensor := False;
end;

{ Tb2Shape }

constructor Tb2Shape.Create(def: Tb2ShapeDef);
begin
   m_userData := def.userData;
   m_friction := def.friction;
   m_restitution := def.restitution;
   m_density := def.density;
   m_body := nil;
   m_sweepRadius := 0.0;

   m_next := nil;
   m_proxyId := b2_nullProxy;
   m_filter := def.filter;
   m_isSensor := def.isSensor;
end;

destructor Tb2Shape.Destroy;
begin
   if Assigned(m_body) then
      m_body.DestroyShape(Self, False);
end;

destructor Tb2Shape.Destroy2;
begin      
end;

procedure Tb2Shape.CreateProxy(broadPhase: Tb2BroadPhase; const xf: Tb2XForm);
var
   aabb: Tb2AABB;
   inRange: Boolean;
begin
   //b2Assert(m_proxyId == b2_nullProxy);
   ComputeAABB(aabb, xf);

   inRange := broadPhase.InRange(aabb);

   // You are creating a shape outside the world box.
   //b2Assert(inRange);

   if inRange then
      m_proxyId := broadPhase.CreateProxy(aabb, Self)
   else
      m_proxyId := b2_nullProxy;
end;

procedure Tb2Shape.DestroyProxy(broadPhase: Tb2BroadPhase);
begin
   if m_proxyId <> b2_nullProxy then
   begin
      broadPhase.DestroyProxy(m_proxyId);
      m_proxyId := b2_nullProxy;
   end;
end;

function Tb2Shape.Synchronize(broadPhase: Tb2BroadPhase; const xf1, xf2: Tb2XForm): Boolean;
var
   aabb: Tb2AABB;
begin
   if m_proxyId = b2_nullProxy then
   begin
      Result := False;
      Exit;
   end;

   // Compute an AABB that covers the swept shape (may miss some rotation effect).
   ComputeSweptAABB(aabb, xf1, xf2);

   if broadPhase.InRange(aabb) then
   begin
      broadPhase.MoveProxy(m_proxyId, aabb);
      Result := True;
   end
   else
      Result := False;
end;

procedure Tb2Shape.RefilterProxy(broadPhase: Tb2BroadPhase; const xf: Tb2XForm);
var
   aabb: Tb2AABB;
begin
   if m_proxyId = b2_nullProxy then
      Exit;

   broadPhase.DestroyProxy(m_proxyId);
   ComputeAABB(aabb, xf);

   if broadPhase.InRange(aabb) then
      m_proxyId := broadPhase.CreateProxy(aabb, Self)
   else
      m_proxyId := b2_nullProxy;
end;

//////////////////////////////////////////////////////////////
// Joints    

{ Tb2Jacobian }

{$IFDEF OP_OVERLOAD}
procedure Tb2Jacobian.SetZero;
begin
   linear1.SetZero;
   linear2.SetZero;
   angular1 := 0.0;   	 
   angular2 := 0.0;
end;
      
procedure Tb2Jacobian.SetValue(const x1, x2: TVector2; a1, a2: Float);
begin
	 linear1 := x1; 
   linear2 := x2;  
   angular1 := a1;    
   angular2 := a2;
end;
      
function Tb2Jacobian.Compute(const x1, x2: TVector2; a1, a2: Float): Float;
begin
   Result := b2Dot(linear1, x1) + angular1 * a1 + b2Dot(linear2, x2) + angular2 * a2;
end;  
{$ELSE}
// unfinished
{$ENDIF}        

{ Tb2JointDef }

constructor Tb2JointDef.Create;
begin
		JointType := e_unknownJoint;
		userData := nil;
		body1 := nil;
		body2 := nil;
		collideConnected := False;
end;

{ Tb2Joint }

constructor Tb2Joint.Create(def: Tb2JointDef);
begin
   m_type := def.JointType;
   m_prev := nil;
   m_next := nil;
   m_body1 := def.body1;
   m_body2 := def.body2;
   m_collideConnected := def.collideConnected;
   m_islandFlag := False;
   m_userData := def.userData;
end;

procedure Tb2Joint.InitPositionConstraints;
begin

end;

//////////////////////////////////////////////////////////////
// Body

{ Tb2BodyDef }

constructor Tb2BodyDef.Create;
begin
   SetZero(massData.center);
   massData.mass := 0.0;
   massData.I := 0.0;
   userData := nil;
   SetZero(position);
   angle := 0.0;
   linearDamping := 0.0;
   angularDamping := 0.0;
   allowSleep := True;
   isSleeping := False;
   fixedRotation := False;
   isBullet := False;
end;

{ Tb2Body }

constructor Tb2Body.Create(bd: Tb2BodyDef; world: Tb2World);
begin
   //b2Assert(world->m_lock == False);
   m_flags := 0;

   if bd.isBullet then
      m_flags := m_flags or e_bulletFlag;
   if bd.fixedRotation then
      m_flags := m_flags or e_fixedRotationFlag;
   if bd.allowSleep then
      m_flags := m_flags or e_allowSleepFlag;
   if bd.isSleeping then
      m_flags := m_flags or e_sleepFlag;

   m_world := world;

   m_xf.position := bd.position;
   {$IFDEF OP_OVERLOAD}
   m_xf.R.SetValue(bd.angle);
   {$ELSE}
   SetValue(m_xf.R, bd.angle);
   {$ENDIF}

   m_sweep.localCenter := bd.massData.center;
   m_sweep.t0 := 1.0;
   m_sweep.a := bd.angle;
   m_sweep.a0 := bd.angle;
   m_sweep.c0 := b2Mul(m_xf, m_sweep.localCenter);
   m_sweep.c := m_sweep.c0;

   m_jointList := nil;
   m_contactList := nil;
   m_prev := nil;
   m_next := nil;

   m_linearDamping := bd.linearDamping;
   m_angularDamping := bd.angularDamping;

   SetZero(m_force);
   m_torque := 0.0;

   SetZero(m_linearVelocity);
   m_angularVelocity := 0.0;

   m_sleepTime := 0.0;

   m_invMass := 0.0;
   m_I := 0.0;
   m_invI := 0.0;

   m_mass := bd.massData.mass;

   if m_mass > 0.0 then
      m_invMass := 1.0 / m_mass;

   if (m_flags and e_fixedRotationFlag) = 0 then
      m_I := bd.massData.I;

   if m_I > 0.0 then
      m_invI := 1.0 / m_I;

   if (m_invMass = 0.0) and (m_invI = 0.0) then
      m_type := e_staticType
   else
      m_type := e_dynamicType;

   m_userData := bd.userData;

   m_shapeList := nil;
   m_shapeCount := 0;
end;

destructor Tb2Body.Destroy;
begin
   if Assigned(m_world) then 
      m_world.DestroyBody(Self, False);
end;

destructor Tb2Body.Destroy2;
begin 
end;

function Tb2Body.SynchronizeShapes: Boolean;
var
   s: Tb2Shape;
   xf1: Tb2XForm;
   inRange: Boolean;
begin
   {$IFDEF OP_OVERLOAD}
   xf1.R.SetValue(m_sweep.a0);
   xf1.position := m_sweep.c0 - b2Mul(xf1.R, m_sweep.localCenter);
   {$ELSE}
   SetValue(xf1.R, m_sweep.a0);
   xf1.position := Subtract(m_sweep.c0, b2Mul(xf1.R, m_sweep.localCenter));
   {$ENDIF}

   inRange := True;
   s := m_shapeList;
   while Assigned(s) do
   begin
      inRange := s.Synchronize(m_world.m_broadPhase, xf1, m_xf);
      if not inRange then
         Break;
      s := s.m_next;
   end;

   if not inRange then
   begin
      m_flags := m_flags or e_frozenFlag;
      SetZero(m_linearVelocity);
      m_angularVelocity := 0.0;

      s := m_shapeList;
      while Assigned(s) do
      begin
         s.DestroyProxy(m_world.m_broadPhase);
         s := s.m_next;
      end;

      // Failure
      Result := False;
      Exit;
   end;

   // Success
   Result := True;
end;

procedure Tb2Body.SynchronizeTransform;
begin
   {$IFDEF OP_OVERLOAD}
   m_xf.R.SetValue(m_sweep.a);
   m_xf.position := m_sweep.c - b2Mul(m_xf.R, m_sweep.localCenter);
   {$ELSE}
   SetValue(m_xf.R, m_sweep.a);
   m_xf.position := Subtract(m_sweep.c, b2Mul(m_xf.R, m_sweep.localCenter));
   {$ENDIF}
end;

function Tb2Body.IsConnected(other: Tb2Body): Boolean;
var
   jn: Pb2JointEdge;
begin
   jn := m_jointList;
   while Assigned(jn) do
   begin
      if jn^.other = other then
      begin
         Result := jn^.joint.m_collideConnected = False;
         Exit;
      end;
      jn := jn^.next;
   end;
   Result := False;
end;

procedure Tb2Body.Advance(t: Float);
begin
   // Advance to the new safe time.
   {$IFDEF OP_OVERLOAD}
   m_sweep.Advance(t);
   {$ELSE}
   UPhysics2DTypes.Advance(m_sweep, t);
   {$ENDIF}
   m_sweep.c := m_sweep.c0;
   m_sweep.a := m_sweep.a0;
   SynchronizeTransform;
end;

function Tb2Body.CreateShape(shapeDef: Tb2ShapeDef; AutoFreeShapeDef: Boolean = True): Tb2Shape;
begin
   //b2Assert(m_world->m_lock == False);
   if m_world.m_lock then
   begin
      Result := nil;
      Exit;
   end;

   case shapeDef.ShapeType of
      e_circleShape: Result := Tb2CircleShape.Create(shapeDef);
      e_polygonShape: Result := Tb2PolygonShape.Create(shapeDef);
   end;

   Result.m_next := m_shapeList;
   m_shapeList := Result;
   Inc(m_shapeCount);

   Result.m_body := Self;

   // Add the shape to the world's broad-phase.
   Result.CreateProxy(m_world.m_broadPhase, m_xf);

   // Compute the sweep radius for CCD.
   Result.UpdateSweepRadius(m_sweep.localCenter);

   if AutoFreeShapeDef then
      shapeDef.Free;
end;

procedure Tb2Body.DestroyShape(s: Tb2Shape; DoFree: Boolean = True);
var
   node: Pb2Shape;
   found: Boolean;
begin
   //b2Assert(m_world->m_lock == False);
   if m_world.m_lock then
      Exit;

   //b2Assert(s->GetBody() == this);
   s.DestroyProxy(m_world.m_broadPhase);

   //b2Assert(m_shapeCount > 0);
   node := @m_shapeList;
   found := False;
   while (node^ <> nil) do
   begin
      if node^ = s then
      begin
         node^ := s.m_next;
         found := True;
         Break;
      end;
      node := @node^.m_next;
   end;

   // You tried to remove a shape that is not attached to this body.
   //b2Assert(found);
   Dec(m_shapeCount);
   if DoFree then      
      s.Destroy2; // Call a destructor without side effects.
end;

procedure Tb2Body.SetMass(const massData: Tb2MassData);
var
   s: Tb2Shape;
   oldType: Tb2BodyType;
begin
   //b2Assert(m_world->m_lock == False);
   if m_world.m_lock then
      Exit;

   m_invMass := 0.0;
   m_I := 0.0;
   m_invI := 0.0;

   m_mass := massData.mass;

   if m_mass > 0.0 then
      m_invMass := 1.0 / m_mass;

   if (m_flags and e_fixedRotationFlag) = 0 then
      m_I := massData.I;

   if m_I > 0.0 then
      m_invI := 1.0 / m_I;

   // Move center of mass.
   m_sweep.localCenter := massData.center;
   m_sweep.c0 := b2Mul(m_xf, m_sweep.localCenter);
   m_sweep.c := m_sweep.c0;

   // Update the sweep radii of all child shapes.
   s := m_shapeList;
   while Assigned(s) do
   begin
      s.UpdateSweepRadius(m_sweep.localCenter);
      s := s.m_next;
   end;

   oldType := m_type;
   if (m_invMass = 0.0) and (m_invI = 0.0) then
      m_type := e_staticType
   else
      m_type := e_dynamicType;

   // If the body type changed, we need to refilter the broad-phase proxies.
   if oldType <> m_type then
   begin
      s := m_shapeList;
      while Assigned(s) do
      begin
         s.RefilterProxy(m_world.m_broadPhase, m_xf);
         s := s.m_next;
      end;
   end;
end;

procedure Tb2Body.SetMassFromShapes;
var
   s: Tb2Shape;
   center: TVector2;
   massData: Tb2MassData;
   oldType: Tb2BodyType;
begin
   //b2Assert(m_world->m_lock == False);
   if m_world.m_lock then
      Exit;

   // Compute mass data from shapes. Each shape has its own density.
   m_mass := 0.0;
   m_invMass := 0.0;
   m_I := 0.0;
   m_invI := 0.0;

   center := b2Vec2_zero;
   s := m_shapeList;
   while Assigned(s) do
   begin
      s.ComputeMass(massData);
      m_mass := m_mass + massData.mass;
      {$IFDEF OP_OVERLOAD}
      center.AddBy(massData.mass * massData.center);
      {$ELSE}
      AddBy(center, Multiply(massData.center, massData.mass));
      {$ENDIF}
      m_I := m_I + massData.I;
      s := s.m_next;
   end;

   // Compute center of mass, and shift the origin to the COM.
   if m_mass > 0.0 then
   begin
      m_invMass := 1.0 / m_mass;
      {$IFDEF OP_OVERLOAD}
      center.MultiplyBy(m_invMass);
      {$ELSE}
      MultiplyBy(center, m_invMass);
      {$ENDIF}
   end;

   if (m_I > 0.0) and ((m_flags and e_fixedRotationFlag) = 0) then
   begin
      // Center the inertia about the center of mass.
      m_I := m_I - m_mass * b2Dot(center, center);
      //b2Assert(m_I > 0.0f);
      m_invI := 1.0 / m_I;
   end
   else
   begin
      m_I := 0.0;
      m_invI := 0.0;
   end;

   // Move center of mass.
   m_sweep.localCenter := center;
   m_sweep.c := b2Mul(m_xf, m_sweep.localCenter);
   m_sweep.c0 := m_sweep.c;

   // Update the sweep radii of all child shapes.
   s := m_shapeList;
   while Assigned(s) do
   begin
      s.UpdateSweepRadius(m_sweep.localCenter);
      s := s.m_next;
   end;

   oldType := m_type;
   if (m_invMass = 0.0) and (m_invI = 0.0) then
      m_type := e_staticType
   else
      m_type := e_dynamicType;

   // If the body type changed, we need to refilter the broad-phase proxies.
   if oldType <> m_type then
   begin
      s := m_shapeList;
      while Assigned(s) do
      begin
         s.RefilterProxy(m_world.m_broadPhase, m_xf);
         s := s.m_next;
      end;
   end;
end;

function Tb2Body.SetXForm(const position: TVector2; angle: Float): Boolean;
var
   s: Tb2Shape;
   freeze: Boolean;
begin
   //b2Assert(m_world->m_lock == False);
   if m_world.m_lock then
   begin
      Result := True;
      Exit;
   end;

   if IsFrozen() then
   begin
      Result := False;
      Exit;
   end;

   {$IFDEF OP_OVERLOAD}
   m_xf.R.SetValue(angle);
   {$ELSE}
   SetValue(m_xf.R, angle);
   {$ENDIF}
   m_xf.position := position;

   m_sweep.c0 := b2Mul(m_xf, m_sweep.localCenter);
   m_sweep.c := m_sweep.c0;
   m_sweep.a0 := angle;
   m_sweep.a := angle;

   freeze := False;
   s := m_shapeList;
   while Assigned(s) do
   begin
      if not s.Synchronize(m_world.m_broadPhase, m_xf, m_xf) then
      begin
         freeze := True;
         Break;
      end;
      s := s.m_next;
   end;

   if freeze then
   begin
      m_flags := m_flags or e_frozenFlag;
      SetZero(m_linearVelocity);
      m_angularVelocity := 0.0;

      s := m_shapeList;
      while Assigned(s) do
      begin
         s.DestroyProxy(m_world.m_broadPhase);
         s := s.m_next;
      end;

      // Failure
      Result := False;
      Exit;
   end;

   // Success
   m_world.m_broadPhase.Commit;
   Result := True;
end;

procedure Tb2Body.SetLinearVelocity(const v: TVector2);
begin
   m_linearVelocity := v;
end;

procedure Tb2Body.SetAngularVelocity(omega: Float);
begin
   m_angularVelocity := omega;
end;

procedure Tb2Body.ApplyForce(const force, point: TVector2);
begin
   if IsSleeping() then
      WakeUp;
   {$IFDEF OP_OVERLOAD}
	 m_force.AddBy(force);
   m_torque := m_torque + b2Cross(point - m_sweep.c, force);
   {$ELSE}
   AddBy(m_force, force);
   m_torque := m_torque + b2Cross(Subtract(point, m_sweep.c), force);
   {$ENDIF}
end;

procedure Tb2Body.ApplyTorque(torque: Float);
begin
   if IsSleeping() then
      WakeUp;
	 m_torque := m_torque + torque;
end;

procedure Tb2Body.ApplyImpulse(const impulse, point: TVector2);
begin
   if IsSleeping() then
      WakeUp;
   {$IFDEF OP_OVERLOAD}
   m_linearVelocity.AddBy(m_invMass * impulse);
   m_angularVelocity := m_angularVelocity + m_invI * b2Cross(point -
      m_sweep.c, impulse);
   {$ELSE}
   AddBy(m_linearVelocity, Multiply(impulse, m_invMass));
   m_angularVelocity := m_angularVelocity + m_invI * b2Cross(Subtract(point,
      m_sweep.c), impulse);
   {$ENDIF}
end;

function Tb2Body.GetWorldPoint(const localPoint: TVector2): TVector2;
begin
   Result := b2Mul(m_xf, localPoint);
end;

function Tb2Body.GetWorldVector(const localVector: TVector2): TVector2;
begin
   Result := b2Mul(m_xf.R, localVector);
end;

function Tb2Body.GetLocalPoint(const worldPoint: TVector2): TVector2;
begin
   Result := b2MulT(m_xf, worldPoint);
end;

function Tb2Body.GetLocalVector(const worldVector: TVector2): TVector2;
begin
   Result := b2MulT(m_xf.R, worldVector);
end;

function Tb2Body.GetLinearVelocityFromWorldPoint(
   const worldPoint: TVector2): TVector2;
begin
   {$IFDEF OP_OVERLOAD}
   Result := m_linearVelocity + b2Cross(m_angularVelocity, worldPoint - m_sweep.c);
   {$ELSE}
   Result := Add(m_linearVelocity, b2Cross(m_angularVelocity,
      Subtract(worldPoint, m_sweep.c)));
   {$ENDIF}
end;

function Tb2Body.GetLinearVelocityFromLocalPoint(
   const localPoint: TVector2): TVector2;
begin
   Result := GetLinearVelocityFromWorldPoint(GetWorldPoint(localPoint));
end;

function Tb2Body.IsBullet: Boolean;
begin
   Result := (m_flags and e_bulletFlag) = e_bulletFlag;
end;

procedure Tb2Body.SetBullet(flag: Boolean);
begin
   if flag then
      m_flags := m_flags or e_bulletFlag
   else
      m_flags := m_flags and (not e_bulletFlag);
end;

function Tb2Body.IsStatic: Boolean;
begin
   Result := m_type = e_staticType;
end;

function Tb2Body.IsDynamic: Boolean;
begin
   Result := m_type = e_dynamicType;
end;

function Tb2Body.IsFrozen: Boolean;
begin
   Result := (m_flags and e_frozenFlag) = e_frozenFlag;
end;

function Tb2Body.IsSleeping: Boolean;
begin
   Result := (m_flags and e_sleepFlag) = e_sleepFlag;
end;

procedure Tb2Body.AllowSleeping(flag: Boolean);
begin
   if flag then
      m_flags := m_flags or e_allowSleepFlag
   else
   begin
      m_flags := m_flags and (not e_allowSleepFlag);
      WakeUp;
   end;
end;

procedure Tb2Body.WakeUp;
begin
	 m_flags := m_flags and (not e_sleepFlag);
	 m_sleepTime := 0.0;
end;

procedure Tb2Body.PutToSleep;
begin
   m_flags := m_flags or e_sleepFlag;
   m_sleepTime := 0.0;
   SetZero(m_linearVelocity);
   m_angularVelocity := 0.0;
   SetZero(m_force);
   m_torque := 0.0;
end;

///////////////////////////////////////////////
// Specific implementations

procedure b2CollideCircles(var manifold: Tb2Manifold;
   circle1, circle2: Tb2CircleShape; const xf1, xf2: Tb2XForm);
var
   p1, p2, d: TVector2;
   distSqr, r1, r2, radiusSum, separation, dist, a: Float;
begin
   manifold.pointCount := 0;

   p1 := b2Mul(xf1, circle1.m_localPosition);
   p2 := b2Mul(xf2, circle2.m_localPosition);

   {$IFDEF OP_OVERLOAD}
   d := p2 - p1;
   {$ELSE}
   d := Subtract(p2, p1);
   {$ENDIF}
   distSqr := b2Dot(d, d);
   r1 := circle1.GetRadius;
   r2 := circle2.GetRadius;
   radiusSum := r1 + r2;
   if distSqr > radiusSum * radiusSum then
      Exit;

   if (distSqr < FLT_EPSILON) then
   begin
      separation := -radiusSum;
      {$IFDEF OP_OVERLOAD}
      manifold.normal.SetValue(0.0, 1.0);
      {$ELSE}
      SetValue(manifold.normal, 0.0, 1.0);
      {$ENDIF}
   end
   else
   begin
      dist := Sqrt(distSqr);
      separation := dist - radiusSum;
      a := 1.0 / dist;
      manifold.normal.x := a * d.x;
      manifold.normal.y := a * d.y;
   end;

   manifold.pointCount := 1;
   manifold.points[0].id.key := 0;
   manifold.points[0].separation := separation;

   {$IFDEF OP_OVERLOAD}
   p1.AddBy(r1 * manifold.normal);
   p2.SubtractBy(r2 * manifold.normal);
   d := 0.5 * (p1 + p2);
   {$ELSE}
   AddBy(p1, Multiply(manifold.normal, r1));
   SubtractBy(p2, Multiply(manifold.normal, r2));
   d := Add(p1, p2);
   MultiplyBy(d, 0.5);
   {$ENDIF}

   manifold.points[0].localPoint1 := b2MulT(xf1, d);
   manifold.points[0].localPoint2 := b2MulT(xf2, d);
end;

procedure b2CollidePolygonAndCircle(var manifold: Tb2Manifold;
   polygon: Tb2PolygonShape; circle: Tb2CircleShape; const xf1, xf2: Tb2XForm);
var
   i: Integer;
   c, cLocal, position, e, p, d: TVector2;
   normalIndex: Int32;
   _separation, radius, length, dist: Float;
   s, u: Float;
   vertIndex1, vertIndex2: Int32;
begin
   manifold.pointCount := 0;

   // Compute circle position in the frame of the polygon.
   c := b2Mul(xf2, circle.m_localPosition);
   cLocal := b2MulT(xf1, c);

   // Find the min separating edge.
   normalIndex := 0;
   _separation := -FLT_MAX;
   radius := circle.GetRadius;

   for i := 0 to polygon.GetVertexCount - 1 do
   begin
      {$IFDEF OP_OVERLOAD}
      s := b2Dot(polygon.m_normals[i], cLocal - polygon.m_vertices[i]);
      {$ELSE}
      s := b2Dot(polygon.m_normals[i], Subtract(cLocal, polygon.m_vertices[i]));
      {$ENDIF}
      if s > radius then
         Exit;

      if s > _separation then
      begin
         _separation := s;
         normalIndex := i;
      end;
   end;

   // If the center is inside the polygon ...
   if _separation < FLT_EPSILON then
      with manifold do
      begin
         pointCount := 1;
         normal := b2Mul(xf1.R, polygon.m_normals[normalIndex]);

         with points[0], id do
         begin
            incidentEdge := normalIndex;
            incidentVertex := b2_nullFeature;
            referenceEdge := 0;
            flip := 0;

            {$IFDEF OP_OVERLOAD}
            position := c - radius * manifold.normal;
            {$ELSE}
            position := Subtract(c, Multiply(manifold.normal, radius));
            {$ENDIF}
            localPoint1 := b2MulT(xf1, position);
            localPoint2 := b2MulT(xf2, position);
            separation := _separation - radius;
         end;
         Exit;
      end;

   // Project the circle center onto the edge segment.
   vertIndex1 := normalIndex;
   if vertIndex1 + 1 < polygon.GetVertexCount then
      vertIndex2 := vertIndex1 + 1
   else
      vertIndex2 := 0;

   {$IFDEF OP_OVERLOAD}
   e := polygon.m_vertices[vertIndex2] - polygon.m_vertices[vertIndex1];
   length := e.Normalize;
   //b2Assert(length > B2_FLT_EPSILON);

   // Project the center onto the edge.
   u := b2Dot(cLocal - polygon.m_vertices[vertIndex1], e);
   {$ELSE}
   e := Subtract(polygon.m_vertices[vertIndex2], polygon.m_vertices[vertIndex1]);
   length := Normalize(e);
   //b2Assert(length > B2_FLT_EPSILON);

   // Project the center onto the edge.
   u := b2Dot(Subtract(cLocal, polygon.m_vertices[vertIndex1]), e);
   {$ENDIF}

   if u <= 0.0 then
   begin
      p := polygon.m_vertices[vertIndex1];
      manifold.points[0].id.incidentEdge := b2_nullFeature;
      manifold.points[0].id.incidentVertex := vertIndex1;
   end
   else if u >= length then
   begin
      p := polygon.m_vertices[vertIndex2];
      manifold.points[0].id.incidentEdge := b2_nullFeature;
      manifold.points[0].id.incidentVertex := vertIndex2;
   end
   else
   begin
      {$IFDEF OP_OVERLOAD}
      p := polygon.m_vertices[vertIndex1] + u * e;
      {$ELSE}
      p := Add(polygon.m_vertices[vertIndex1], Multiply(e, u));
      {$ENDIF}
      manifold.points[0].id.incidentEdge := normalIndex;
      manifold.points[0].id.incidentVertex := 0;
   end;

   {$IFDEF OP_OVERLOAD}
   d := cLocal - p;
   dist := d.Normalize;
   {$ELSE}
   d := Subtract(cLocal, p);
   dist := Normalize(d);
   {$ENDIF}
   if dist > radius then
      Exit;

   manifold.pointCount := 1;
   manifold.normal := b2Mul(xf1.R, d);
   {$IFDEF OP_OVERLOAD}
   position := c - radius * manifold.normal;
   {$ELSE}
   position := Subtract(c, Multiply(manifold.normal, radius));
   {$ENDIF}
   with manifold.points[0] do
   begin
      localPoint1 := b2MulT(xf1, position);
      localPoint2 := b2MulT(xf2, position);
      separation := dist - radius;
      id.referenceEdge := 0;
      id.flip := 0;
   end;
end;

type
   PClipVertex = ^TClipVertex;
   TClipVertex = record
      v: TVector2;
      id: Tb2ContactID;
   end;

   PClipVertices = ^TClipVertices;
   TClipVertices = array[0..1] of TClipVertex;

function ClipSegmentToLine(vOut, vIn: PClipVertices; const normal: TVector2;
   offset: Float): Int32;
var
   distance0, distance1, interp: Float;
begin
   Result := 0; // Start with no output points

   // Calculate the distance of end points to the line
   distance0 := b2Dot(normal, vIn^[0].v) - offset;
   distance1 := b2Dot(normal, vIn^[1].v) - offset;

   // If the points are behind the plane
   if distance0 <= 0.0 then
   begin
      vOut^[Result] := vIn^[0];
      Inc(Result);
   end;
   if distance1 <= 0.0 then
   begin
      vOut^[Result] := vIn^[1];
      Inc(Result);
   end;

   // If the points are on different sides of the plane
   if distance0 * distance1 < 0.0 then
   begin
      // Find intersection point of edge and plane
      interp := distance0 / (distance0 - distance1);
      {$IFDEF OP_OVERLOAD}
      vOut^[Result].v := vIn^[0].v + interp * (vIn^[1].v - vIn^[0].v);
      {$ELSE}
      vOut^[Result].v := Add(vIn^[0].v, Multiply(Subtract(vIn^[1].v, vIn^[0].v), interp));
      {$ENDIF}
      if distance0 > 0.0 then
         vOut^[Result].id := vIn^[0].id
      else
         vOut^[Result].id := vIn^[1].id;
      Inc(Result);
   end;
end;

// Find the separation between poly1 and poly2 for a give edge normal on poly1.
function EdgeSeparation(poly1, poly2: Tb2PolygonShape; const xf1,
   xf2: Tb2XForm; edge1: Int32): Float;
var
   i: Integer;
   index: Int32;
   minDot, dot: Float;
   normal1World, normal1: TVector2;
   v1, v2: TVector2;
begin
   //b2Assert(0 <= edge1 && edge1 < count1);

   // Convert normal from poly1's frame into poly2's frame.
   normal1World := b2Mul(xf1.R, poly1.m_normals[edge1]);
   normal1 := b2MulT(xf2.R, normal1World);

   // Find support vertex on poly2 for -normal.
   index := 0;
   minDot := FLT_MAX;

   for i := 0 to poly2.GetVertexCount - 1 do
   begin
      dot := b2Dot(poly2.m_vertices[i], normal1);
      if dot < minDot then
      begin
         minDot := dot;
         index := i;
      end;
   end;

   v1 := b2Mul(xf1, poly1.m_vertices[edge1]);
   v2 := b2Mul(xf2, poly2.m_vertices[index]);
   {$IFDEF OP_OVERLOAD}
   Result := b2Dot(v2 - v1, normal1World);
   {$ELSE}
   Result := b2Dot(Subtract(v2, v1), normal1World);
   {$ENDIF}
end;

// Find the max separation between poly1 and poly2 using edge normals from poly1.
function FindMaxSeparation(var edgeIndex: Int32;
   poly1, poly2: Tb2PolygonShape; const xf1, xf2: Tb2XForm): Float;
var
   i: Integer;
   edge, prevEdge, nextEdge, bestEdge, increment: Int32;
   d, dLocal1: TVector2;
   maxDot, dot, s, sPrev, sNext, bestSeparation: Float;
begin
   // Vector pointing from the centroid of poly1 to the centroid of poly2.
   {$IFDEF OP_OVERLOAD}
   d := b2Mul(xf2, poly2.m_centroid) - b2Mul(xf1, poly1.m_centroid);
   {$ELSE}
   d := Subtract(b2Mul(xf2, poly2.m_centroid), b2Mul(xf1, poly1.m_centroid));
   {$ENDIF}
   dLocal1 := b2MulT(xf1.R, d);

   // Find edge normal on poly1 that has the largest projection onto d.
   edge := 0;
   maxDot := -FLT_MAX;
   for i := 0 to poly1.GetVertexCount - 1 do
   begin
      dot := b2Dot(poly1.m_normals[i], dLocal1);
      if dot > maxDot then
      begin
         maxDot := dot;
         edge := i;
      end;
   end;

   // Get the separation for the edge normal.
   s := EdgeSeparation(poly1, poly2, xf1, xf2, edge);
   if s > 0.0 then
   begin
      Result := s;
      Exit;
   end;

   // Check the separation for the previous edge normal.
   if edge - 1 >= 0 then
      prevEdge := edge - 1
   else
      prevEdge := poly1.GetVertexCount - 1;

   sPrev := EdgeSeparation(poly1, poly2, xf1, xf2, prevEdge);
   if sPrev > 0.0 then
   begin
      Result := sPrev;
      Exit;
   end;

   // Check the separation for the next edge normal.
   if edge + 1 < poly1.GetVertexCount then
      nextEdge := edge + 1
   else
      nextEdge := 0;

   sNext := EdgeSeparation(poly1, poly2, xf1, xf2, nextEdge);
   if sNext > 0.0 then
   begin
      Result := sNext;
      Exit;
   end;

   // Find the best edge and the search direction.
   if (sPrev > s) and (sPrev > sNext) then
   begin
      increment := -1;
      bestEdge := prevEdge;
      bestSeparation := sPrev;
   end
   else if sNext > s then
   begin
      increment := 1;
      bestEdge := nextEdge;
      bestSeparation := sNext;
   end
   else
   begin
      edgeIndex := edge;
      Result := s;
      Exit;
   end;

   // Perform a local search for the best edge normal.
   while True do
   begin
      if increment = -1 then
      begin
         if bestEdge - 1 >= 0 then
            edge := bestEdge - 1
         else
            edge := poly1.GetVertexCount - 1;
      end
      else
         if bestEdge + 1 < poly1.GetVertexCount then
            edge := bestEdge + 1
         else
            edge := 0;

      s := EdgeSeparation(poly1, poly2, xf1, xf2, edge);
      if s > 0.0 then
      begin
         Result := s;
         Exit;
      end;

      if s > bestSeparation then
      begin
         bestEdge := edge;
         bestSeparation := s;
      end
      else
         Break;
   end;

   edgeIndex := bestEdge;
   Result := bestSeparation;
end;

procedure FindIncidentEdge(c: PClipVertices; poly1, poly2: Tb2PolygonShape;
   const xf1, xf2: Tb2XForm; edge1: Int32);
var
   i: Integer;
   index, i1, i2: Int32;
   normal1: TVector2;
   minDot, dot: Float;
begin
   //b2Assert(0 <= edge1 && edge1 < poly1.GetVertexCount);

   // Get the normal of the reference edge in poly2's frame.
   normal1 := b2MulT(xf2.R, b2Mul(xf1.R, poly1.m_normals[edge1]));

   // Find the incident edge on poly2.
   index := 0;
   minDot := FLT_MAX;
   for i := 0 to poly2.GetVertexCount - 1 do
   begin
      dot := b2Dot(normal1, poly2.m_normals[i]);
      if dot < minDot then
      begin
         minDot := dot;
         index := i;
      end;
   end;

   // Build the clip vertices for the incident edge.
   i1 := index;
   if i1 + 1 < poly2.GetVertexCount then
      i2 := i1 + 1
   else
      i2 := 0;

   c^[0].v := b2Mul(xf2, poly2.m_vertices[i1]);
   c^[0].id.referenceEdge := edge1;
   c^[0].id.incidentEdge := i1;
   c^[0].id.incidentVertex := 0;

   c^[1].v := b2Mul(xf2, poly2.m_vertices[i2]);
   c^[1].id.referenceEdge := edge1;
   c^[1].id.incidentEdge := i2;
   c^[1].id.incidentVertex := 1;
end;

// Find edge normal of max separation on A - return if separating axis is found
// Find edge normal of max separation on B - return if separation axis is found
// Choose reference edge as min(minA, minB)
// Find incident edge
// Clip

// The normal points from 1 to 2
procedure b2CollidePolygons(var manifold: Tb2Manifold;
   polyA, polyB: Tb2PolygonShape; xfA, xfB: Tb2XForm);
var
   i: Integer;
   edgeA, edgeB: Int32;
   edge1: Int32; // reference edge
   flip: UInt8;
   separationA, separationB: Float;
   poly1, poly2: Tb2PolygonShape; // reference poly and incident poly
   xf1, xf2: Tb2XForm;
   k_relativeTol, k_absoluteTol: Float;
   incidentEdge, clipPoints1, clipPoints2: TClipVertices;
   v11, v12, sideNormal, frontNormal: TVector2;
   frontOffset, sideOffset1, sideOffset2: Float;
   np: Int32; // Clip incident edge against extruded edge1 side edges.
   pointCount: Int32;
   _separation: Float;
begin
   manifold.pointCount := 0;

   separationA := FindMaxSeparation(edgeA, polyA, polyB, xfA, xfB);
   if separationA > 0.0 then
      Exit;

   edgeB := 0;
   separationB := FindMaxSeparation(edgeB, polyB, polyA, xfB, xfA);
   if separationB > 0.0 then
      Exit;

   k_relativeTol := 0.98;
   k_absoluteTol := 0.001;

   // TODO_ERIN use "radius" of poly for absolute tolerance.
   if separationB > k_relativeTol * separationA + k_absoluteTol then
   begin
      poly1 := polyB;
      poly2 := polyA;
      xf1 := xfB;
      xf2 := xfA;
      edge1 := edgeB;
      flip := 1;
   end
   else
   begin
      poly1 := polyA;
      poly2 := polyB;
      xf1 := xfA;
      xf2 := xfB;
      edge1 := edgeA;
      flip := 0;
   end;

   FindIncidentEdge(@incidentEdge, poly1, poly2, xf1, xf2, edge1);

   v11 := poly1.m_vertices[edge1];
   if edge1 + 1 < poly1.GetVertexCount then
      v12 := poly1.m_vertices[edge1+1]
   else
      v12 := poly1.m_vertices[0];

   {$IFDEF OP_OVERLOAD}
   sideNormal := b2Mul(xf1.R, v12 - v11);
   sideNormal.Normalize;
   {$ELSE}
   sideNormal := b2Mul(xf1.R, Subtract(v12, v11));
   Normalize(sideNormal);
   {$ENDIF}
   frontNormal := b2Cross(sideNormal, 1.0);

   v11 := b2Mul(xf1, v11);
   v12 := b2Mul(xf1, v12);

   frontOffset := b2Dot(frontNormal, v11);
   sideOffset1 := -b2Dot(sideNormal, v11);
   sideOffset2 := b2Dot(sideNormal, v12);

   // Clip to box side 1
   {$IFDEF OP_OVERLOAD}
   np := ClipSegmentToLine(@clipPoints1, @incidentEdge, -sideNormal, sideOffset1);
   {$ELSE}
   np := ClipSegmentToLine(@clipPoints1, @incidentEdge, Negative(sideNormal), sideOffset1);
   {$ENDIF}
   if np < 2 then
      Exit;

   // Clip to negative box side 1
   np := ClipSegmentToLine(@clipPoints2, @clipPoints1, sideNormal, sideOffset2);
   if np < 2 then
      Exit;

   // Now clipPoints2 contains the clipped points.
   if flip <> 0 then
      {$IFDEF OP_OVERLOAD}
      manifold.normal := -frontNormal
      {$ELSE}
      manifold.normal := Negative(frontNormal)
      {$ENDIF}
   else
      manifold.normal := frontNormal;

   pointCount := 0;
   for i := 0 to b2_maxManifoldPoints - 1 do
   begin
      _separation := b2Dot(frontNormal, clipPoints2[i].v) - frontOffset;

      if _separation <= 0.0 then
         with manifold.points[pointCount] do
         begin
            separation := _separation;
            localPoint1 := b2MulT(xfA, clipPoints2[i].v);
            localPoint2 := b2MulT(xfB, clipPoints2[i].v);
            id := clipPoints2[i].id;
            id.flip := flip;
            Inc(pointCount);
         end;
   end;

   manifold.pointCount := pointCount;
end;

{ Tb2CircleDef }

constructor Tb2CircleDef.Create;
begin
   inherited Create;
   ShapeType := e_circleShape;
   SetZero(localPosition);
	 radius := 1.0;
end;

{ Tb2CircleShape }

constructor Tb2CircleShape.Create(def: Tb2ShapeDef);
var
   circleDef: Tb2CircleDef;
begin
   inherited Create(def);
   //b2Assert(def->type == e_circleShape);
   circleDef := Tb2CircleDef(def);
   m_type := e_circleShape;
   m_localPosition := circleDef.localPosition;
   m_radius := circleDef.radius;
end;

procedure Tb2CircleShape.UpdateSweepRadius(const center: TVector2);
begin
 	 // Update the sweep radius (maximum radius) as measured from a local center point.
   {$IFDEF OP_OVERLOAD}
	 m_sweepRadius := (m_localPosition - center).Length() + m_radius - b2_toiSlop;
   {$ELSE}
   m_sweepRadius := Length(Subtract(m_localPosition, center)) + m_radius - b2_toiSlop;
   {$ENDIF}
end;

function Tb2CircleShape.TestPoint(const transform: Tb2XForm; const p: TVector2): Boolean;
var
   center, d: TVector2;
begin
   {$IFDEF OP_OVERLOAD}
 	 center := transform.position + b2Mul(transform.R, m_localPosition);
   d := p - center;
   {$ELSE}
 	 center := Add(transform.position, b2Mul(transform.R, m_localPosition));
   d := Subtract(p, center);
   {$ENDIF}
	 Result := b2Dot(d, d) <= m_radius * m_radius;
end;

function Tb2CircleShape.TestSegment(const xf: Tb2XForm; var lambda: Float;
   var normal: TVector2; const segment: Tb2Segment; maxLambda: Float): Boolean;
var
   position, s, r: TVector2;
   b, c, rr, sigma, a: Float;
begin
   {$IFDEF OP_OVERLOAD}
   position := xf.position + b2Mul(xf.R, m_localPosition);
   s := segment.p1 - position;
   {$ELSE}
   position := Add(xf.position, b2Mul(xf.R, m_localPosition));
   s := Subtract(segment.p1, position);
   {$ENDIF}
   b := b2Dot(s, s) - m_radius * m_radius;

   // Does the segment start inside the circle?
   if b < 0.0 then
   begin
      Result := False;
      Exit;
   end;

   // Solve quadratic equation.
   {$IFDEF OP_OVERLOAD}
   r := segment.p2 - segment.p1;
   {$ELSE}
   r := Subtract(segment.p2, segment.p1);
   {$ENDIF}
   c :=  b2Dot(s, r);
   rr := b2Dot(r, r);
   sigma := c * c - rr * b;

   // Check for negative discriminant and short segment.
   if (sigma < 0.0) or (rr < FLT_EPSILON) then
   begin
      Result := False;
      Exit;
   end;

   // Find the point of intersection of the line with the circle.
   a := -(c + Sqrt(sigma));

   // Is the intersection point on the segment?
   if (0.0 <= a) and (a <= maxLambda * rr) then
   begin
      a := a / rr;
      lambda := a;
      {$IFDEF OP_OVERLOAD}
      normal := s + a * r;
      normal.Normalize;
      {$ELSE}
      normal := Add(s, Multiply(r, a));
      Normalize(normal);
      {$ENDIF}
      Result := True;
   end
   else
      Result := False;
end;

procedure Tb2CircleShape.ComputeAABB(var aabb: Tb2AABB; const xf: Tb2XForm);
var
   p: TVector2;
begin
   {$IFDEF OP_OVERLOAD}
   p := xf.position + b2Mul(xf.R, m_localPosition);
   aabb.lowerBound.SetValue(p.x - m_radius, p.y - m_radius);
   aabb.upperBound.SetValue(p.x + m_radius, p.y + m_radius);
   {$ELSE}
   p := Add(xf.position, b2Mul(xf.R, m_localPosition));
   SetValue(aabb.lowerBound, p.x - m_radius, p.y - m_radius);
   SetValue(aabb.upperBound, p.x + m_radius, p.y + m_radius);
   {$ENDIF}
end;

procedure Tb2CircleShape.ComputeSweptAABB(var aabb: Tb2AABB; const xf1, xf2: Tb2XForm);
var
   p1, p2, lower, upper: TVector2;
begin
   {$IFDEF OP_OVERLOAD}
   p1 := xf1.position + b2Mul(xf1.R, m_localPosition);
   p2 := xf2.position + b2Mul(xf2.R, m_localPosition);
   {$ELSE}
   p1 := Add(xf1.position, b2Mul(xf1.R, m_localPosition));
   p2 := Add(xf2.position, b2Mul(xf2.R, m_localPosition));
   {$ENDIF}
   lower := b2Min(p1, p2);
   upper := b2Max(p1, p2);

   {$IFDEF OP_OVERLOAD}
   aabb.lowerBound.SetValue(lower.x - m_radius, lower.y - m_radius);
   aabb.upperBound.SetValue(upper.x + m_radius, upper.y + m_radius);
   {$ELSE}
   SetValue(aabb.lowerBound, lower.x - m_radius, lower.y - m_radius);
   SetValue(aabb.upperBound, upper.x + m_radius, upper.y + m_radius);
   {$ENDIF}
end;

procedure Tb2CircleShape.ComputeMass(var massData: Tb2MassData);
begin
   massData.mass := m_density * Pi * m_radius * m_radius;
   massData.center := m_localPosition;

   // inertia about the local origin
   massData.I := massData.mass * (0.5 * m_radius * m_radius +
      b2Dot(m_localPosition, m_localPosition));
end;

{ Tb2PolygonDef }

constructor Tb2PolygonDef.Create;
begin
   inherited Create;
	 ShapeType := e_polygonShape;
	 vertexCount := 0;
end;

procedure Tb2PolygonDef.SetAsBox(hx, hy: Float);
begin
   vertexCount := 4;
   {$IFDEF OP_OVERLOAD}
   vertices[0].SetValue(-hx, -hy);
   vertices[1].SetValue(hx, -hy);
   vertices[2].SetValue(hx, hy);
   vertices[3].SetValue(-hx, hy);
   {$ELSE}
   SetValue(vertices[0], -hx, -hy);
   SetValue(vertices[1], hx, -hy);
   SetValue(vertices[2], hx, hy);
   SetValue(vertices[3], -hx, hy);
   {$ENDIF}
end;

procedure Tb2PolygonDef.SetAsBox(hx, hy: Float; const center: TVector2; angle: Float);
var
   i: Integer;
   xf: Tb2XForm;
begin
   SetAsBox(hx, hy);
   xf.position := center;
   {$IFDEF OP_OVERLOAD}
   xf.R.SetValue(angle);
   {$ELSE}
   SetValue(xf.R, angle);
   {$ENDIF}

   for i := 0 to vertexCount - 1 do
      vertices[i] := b2Mul(xf, vertices[i]);
end;

{ Tb2PolygonShape }

constructor Tb2PolygonShape.Create(const def: Tb2ShapeDef);
var
   i: Integer;
   i2: Int32;
   poly: Tb2PolygonDef;
   edge, n1, n2, v, d: TVector2;
   A: TMatrix22;
begin
   inherited Create(def);
   pm_vertices := @m_vertices[0];
   pm_normals := @m_normals[0];
   pm_coreVertices := @m_coreVertices[0];

   //b2Assert(def.type == e_polygonShape);
   m_type := e_polygonShape;
   poly := Tb2PolygonDef(def);

   // Get the vertices transformed into the body frame.
   m_vertexCount := poly.vertexCount;
   //b2Assert(3 <= m_vertexCount && m_vertexCount <= b2_maxPolygonVertices);

   // Copy vertices.
   m_vertices := poly.vertices;

   // Compute normals. Ensure the edges have non-zero length.
   for i := 0 to m_vertexCount - 1 do
   begin
      if i + 1 < m_vertexCount then
         i2 := i + 1
      else
         i2 := 0;      

      {$IFDEF OP_OVERLOAD}
      edge := m_vertices[i2] - m_vertices[i];
      //b2Assert(edge.LengthSquared() > B2_FLT_EPSILON * B2_FLT_EPSILON);
      m_normals[i] := b2Cross(edge, 1.0);
      m_normals[i].Normalize;
      {$ELSE}
      edge := Subtract(m_vertices[i2], m_vertices[i]);
      //b2Assert(edge.LengthSquared() > B2_FLT_EPSILON * B2_FLT_EPSILON);
      m_normals[i] := b2Cross(edge, 1.0);
      Normalize(m_normals[i]);
      {$ENDIF}
   end;

   // Compute the polygon centroid.
   m_centroid := ComputeCentroid(poly.vertices, poly.vertexCount);

   // Compute the oriented bounding box.
   ComputeOBB(m_obb, m_vertices, m_vertexCount);

   // Create core polygon shape by shifting edges inward.
   // Also compute the min/max radius for CCD.
   for i := 0 to m_vertexCount - 1 do
   begin
      if i - 1 >= 0 then
         i2 := i - 1
      else
         i2 := m_vertexCount - 1; 

      n1 := m_normals[i2];
      n2 := m_normals[i];
      {$IFDEF OP_OVERLOAD}
      v := m_vertices[i] - m_centroid;
      {$ELSE}
      v := Subtract(m_vertices[i], m_centroid);
      {$ENDIF}

      d.x := b2Dot(n1, v) - b2_toiSlop;
      d.y := b2Dot(n2, v) - b2_toiSlop;

      // Shifting the edge inward by b2_toiSlop should
      // not cause the plane to pass the centroid.

      // Your shape has a radius/extent less than b2_toiSlop.
      //b2Assert(d.x >= 0.0f);
      //b2Assert(d.y >= 0.0f);
      A.col1.x := n1.x;
      A.col2.x := n1.y;
      A.col1.y := n2.x; 
      A.col2.y := n2.y;
      {$IFDEF OP_OVERLOAD}
      m_coreVertices[i] := A.Solve(d) + m_centroid;
      {$ELSE}
      m_coreVertices[i] := Add(Solve(A, d), m_centroid);
      {$ENDIF}
   end;
end;

procedure Tb2PolygonShape.UpdateSweepRadius(const center: TVector2);
var
   i: Integer;
   d: TVector2;
begin
   // Update the sweep radius (maximum radius) as measured from a local center point.
   m_sweepRadius := 0.0;
   for i := 0 to m_vertexCount - 1 do
   begin
      {$IFDEF OP_OVERLOAD}
      d := m_coreVertices[i] - center;
      m_sweepRadius := b2Max(m_sweepRadius, d.Length());
      {$ELSE}
      d := Subtract(m_coreVertices[i], center);
      m_sweepRadius := b2Max(m_sweepRadius, Length(d));
      {$ENDIF}
   end;
end;

function Tb2PolygonShape.TestPoint(const transform: Tb2XForm;
   const p: TVector2): Boolean;
var
   i: Integer;
   pLocal: TVector2;
begin
   {$IFDEF OP_OVERLOAD}
   pLocal := b2MulT(transform.R, p - transform.position);
   {$ELSE}
   pLocal := b2MulT(transform.R, Subtract(p, transform.position));
   {$ENDIF}

   for i := 0 to m_vertexCount - 1 do
      {$IFDEF OP_OVERLOAD}
      if b2Dot(m_normals[i], pLocal - m_vertices[i]) > 0.0 then
      {$ELSE}
      if b2Dot(m_normals[i], Subtract(pLocal, m_vertices[i])) > 0.0 then
      {$ENDIF}
      begin
         Result := False;
         Exit;
      end;

   Result := True;
end;

function Tb2PolygonShape.TestSegment(const xf: Tb2XForm; var lambda: Float;
   var normal: TVector2; const segment: Tb2Segment; maxLambda: Float): Boolean;
var
   lower, upper: Float;
   p1, p2, d: TVector2;
   index: Int32;
   i: Integer;
   numerator, denominator: Float;
begin
   lower := 0.0;
   upper := maxLambda;

   {$IFDEF OP_OVERLOAD}
   p1 := b2MulT(xf.R, segment.p1 - xf.position);
   p2 := b2MulT(xf.R, segment.p2 - xf.position);
   d := p2 - p1;
   {$ELSE}
   p1 := b2MulT(xf.R, Subtract(segment.p1, xf.position));
   p2 := b2MulT(xf.R, Subtract(segment.p2, xf.position));
   d := Subtract(p2, p1);
   {$ENDIF}
   index := -1;

   for i := 0 to m_vertexCount - 1 do
   begin
      // p := p1 + a * d
      // dot(normal, p - v) := 0
      // dot(normal, p1 - v) + a * dot(normal, d) := 0
      {$IFDEF OP_OVERLOAD}
      numerator := b2Dot(m_normals[i], m_vertices[i] - p1);
      {$ELSE}
      numerator := b2Dot(m_normals[i], Subtract(m_vertices[i], p1));
      {$ENDIF}
      denominator := b2Dot(m_normals[i], d);

      // Note: we want this predicate without division:
      // lower < numerator / denominator, where denominator < 0
      // Since denominator < 0, we have to flip the inequality:
      // lower < numerator / denominator <==> denominator * lower > numerator.

      if (denominator < 0.0) and (numerator < lower * denominator) then
      begin
         // Increase lower.
         // The segment enters this half-space.
         lower := numerator / denominator;
         index := i;
      end
      else if (denominator > 0.0) and (numerator < upper * denominator) then
      begin
         // Decrease upper.
         // The segment exits this half-space.
         upper := numerator / denominator;
      end;

      if upper < lower then
      begin
         Result := False;
         Exit;
      end;
   end;

   //b2Assert(0.0f <= lower && lower <= maxLambda);

   if index >= 0 then
   begin
      lambda := lower;
      normal := b2Mul(xf.R, m_normals[index]);
      Result := True;
   end
   else
      Result := False;
end;

procedure Tb2PolygonShape.ComputeAABB(var aabb: Tb2AABB; const xf: Tb2XForm);
var
   R, absR: TMatrix22;
   h, position: TVector2;
begin
   R := b2Mul(xf.R, m_obb.R);
   absR := b2Abs(R);
   h := b2Mul(absR, m_obb.extents);
   {$IFDEF OP_OVERLOAD}
   position := xf.position + b2Mul(xf.R, m_obb.center);
   aabb.lowerBound := position - h;
   aabb.upperBound := position + h;
   {$ELSE}
   position := Add(xf.position, b2Mul(xf.R, m_obb.center));
   aabb.lowerBound := Subtract(position, h);
   aabb.upperBound := Add(position, h);
   {$ENDIF}
end;

procedure Tb2PolygonShape.ComputeMass(var massData: Tb2MassData);
const
   k_inv3 = 1.0 / 3.0;
var
   i: Integer;
   center, pRef, p1, p2, p3, e1, e2: TVector2;
   area, inertia: Float;
   D, triangleArea: Float;
   px, py, ex1, ey1, ex2, ey2: Float;
   intx2, inty2: Float;
begin
   // Polygon mass, centroid, and inertia.
   // Let rho be the polygon density in mass per unit area.
   // Then:
   // mass = rho * int(dA)
   // centroid.x = (1/mass) * rho * int(x * dA)
   // centroid.y = (1/mass) * rho * int(y * dA)
   // I = rho * int((x*x + y*y) * dA)
   //
   // We can compute these integrals by summing all the integrals
   // for each triangle of the polygon. To evaluate the integral
   // for a single triangle, we make a change of variables to
   // the (u,v) coordinates of the triangle:
   // x = x0 + e1x * u + e2x * v
   // y = y0 + e1y * u + e2y * v
   // where 0 <= u && 0 <= v && u + v <= 1.
   //
   // We integrate u from [0,1-v] and then v from [0,1].
   // We also need to use the Jacobian of the transformation:
   // D = cross(e1, e2)
   //
   // Simplification: triangle centroid = (1/3) * (p1 + p2 + p3)
   //
   // The rest of the derivation is handled by computer algebra.

   //b2Assert(m_vertexCount >= 3);

   SetZero(center);
   area := 0.0;
   inertia := 0.0;

   // pRef is the reference point for forming triangles.
   // It's location doesn't change the result (except for rounding error).
   SetZero(pRef);

   for i := 0 to m_vertexCount - 1 do
   begin
      // Triangle vertices.
      p1 := pRef;
      p2 := m_vertices[i];
      if (i + 1 < m_vertexCount) then
         p3 := m_vertices[i + 1]
      else
         p3 := m_vertices[0];

      {$IFDEF OP_OVERLOAD}
      e1 := p2 - p1;
      e2 := p3 - p1;
      {$ELSE}
      e1 := Subtract(p2, p1);
      e2 := Subtract(p3, p1);
      {$ENDIF}

      D := b2Cross(e1, e2);

      triangleArea := 0.5 * D;
      area := area + triangleArea;

      // Area weighted centroid
      {$IFDEF OP_OVERLOAD}
      center.AddBy(triangleArea * k_inv3 * (p1 + p2 + p3));
      {$ELSE}
      AddBy(center, Multiply(Add(p1, p2, p3), triangleArea * k_inv3));
      {$ENDIF}

      px := p1.x;
      py := p1.y;
      ex1 := e1.x;
      ey1 := e1.y;
      ex2 := e2.x;
      ey2 := e2.y;

      intx2 := k_inv3 * (0.25 * (ex1 * ex1 + ex2 * ex1 + ex2 * ex2) +
         (px * ex1 + px * ex2)) + 0.5 * px * px;
      inty2 := k_inv3 * (0.25 * (ey1 * ey1 + ey2 * ey1 + ey2 * ey2) +
         (py * ey1 + py * ey2)) + 0.5 * py * py;

      inertia := inertia + D * (intx2 + inty2);
   end;

   // Total mass
   massData.mass := m_density * area;

   // Center of mass
   //b2Assert(area > B2_FLT_EPSILON);
   {$IFDEF OP_OVERLOAD}
   center.DivideBy(area);
   {$ELSE}
   DivideBy(center, area);
   {$ENDIF}
   massData.center := center;

   // Inertia tensor relative to the local origin.
   massData.I := m_density * inertia;
end;

procedure Tb2PolygonShape.ComputeSweptAABB(var aabb: Tb2AABB; const xf1,
   xf2: Tb2XForm);
var
   aabb1, aabb2: Tb2AABB;
begin
   ComputeAABB(aabb1, xf1);
   ComputeAABB(aabb2, xf2);
   aabb.lowerBound := b2Min(aabb1.lowerBound, aabb2.lowerBound);
   aabb.upperBound := b2Max(aabb1.upperBound, aabb2.upperBound);
end;

function Tb2PolygonShape.GetFirstVertex(const xf: Tb2XForm): TVector2;
begin
   Result := b2Mul(xf, m_coreVertices[0]);
end;

function Tb2PolygonShape.Centroid(const xf: Tb2XForm): TVector2;
begin
   Result := b2Mul(xf, m_centroid);
end;

function Tb2PolygonShape.Support(const xf: Tb2XForm; const d: TVector2): TVector2;
var
   i: Integer;
   dLocal: TVector2;
   bestIndex: Int32;
   bestValue, value: Float;
begin
   dLocal := b2MulT(xf.R, d);

   bestIndex := 0;
   bestValue := b2Dot(m_coreVertices[0], dLocal);
   for i := 1 to m_vertexCount - 1 do
   begin
      value := b2Dot(m_coreVertices[i], dLocal);
      if value > bestValue then
      begin
         bestIndex := i;
         bestValue := value;
      end;
   end;

   Result := b2Mul(xf, m_coreVertices[bestIndex]);
end;

{ Tb2CircleContact }

constructor Tb2CircleContact.Create(shape1, shape2: Tb2Shape);
begin
   inherited Create(shape1, shape2);
   //b2Assert(m_shape1->GetType() == e_circleShape);
   //b2Assert(m_shape2->GetType() == e_circleShape);
   m_manifold.pointCount := 0;
   m_manifold.points[0].normalImpulse := 0.0;
   m_manifold.points[0].tangentImpulse := 0.0;
end;

procedure Tb2CircleContact.Evaluate(listener: Tb2ContactListener);
var
   b1, b2: Tb2Body;
   m0: Tb2Manifold;
   cp: Tb2ContactPoint;
   v1, v2: TVector2;
begin
   b1 := m_shape1.GetBody;
   b2 := m_shape2.GetBody;

   m0 := m_manifold;

   b2CollideCircles(m_manifold, Tb2CircleShape(m_shape1), Tb2CircleShape(m_shape2),
      b1.m_xf, b2.m_xf);

   cp.shape1 := m_shape1;
   cp.shape2 := m_shape2;
   cp.friction := m_friction;
   cp.restitution := m_restitution;

   if m_manifold.pointCount > 0 then
   begin
      m_manifoldCount := 1;
      with m_manifold.points[0] do
      begin
         if m0.pointCount = 0 then
         begin
            normalImpulse := 0.0;
            tangentImpulse := 0.0;

            if Assigned(listener) then
            begin
               cp.position := b1.GetWorldPoint(localPoint1);
               v1 := b1.GetLinearVelocityFromLocalPoint(localPoint1);
               v2 := b2.GetLinearVelocityFromLocalPoint(localPoint2);
               {$IFDEF OP_OVERLOAD}
               cp.velocity := v2 - v1;
               {$ELSE}
               cp.velocity := Subtract(v2, v1);
               {$ENDIF}
               cp.normal := m_manifold.normal;
               cp.separation := separation;
               cp.id := id;
               listener.Add(cp);
            end;
         end
         else
         begin
            normalImpulse := m0.points[0].normalImpulse;
            tangentImpulse := m0.points[0].tangentImpulse;

            if Assigned(listener) then
            begin
               cp.position := b1.GetWorldPoint(localPoint1);
               v1 := b1.GetLinearVelocityFromLocalPoint(localPoint1);
               v2 := b2.GetLinearVelocityFromLocalPoint(localPoint2);
               {$IFDEF OP_OVERLOAD}
               cp.velocity := v2 - v1;
               {$ELSE}
               cp.velocity := Subtract(v2, v1);
               {$ENDIF}
               cp.normal := m_manifold.normal;
               cp.separation := separation;
               cp.id := id;
               listener.Persist(cp);
            end;
         end;
      end;
   end
   else
   begin
      m_manifoldCount := 0;
      if (m0.pointCount > 0) and Assigned(listener) then
      begin
         with m0.points[0] do
         begin
            cp.position := b1.GetWorldPoint(localPoint1);
            v1 := b1.GetLinearVelocityFromLocalPoint(localPoint1);
            v2 := b2.GetLinearVelocityFromLocalPoint(localPoint2);
            {$IFDEF OP_OVERLOAD}
            cp.velocity := v2 - v1;
            {$ELSE}
            cp.velocity := Subtract(v2, v1);
            {$ENDIF}
            cp.normal := m0.normal;
            cp.separation := separation;
            cp.id := id;
            listener.Remove(cp);
         end;
      end;
   end;
end;

function Tb2CircleContact.GetManifolds: Pb2Manifold;
begin
   Result := @m_manifold;
end;

{ Tb2PolyAndCircleContact }

constructor Tb2PolyAndCircleContact.Create(shape1, shape2: Tb2Shape);
begin
   inherited Create(shape1, shape2);
   //b2Assert(m_shape1->GetType() == e_polygonShape);
	 //b2Assert(m_shape2->GetType() == e_circleShape);
	 m_manifold.pointCount := 0;
   m_manifold.points[0].normalImpulse := 0.0;
	 m_manifold.points[0].tangentImpulse := 0.0;
end;

procedure Tb2PolyAndCircleContact.Evaluate(listener: Tb2ContactListener);
var
   i, j: Integer;
   b1, b2: Tb2Body;
   m0: Tb2Manifold;
   persisted: array[0..b2_maxManifoldPoints - 1] of Boolean;
   cp: Tb2ContactPoint;
   found: Boolean;
   mp0: Pb2ManifoldPoint;
   local_id: Tb2ContactID;
   v1, v2: TVector2;
begin
   b1 := m_shape1.GetBody;
   b2 := m_shape2.GetBody;

   m0 := m_manifold;

   b2CollidePolygonAndCircle(m_manifold, Tb2PolygonShape(m_shape1),
      Tb2CircleShape(m_shape2), b1.m_xf, b2.m_xf);

   persisted[0] := False;
   persisted[1] := False;

   cp.shape1 := m_shape1;
   cp.shape2 := m_shape2;
   cp.friction := m_friction;
   cp.restitution := m_restitution;

   // Match contact ids to facilitate warm starting.
   if m_manifold.pointCount > 0 then
   begin
      // Match old contact ids to new contact ids and copy the
      // stored impulses to warm start the solver.
      for i := 0 to m_manifold.pointCount - 1 do
         with m_manifold.points[i] do
         begin
            normalImpulse := 0.0;
            tangentImpulse := 0.0;
            found := False;
            local_id := id;

            for j := 0 to m0.pointCount - 1 do
            begin
               if persisted[j] then
                  Continue;

               mp0 := @m0.points[j];
               if mp0.id.key = local_id.key then
               begin
                  persisted[j] := True;
                  normalImpulse := mp0.normalImpulse;
                  tangentImpulse := mp0.tangentImpulse;

                  // A persistent point.
                  found := True;

                  // Report persistent point.
                  if Assigned(listener) then
                  begin
                     cp.position := b1.GetWorldPoint(localPoint1);
                     v1 := b1.GetLinearVelocityFromLocalPoint(localPoint1);
                     v2 := b2.GetLinearVelocityFromLocalPoint(localPoint2);
                     {$IFDEF OP_OVERLOAD}
                     cp.velocity := v2 - v1;
                     {$ELSE}
                     cp.velocity := Subtract(v2, v1);
                     {$ENDIF}
                     cp.normal := m_manifold.normal;
                     cp.separation := separation;
                     cp.id := local_id;
                     listener.Persist(cp);
                  end;
                  Break;
               end;
            end;

            // Report added point.
            if (not found) and Assigned(listener) then
            begin
               cp.position := b1.GetWorldPoint(localPoint1);
               v1 := b1.GetLinearVelocityFromLocalPoint(localPoint1);
               v2 := b2.GetLinearVelocityFromLocalPoint(localPoint2);
               {$IFDEF OP_OVERLOAD}
               cp.velocity := v2 - v1;
               {$ELSE}
               cp.velocity := Subtract(v2, v1);
               {$ENDIF}
               cp.normal := m_manifold.normal;
               cp.separation := separation;
               cp.id := local_id;
               listener.Add(cp);
            end;
         end;

      m_manifoldCount := 1;
   end
   else
      m_manifoldCount := 0;

   if not Assigned(listener) then
      Exit;

   // Report removed points.
   for i := 0 to m0.pointCount - 1 do
   begin
      if persisted[i] then
         Continue;

      with m0.points[i] do
      begin
         cp.position := b1.GetWorldPoint(localPoint1);
         v1 := b1.GetLinearVelocityFromLocalPoint(localPoint1);
         v2 := b2.GetLinearVelocityFromLocalPoint(localPoint2);
         {$IFDEF OP_OVERLOAD}
         cp.velocity := v2 - v1;
         {$ELSE}
         cp.velocity := Subtract(v2, v1);
         {$ENDIF}
         cp.normal := m0.normal;
         cp.separation := separation;
         cp.id := id;
         listener.Remove(cp);
      end;
   end;
end;

function Tb2PolyAndCircleContact.GetManifolds: Pb2Manifold;
begin
   Result := @m_manifold;
end;

{ Tb2PolygonContact }

constructor Tb2PolygonContact.Create(shape1, shape2: Tb2Shape);
begin
   inherited Create(shape1, shape2);
	 //b2Assert(m_shape1->GetType() == e_polygonShape);
	 //b2Assert(m_shape2->GetType() == e_polygonShape);
	 m_manifold.pointCount := 0;
end;

procedure Tb2PolygonContact.Evaluate(listener: Tb2ContactListener);
var
   i, j: Integer;
   b1, b2: Tb2Body;
   m0: Tb2Manifold;
   persisted: array[0..b2_maxManifoldPoints - 1] of Boolean;
   cp: Tb2ContactPoint;
   v1, v2: TVector2;
   found: Boolean;
   local_id: Tb2ContactID;
   mp0: Pb2ManifoldPoint;
begin
   b1 := m_shape1.GetBody;
   b2 := m_shape2.GetBody;

   m0 := m_manifold;

   b2CollidePolygons(m_manifold, Tb2PolygonShape(m_shape1), Tb2PolygonShape(m_shape2),
      b1.m_xf, b2.m_xf);

   persisted[0] := False;
   persisted[1] := False;

   cp.shape1 := m_shape1;
   cp.shape2 := m_shape2;
   cp.friction := m_friction;
   cp.restitution := m_restitution;

   // Match contact ids to facilitate warm starting.
   if m_manifold.pointCount > 0 then
   begin
      // Match old contact ids to new contact ids and copy the
      // stored impulses to warm start the solver.
      for i := 0 to m_manifold.pointCount - 1 do
         with m_manifold.points[i] do
         begin
            normalImpulse := 0.0;
            tangentImpulse := 0.0;
            found := False;
            local_id := id;

            for j := 0 to m0.pointCount - 1 do
            begin
               if persisted[j] then
                  Continue;

               mp0 := @m0.points[j];

               if mp0.id.key = id.key then
               begin
                  persisted[j] := True;
                  normalImpulse := mp0.normalImpulse;
                  tangentImpulse := mp0.tangentImpulse;

                  // A persistent point.
                  found := True;

                  // Report persistent point.
                  if Assigned(listener) then
                  begin
                     cp.position := b1.GetWorldPoint(localPoint1);
                     v1 := b1.GetLinearVelocityFromLocalPoint(localPoint1);
                     v2 := b2.GetLinearVelocityFromLocalPoint(localPoint2);
                     {$IFDEF OP_OVERLOAD}
                     cp.velocity := v2 - v1;
                     {$ELSE}
                     cp.velocity := Subtract(v2, v1);
                     {$ENDIF}
                     cp.normal := m_manifold.normal;
                     cp.separation := separation;
                     cp.id := local_id;
                     listener.Persist(cp);
                  end;
                  Break;
               end;
            end;

            // Report added point.
            if (not found) and Assigned(listener) then
            begin
               cp.position := b1.GetWorldPoint(localPoint1);
               v1 := b1.GetLinearVelocityFromLocalPoint(localPoint1);
               v2 := b2.GetLinearVelocityFromLocalPoint(localPoint2);
               {$IFDEF OP_OVERLOAD}
               cp.velocity := v2 - v1;
               {$ELSE}
               cp.velocity := Subtract(v2, v1);
               {$ENDIF}
               cp.normal := m_manifold.normal;
               cp.separation := separation;
               cp.id := local_id;
               listener.Add(cp);
            end;
         end;

      m_manifoldCount := 1;
   end
   else
      m_manifoldCount := 0;

   if not Assigned(listener) then
      Exit;

   // Report removed points.
   for i := 0 to m0.pointCount - 1 do
   begin
      if persisted[i] then
         Continue;

      with m0.points[i] do
      begin
         cp.position := b1.GetWorldPoint(localPoint1);
         v1 := b1.GetLinearVelocityFromLocalPoint(localPoint1);
         v2 := b2.GetLinearVelocityFromLocalPoint(localPoint2);
         {$IFDEF OP_OVERLOAD}
         cp.velocity := v2 - v1;
         {$ELSE}
         cp.velocity := Subtract(v2, v1);
         {$ENDIF}
         cp.normal := m0.normal;
         cp.separation := separation;
         cp.id := id;
         listener.Remove(cp);
      end;
   end;
end;

function Tb2PolygonContact.GetManifolds: Pb2Manifold;
begin
   Result := @m_manifold;
end;

{ Tb2DistanceJointDef }

// 1-D constrained system
// m (v2 - v1) = lambda
// v2 + (beta/h) * x1 + gamma * lambda = 0, gamma has units of inverse mass.
// x2 = x1 + h * v2

// 1-D mass-damper-spring system
// m (v2 - v1) + h * d * v2 + h * k *

// C = norm(p2 - p1) - L
// u = (p2 - p1) / norm(p2 - p1)
// Cdot = dot(u, v2 + cross(w2, r2) - v1 - cross(w1, r1))
// J = [-u -cross(r1, u) u cross(r2, u)]
// K = J * invM * JT
//   = invMass1 + invI1 * cross(r1, u)^2 + invMass2 + invI2 * cross(r2, u)^2

constructor Tb2DistanceJointDef.Create;
begin
   JointType := e_distanceJoint;
   SetZero(localAnchor1);
   SetZero(localAnchor2);
   length := 1.0;
   frequencyHz := 0.0;
   dampingRatio := 0.0;
end;

procedure Tb2DistanceJointDef.Initialize(body1, body2: Tb2Body; const anchor1,
   anchor2: TVector2);
begin
   Self.body1 := body1;
   Self.body2 := body2;
   localAnchor1 := body1.GetLocalPoint(anchor1);
   localAnchor2 := body2.GetLocalPoint(anchor2);
   {$IFDEF OP_OVERLOAD}
   length := (anchor2 - anchor1).Length;
   {$ELSE}
   length := UPhysics2DTypes.Length(Subtract(anchor2, anchor1));
   {$ENDIF}
end;

{ Tb2DistanceJoint }

constructor Tb2DistanceJoint.Create(def: Tb2DistanceJointDef);
begin
   inherited Create(def);
   m_localAnchor1 := def.localAnchor1;
   m_localAnchor2 := def.localAnchor2;
   m_length := def.length;
   m_frequencyHz := def.frequencyHz;
   m_dampingRatio := def.dampingRatio;
   m_impulse := 0.0;
   m_gamma := 0.0;
   m_bias := 0.0;
   m_inv_dt := 0.0;
end;

function Tb2DistanceJoint.GetAnchor1: TVector2;
begin
   Result := m_body1.GetWorldPoint(m_localAnchor1);
end;

function Tb2DistanceJoint.GetAnchor2: TVector2;
begin
   Result := m_body2.GetWorldPoint(m_localAnchor2);
end;

function Tb2DistanceJoint.GetReactionForce: TVector2;
begin
   {$IFDEF OP_OVERLOAD}
   Result := (m_inv_dt * m_impulse) * m_u;
   {$ELSE}
   Result := Multiply(m_u, m_inv_dt * m_impulse);
   {$ENDIF}
end;

function Tb2DistanceJoint.GetReactionTorque: Float;
begin
   Result := 0.0;
end;

procedure Tb2DistanceJoint.InitVelocityConstraints(const step: Tb2TimeStep);
var
   r1, r2, P: TVector2;
   length, cr1u, cr2u, invMass: Float;
   C, omega, d, k: Float;
begin
   m_inv_dt := step.inv_dt;

   // Compute the effective mass matrix.
   {$IFDEF OP_OVERLOAD}
   r1 := b2Mul(m_body1.m_xf.R, m_localAnchor1 - m_body1.GetLocalCenter);
   r2 := b2Mul(m_body2.m_xf.R, m_localAnchor2 - m_body2.GetLocalCenter);
   m_u := m_body2.m_sweep.c + r2 - m_body1.m_sweep.c - r1;
   {$ELSE}
   r1 := b2Mul(m_body1.m_xf.R, Subtract(m_localAnchor1, m_body1.GetLocalCenter));
   r2 := b2Mul(m_body2.m_xf.R, Subtract(m_localAnchor2, m_body2.GetLocalCenter));
   m_u := Subtract(r2, r1);
   AddBy(m_u, m_body2.m_sweep.c);
   SubtractBy(m_u, m_body1.m_sweep.c);
   {$ENDIF}

   // Handle singularity.
   {$IFDEF OP_OVERLOAD}
   length := m_u.Length;
   if length > b2_linearSlop then
      m_u := m_u / length
   else
      m_u.SetZero;
   {$ELSE}
   length := UPhysics2DTypes.Length(m_u);
   if length > b2_linearSlop then
      DivideBy(m_u, length)
   else
      SetZero(m_u);
   {$ENDIF}

   cr1u := b2Cross(r1, m_u);
   cr2u := b2Cross(r2, m_u);
   invMass := m_body1.m_invMass + m_body1.m_invI * cr1u * cr1u +
      m_body2.m_invMass + m_body2.m_invI * cr2u * cr2u;
   //b2Assert(invMass > B2_FLT_EPSILON);
   m_mass := 1.0 / invMass;

   if m_frequencyHz > 0.0 then
   begin
      C := length - m_length;
      omega := 2.0 * Pi * m_frequencyHz; // Frequency
      d := 2.0 * m_mass * m_dampingRatio * omega; // Damping coefficient
      k := m_mass * omega * omega; // Spring stiffness
      m_gamma := 1.0 / (step.dt * (d + step.dt * k)); // magic formulas
      m_bias := C * step.dt * k * m_gamma;
      m_mass := 1.0 / (invMass + m_gamma);
   end;

   if step.warmStarting then
   begin
      m_impulse := m_impulse * step.dtRatio;
      {$IFDEF OP_OVERLOAD}
      P := m_impulse * m_u;
      m_body1.m_linearVelocity.SubtractBy(m_body1.m_invMass * P);
      m_body2.m_linearVelocity.AddBy(m_body2.m_invMass * P);
      {$ELSE}
      P := Multiply(m_u, m_impulse);
      SubtractBy(m_body1.m_linearVelocity, Multiply(P, m_body1.m_invMass));
      AddBy(m_body2.m_linearVelocity, Multiply(P, m_body2.m_invMass));
      {$ENDIF}
      m_body1.m_angularVelocity := m_body1.m_angularVelocity - m_body1.m_invI * b2Cross(r1, P);
      m_body2.m_angularVelocity := m_body2.m_angularVelocity + m_body2.m_invI * b2Cross(r2, P);
   end
   else
      m_impulse := 0.0;
end;

procedure Tb2DistanceJoint.SolveVelocityConstraints(const step: Tb2TimeStep);
var
   r1, r2, v1, v2, P: TVector2;
   Cdot, impulse: Float;
begin
   {$IFDEF OP_OVERLOAD}
   r1 := b2Mul(m_body1.m_xf.R, m_localAnchor1 - m_body1.GetLocalCenter);
   r2 := b2Mul(m_body2.m_xf.R, m_localAnchor2 - m_body2.GetLocalCenter);

   // Cdot = dot(u, v + cross(w, r))
   v1 := m_body1.m_linearVelocity + b2Cross(m_body1.m_angularVelocity, r1);
   v2 := m_body2.m_linearVelocity + b2Cross(m_body2.m_angularVelocity, r2);
   Cdot := b2Dot(m_u, v2 - v1);
   {$ELSE}
   r1 := b2Mul(m_body1.m_xf.R, Subtract(m_localAnchor1, m_body1.GetLocalCenter));
   r2 := b2Mul(m_body2.m_xf.R, Subtract(m_localAnchor2, m_body2.GetLocalCenter));

   // Cdot = dot(u, v + cross(w, r))
   v1 := Add(m_body1.m_linearVelocity, b2Cross(m_body1.m_angularVelocity, r1));
   v2 := Add(m_body2.m_linearVelocity, b2Cross(m_body2.m_angularVelocity, r2));
   Cdot := b2Dot(m_u, Subtract(v2, v1));
   {$ENDIF}

   impulse := -m_mass * (Cdot + m_bias + m_gamma * m_impulse);
   m_impulse := m_impulse + impulse;

   {$IFDEF OP_OVERLOAD}
   P := impulse * m_u;
   m_body1.m_linearVelocity.SubtractBy(m_body1.m_invMass * P);
   m_body2.m_linearVelocity.AddBy(m_body2.m_invMass * P);
   {$ELSE}
   P := Multiply(m_u, impulse);
   SubtractBy(m_body1.m_linearVelocity, Multiply(P, m_body1.m_invMass));
   AddBy(m_body2.m_linearVelocity, Multiply(P, m_body2.m_invMass));
   {$ENDIF}
   m_body1.m_angularVelocity := m_body1.m_angularVelocity - m_body1.m_invI * b2Cross(r1, P);
   m_body2.m_angularVelocity := m_body2.m_angularVelocity + m_body2.m_invI * b2Cross(r2, P);
end;

function Tb2DistanceJoint.SolvePositionConstraints: Boolean;
var
   r1, r2, d, P: TVector2;
   C, impulse: Float;
begin
   if m_frequencyHz > 0.0 then
   begin
      Result := True;
      Exit;
   end;

   {$IFDEF OP_OVERLOAD}
   r1 := b2Mul(m_body1.m_xf.R, m_localAnchor1 - m_body1.GetLocalCenter);
   r2 := b2Mul(m_body2.m_xf.R, m_localAnchor2 - m_body2.GetLocalCenter);
   d := m_body2.m_sweep.c + r2 - m_body1.m_sweep.c - r1;
   C := d.Normalize() - m_length;
   {$ELSE}
   r1 := b2Mul(m_body1.m_xf.R, Subtract(m_localAnchor1, m_body1.GetLocalCenter));
   r2 := b2Mul(m_body2.m_xf.R, Subtract(m_localAnchor2, m_body2.GetLocalCenter));
   d := Subtract(r2, r1);
   AddBy(d, m_body2.m_sweep.c);
   SubtractBy(d, m_body1.m_sweep.c);
   C := Normalize(d) - m_length;
   {$ENDIF}
   C := b2Clamp(C, -b2_maxLinearCorrection, b2_maxLinearCorrection);

   impulse := -m_mass * C;
   m_u := d;
   {$IFDEF OP_OVERLOAD}
   P := impulse * m_u;

   m_body1.m_sweep.c.SubtractBy(m_body1.m_invMass * P);
   m_body2.m_sweep.c.AddBy(m_body2.m_invMass * P);
   {$ELSE}
   P := Multiply(m_u, impulse);

   SubtractBy(m_body1.m_sweep.c, Multiply(P, m_body1.m_invMass));
   AddBy(m_body2.m_sweep.c, Multiply(P, m_body2.m_invMass));
   {$ENDIF}
   m_body1.m_sweep.a := m_body1.m_sweep.a - m_body1.m_invI * b2Cross(r1, P);
   m_body2.m_sweep.a := m_body2.m_sweep.a + m_body2.m_invI * b2Cross(r2, P);

   m_body1.SynchronizeTransform;
   m_body2.SynchronizeTransform;

   Result := Abs(C) < b2_linearSlop;
end;

{ Tb2PrismaticJointDef }

// Linear constraint (point-to-line)
// d = p2 - p1 = x2 + r2 - x1 - r1
// C = dot(ay1, d)
// Cdot = dot(d, cross(w1, ay1)) + dot(ay1, v2 + cross(w2, r2) - v1 - cross(w1, r1))
//      = -dot(ay1, v1) - dot(cross(d + r1, ay1), w1) + dot(ay1, v2) + dot(cross(r2, ay1), v2)
// J = [-ay1 -cross(d+r1,ay1) ay1 cross(r2,ay1)]
//
// Angular constraint
// C = a2 - a1 + a_initial
// Cdot = w2 - w1
// J = [0 0 -1 0 0 1]

// Motor/Limit linear constraint
// C = dot(ax1, d)
// Cdot = = -dot(ax1, v1) - dot(cross(d + r1, ax1), w1) + dot(ax1, v2) + dot(cross(r2, ax1), v2)
// J = [-ax1 -cross(d+r1,ax1) ax1 cross(r2,ax1)]

constructor Tb2PrismaticJointDef.Create;
begin
   JointType := e_prismaticJoint;
   SetZero(localAnchor1);
   SetZero(localAnchor2);
   {$IFDEF OP_OVERLOAD}
   localAxis1.SetValue(1.0, 0.0);
   {$ELSE}
   SetValue(localAxis1, 1.0, 0.0);
   {$ENDIF}
   referenceAngle := 0.0;
   enableLimit := False;
   lowerTranslation := 0.0;
   upperTranslation := 0.0;
   enableMotor := False;
   maxMotorForce := 0.0;
   motorSpeed := 0.0;
end;

procedure Tb2PrismaticJointDef.Initialize(body1, body2: Tb2Body; const anchor,
   axis: TVector2);
begin
   Self.body1 := body1;
   Self.body2 := body2;
   localAnchor1 := body1.GetLocalPoint(anchor);
   localAnchor2 := body2.GetLocalPoint(anchor);
   localAxis1 := body1.GetLocalVector(axis);
   referenceAngle := body2.GetAngle - body1.GetAngle;
end;

{ Tb2PrismaticJoint }

constructor Tb2PrismaticJoint.Create(def: Tb2PrismaticJointDef);
begin
   inherited Create(def);
   m_localAnchor1 := def.localAnchor1;
   m_localAnchor2 := def.localAnchor2;
   m_localXAxis1 := def.localAxis1;
   m_localYAxis1 := b2Cross(1.0, m_localXAxis1);
   m_refAngle := def.referenceAngle;

   {$IFDEF OP_OVERLOAD}
   m_linearJacobian.SetZero;
   m_motorJacobian.SetZero;
   {$ELSE}
   SetZero(m_linearJacobian);
   SetZero(m_motorJacobian);
   {$ENDIF}
   m_linearMass := 0.0;
   m_force := 0.0;

   m_angularMass := 0.0;
   m_torque := 0.0;

   m_motorMass := 0.0;
   m_motorForce := 0.0;
   m_limitForce := 0.0;
   m_limitPositionImpulse := 0.0;

   m_lowerTranslation := def.lowerTranslation;
   m_upperTranslation := def.upperTranslation;
   m_maxMotorForce := def.maxMotorForce;
   m_motorSpeed := def.motorSpeed;
   m_enableLimit := def.enableLimit;
   m_enableMotor := def.enableMotor;
end;

function Tb2PrismaticJoint.GetAnchor1: TVector2;
begin
   Result := m_body1.GetWorldPoint(m_localAnchor1);
end;

function Tb2PrismaticJoint.GetAnchor2: TVector2;
begin
   Result := m_body2.GetWorldPoint(m_localAnchor2);
end;

function Tb2PrismaticJoint.GetReactionForce: TVector2;
var
   ax1, ay1: TVector2;
begin
   ax1 := b2Mul(m_body1.m_xf.R, m_localXAxis1);
	 ay1 := b2Mul(m_body1.m_xf.R, m_localYAxis1);
   {$IFDEF OP_OVERLOAD}
   Result := m_limitForce * ax1 + m_force * ay1;
   {$ELSE}
   Result := Add(Multiply(ax1, m_limitForce), Multiply(ay1, m_force));
   {$ENDIF}
end;

function Tb2PrismaticJoint.GetReactionTorque: Float;
begin
   Result := m_torque;
end;

procedure Tb2PrismaticJoint.InitVelocityConstraints(const step: Tb2TimeStep);
var
   r1, r2, ay1, e, ax1, d: TVector2;
   invMass1, invMass2, invI1, invI2, jointTranslation, L1, L2: Float;
begin
   // Compute the effective masses.
   {$IFDEF OP_OVERLOAD}
   r1 := b2Mul(m_body1.m_xf.R, m_localAnchor1 - m_body1.GetLocalCenter);
   r2 := b2Mul(m_body2.m_xf.R, m_localAnchor2 - m_body2.GetLocalCenter);
   {$ELSE}
   r1 := b2Mul(m_body1.m_xf.R, Subtract(m_localAnchor1, m_body1.GetLocalCenter));
   r2 := b2Mul(m_body2.m_xf.R, Subtract(m_localAnchor2, m_body2.GetLocalCenter));
   {$ENDIF}

   invMass1 := m_body1.m_invMass;
   invMass2 := m_body2.m_invMass;
   invI1 := m_body1.m_invI;
   invI2 := m_body2.m_invI;

   // Compute point to line constraint effective mass.
   // J := [-ay1 -cross(d+r1,ay1) ay1 cross(r2,ay1)]
   ay1 := b2Mul(m_body1.m_xf.R, m_localYAxis1);
   {$IFDEF OP_OVERLOAD}
   e := m_body2.m_sweep.c + r2 - m_body1.m_sweep.c;	// e := d + r1
   m_linearJacobian.SetValue(-ay1, ay1, -b2Cross(e, ay1), b2Cross(r2, ay1));
   {$ELSE}
   e := Subtract(m_body2.m_sweep.c, m_body1.m_sweep.c);
   AddBy(e, r2);
   SetValue(m_linearJacobian, Negative(ay1), ay1, -b2Cross(e, ay1), b2Cross(r2, ay1));
   {$ENDIF}
   m_linearMass := invMass1 + invI1 * m_linearJacobian.angular1 *
      m_linearJacobian.angular1 + invMass2 + invI2 *
      m_linearJacobian.angular2 * m_linearJacobian.angular2;
   //b2Assert(m_linearMass > B2_FLT_EPSILON);
   m_linearMass := 1.0 / m_linearMass;

   // Compute angular constraint effective mass.
   m_angularMass := invI1 + invI2;
   if m_angularMass > FLT_EPSILON then
      m_angularMass := 1.0 / m_angularMass;

   // Compute motor and limit terms.
   if m_enableLimit or m_enableMotor then
   begin
      // The motor and limit share a Jacobian and effective mass.
      ax1 := b2Mul(m_body1.m_xf.R, m_localXAxis1);
      {$IFDEF OP_OVERLOAD}
      m_motorJacobian.SetValue(-ax1, ax1, -b2Cross(e, ax1), b2Cross(r2, ax1));
      {$ELSE}
      SetValue(m_motorJacobian, Negative(ax1), ax1, -b2Cross(e, ax1), b2Cross(r2, ax1));
      {$ENDIF}
      m_motorMass :=	invMass1 + invI1 * m_motorJacobian.angular1 * m_motorJacobian.angular1 +
              invMass2 + invI2 * m_motorJacobian.angular2 * m_motorJacobian.angular2;
      //b2Assert(m_motorMass > B2_FLT_EPSILON);
      m_motorMass := 1.0 / m_motorMass;

      if m_enableLimit then
      begin
         {$IFDEF OP_OVERLOAD}
         d := e - r1;	// p2 - p1
         {$ELSE}
         d := Subtract(e, r1);	// p2 - p1
         {$ENDIF}
         jointTranslation := b2Dot(ax1, d);
         if Abs(m_upperTranslation - m_lowerTranslation) < 2.0 * b2_linearSlop then
            m_limitState := e_equalLimits
         else if jointTranslation <= m_lowerTranslation then
         begin
            if m_limitState <> e_atLowerLimit then
               m_limitForce := 0.0;
            m_limitState := e_atLowerLimit;
         end
         else if jointTranslation >= m_upperTranslation then
         begin
            if m_limitState <> e_atUpperLimit then
               m_limitForce := 0.0;
            m_limitState := e_atUpperLimit;
         end
         else
         begin
            m_limitState := e_inactiveLimit;
            m_limitForce := 0.0;
         end;
      end;
   end;

   if not m_enableMotor then
      m_motorForce := 0.0;

   if not m_enableLimit then
      m_limitForce := 0.0;

   if step.warmStarting then
   begin
      {$IFDEF OP_OVERLOAD}
      r1 := step.dt * (m_force * m_linearJacobian.linear1 + (m_motorForce +
         m_limitForce) * m_motorJacobian.linear1);
      r2 := step.dt * (m_force * m_linearJacobian.linear2 + (m_motorForce +
         m_limitForce) * m_motorJacobian.linear2);
      {$ELSE}
      r1 := Multiply(m_motorJacobian.linear1, m_motorForce + m_limitForce);
      AddBy(r1, Multiply(m_linearJacobian.linear1, m_force));
      MultiplyBy(r1, step.dt);
      r2 := Multiply(m_motorJacobian.linear2, m_motorForce + m_limitForce);
      AddBy(r2, Multiply(m_linearJacobian.linear2, m_force));
      MultiplyBy(r2, step.dt);
      {$ENDIF}
      L1 := step.dt * (m_force * m_linearJacobian.angular1 - m_torque +
         (m_motorForce + m_limitForce) * m_motorJacobian.angular1);
      L2 := step.dt * (m_force * m_linearJacobian.angular2 + m_torque +
         (m_motorForce + m_limitForce) * m_motorJacobian.angular2);

      {$IFDEF OP_OVERLOAD}
      m_body1.m_linearVelocity.AddBy(invMass1 * r1);
      m_body2.m_linearVelocity.AddBy(invMass2 * r2);
      {$ELSE}
      AddBy(m_body1.m_linearVelocity, Multiply(r1, invMass1));
      AddBy(m_body2.m_linearVelocity, Multiply(r2, invMass2));
      {$ENDIF}

      m_body1.m_angularVelocity := m_body1.m_angularVelocity + invI1 * L1;
      m_body2.m_angularVelocity := m_body2.m_angularVelocity + invI2 * L2;
   end
   else
   begin
      m_force := 0.0;
      m_torque := 0.0;
      m_limitForce := 0.0;
      m_motorForce := 0.0;
   end;

   m_limitPositionImpulse := 0.0;
end;

procedure Tb2PrismaticJoint.SolveVelocityConstraints(const step: Tb2TimeStep);
var
   invMass1, invMass2, invI1, invI2: Float;
   linearCdot, force, P, angularCdot, torque, L: Float;
   motorCdot, motorForce, oldMotorForce: Float;
   limitCdot, limitForce, oldLimitForce: Float;
begin
   invMass1 := m_body1.m_invMass;
   invMass2 := m_body2.m_invMass;
   invI1 := m_body1.m_invI;
   invI2 := m_body2.m_invI;

   // Solve linear constraint.
   {$IFDEF OP_OVERLOAD}
   linearCdot := m_linearJacobian.Compute(m_body1.m_linearVelocity,
      m_body2.m_linearVelocity, m_body1.m_angularVelocity, m_body2.m_angularVelocity);
   {$ELSE}
   linearCdot := Compute(m_linearJacobian, m_body1.m_linearVelocity,
      m_body2.m_linearVelocity, m_body1.m_angularVelocity, m_body2.m_angularVelocity);
   {$ENDIF}
   force := -step.inv_dt * m_linearMass * linearCdot;
   m_force := m_force + force;

   P := step.dt * force;
   {$IFDEF OP_OVERLOAD}
   m_body1.m_linearVelocity.AddBy((invMass1 * P) * m_linearJacobian.linear1);
   m_body2.m_linearVelocity.AddBy((invMass2 * P) * m_linearJacobian.linear2);
   {$ELSE}
   AddBy(m_body1.m_linearVelocity, Multiply(m_linearJacobian.linear1, invMass1 * P));
   AddBy(m_body2.m_linearVelocity, Multiply(m_linearJacobian.linear2, invMass2 * P));
   {$ENDIF}
   m_body1.m_angularVelocity := m_body1.m_angularVelocity + invI1 * P * m_linearJacobian.angular1;
   m_body2.m_angularVelocity := m_body2.m_angularVelocity + invI2 * P * m_linearJacobian.angular2;

   // Solve angular constraint.
   angularCdot := m_body2.m_angularVelocity - m_body1.m_angularVelocity;
   torque := -step.inv_dt * m_angularMass * angularCdot;
   m_torque := m_torque + torque;

   L := step.dt * torque;
   m_body1.m_angularVelocity := m_body1.m_angularVelocity - invI1 * L;
   m_body2.m_angularVelocity := m_body2.m_angularVelocity + invI2 * L;

   // Solve linear motor constraint.
   if m_enableMotor and (m_limitState <> e_equalLimits) then
   begin
      {$IFDEF OP_OVERLOAD}
      motorCdot := m_motorJacobian.Compute(m_body1.m_linearVelocity,
         m_body2.m_linearVelocity, m_body1.m_angularVelocity,
         m_body2.m_angularVelocity) - m_motorSpeed;
      {$ELSE}
      motorCdot := Compute(m_motorJacobian, m_body1.m_linearVelocity,
         m_body2.m_linearVelocity, m_body1.m_angularVelocity,
         m_body2.m_angularVelocity) - m_motorSpeed;
      {$ENDIF}
      motorForce := -step.inv_dt * m_motorMass * motorCdot;
      oldMotorForce := m_motorForce;
      m_motorForce := b2Clamp(m_motorForce + motorForce, -m_maxMotorForce, m_maxMotorForce);
      motorForce := m_motorForce - oldMotorForce;

      P := step.dt * motorForce;
      {$IFDEF OP_OVERLOAD}
      m_body1.m_linearVelocity.AddBy((invMass1 * P) * m_motorJacobian.linear1);
      m_body2.m_linearVelocity.AddBy((invMass2 * P) * m_motorJacobian.linear2);
      {$ELSE}
      AddBy(m_body1.m_linearVelocity, Multiply(m_motorJacobian.linear1, invMass1 * P));
      AddBy(m_body2.m_linearVelocity, Multiply(m_motorJacobian.linear2, invMass2 * P));
      {$ENDIF}
      m_body1.m_angularVelocity := m_body1.m_angularVelocity + invI1 * P * m_motorJacobian.angular1;
      m_body2.m_angularVelocity := m_body2.m_angularVelocity + invI2 * P * m_motorJacobian.angular2;
   end;

   // Solve linear limit constraint.
   if m_enableLimit and (m_limitState <> e_inactiveLimit) then
   begin
      {$IFDEF OP_OVERLOAD}
      limitCdot := m_motorJacobian.Compute(m_body1.m_linearVelocity,
         m_body2.m_linearVelocity, m_body1.m_angularVelocity, m_body2.m_angularVelocity);
      {$ELSE}
      limitCdot := Compute(m_motorJacobian, m_body1.m_linearVelocity,
         m_body2.m_linearVelocity, m_body1.m_angularVelocity, m_body2.m_angularVelocity);
      {$ENDIF}
      limitForce := -step.inv_dt * m_motorMass * limitCdot;

      case m_limitState of
         e_equalLimits: m_limitForce := m_limitForce + limitForce;
         e_atLowerLimit:
            begin
               oldLimitForce := m_limitForce;
               m_limitForce := b2Max(m_limitForce + limitForce, 0.0);
               limitForce := m_limitForce - oldLimitForce;
            end;
         e_atUpperLimit:
            begin
               oldLimitForce := m_limitForce;
               m_limitForce := b2Min(m_limitForce + limitForce, 0.0);
               limitForce := m_limitForce - oldLimitForce;
            end;
      end;

      P := step.dt * limitForce;

      {$IFDEF OP_OVERLOAD}
      m_body1.m_linearVelocity.AddBy((invMass1 * P) * m_motorJacobian.linear1);
      m_body2.m_linearVelocity.AddBy((invMass2 * P) * m_motorJacobian.linear2);
      {$ELSE}
      AddBy(m_body1.m_linearVelocity, Multiply(m_motorJacobian.linear1, invMass1 * P));
      AddBy(m_body2.m_linearVelocity, Multiply(m_motorJacobian.linear2, invMass2 * P));
      {$ENDIF}
      m_body1.m_angularVelocity := m_body1.m_angularVelocity + invI1 * P * m_motorJacobian.angular1;
      m_body2.m_angularVelocity := m_body2.m_angularVelocity + invI2 * P * m_motorJacobian.angular2;
   end;
end;

function Tb2PrismaticJoint.SolvePositionConstraints: Boolean;
var
   invMass1, invMass2, invI1, invI2: Float;
   r1, r2, p1, p2, d, ay1, ax1: TVector2;
   linearC, linearImpulse, positionError, angularC, angularImpulse, angularError: Float;
   translation, limitImpulse, limitC, oldLimitImpulse: Float;
begin
   invMass1 := m_body1.m_invMass;
   invMass2 := m_body2.m_invMass;
   invI1 := m_body1.m_invI;
   invI2 := m_body2.m_invI;

   {$IFDEF OP_OVERLOAD}
   r1 := b2Mul(m_body1.m_xf.R, m_localAnchor1 - m_body1.GetLocalCenter);
   r2 := b2Mul(m_body2.m_xf.R, m_localAnchor2 - m_body2.GetLocalCenter);
   p1 := m_body1.m_sweep.c + r1;
   p2 := m_body2.m_sweep.c + r2;
   d := p2 - p1;
   {$ELSE}
   r1 := b2Mul(m_body1.m_xf.R, Subtract(m_localAnchor1, m_body1.GetLocalCenter));
   r2 := b2Mul(m_body2.m_xf.R, Subtract(m_localAnchor2, m_body2.GetLocalCenter));
   p1 := Add(m_body1.m_sweep.c, r1);
   p2 := Add(m_body2.m_sweep.c, r2);
   d := Subtract(p2, p1);
   {$ENDIF}
   ay1 := b2Mul(m_body1.m_xf.R, m_localYAxis1);

   // Solve linear (point-to-line) constraint.
   linearC := b2Dot(ay1, d);
   // Prevent overly large corrections.
   linearC := b2Clamp(linearC, -b2_maxLinearCorrection, b2_maxLinearCorrection);
   linearImpulse := -m_linearMass * linearC;

   {$IFDEF OP_OVERLOAD}
   m_body1.m_sweep.c.AddBy((invMass1 * linearImpulse) * m_linearJacobian.linear1);
   m_body2.m_sweep.c.AddBy((invMass2 * linearImpulse) * m_linearJacobian.linear2);
   {$ELSE}
   AddBy(m_body1.m_sweep.c, Multiply(m_linearJacobian.linear1, invMass1 * linearImpulse));
   AddBy(m_body2.m_sweep.c, Multiply(m_linearJacobian.linear2, invMass2 * linearImpulse));
   {$ENDIF}

   m_body1.m_sweep.a := m_body1.m_sweep.a + invI1 * linearImpulse * m_linearJacobian.angular1;
   //m_body1.SynchronizeTransform; // updated by angular constraint
   m_body2.m_sweep.a := m_body2.m_sweep.a + invI2 * linearImpulse * m_linearJacobian.angular2;
   //m_body2.SynchronizeTransform; // updated by angular constraint

   positionError := Abs(linearC);

   // Solve angular constraint.
   angularC := m_body2.m_sweep.a - m_body1.m_sweep.a - m_refAngle;
   // Prevent overly large corrections.
   angularC := b2Clamp(angularC, -b2_maxAngularCorrection, b2_maxAngularCorrection);
   angularImpulse := -m_angularMass * angularC;

   m_body1.m_sweep.a := m_body1.m_sweep.a - m_body1.m_invI * angularImpulse;
   m_body2.m_sweep.a := m_body2.m_sweep.a + m_body2.m_invI * angularImpulse;

   m_body1.SynchronizeTransform;
   m_body2.SynchronizeTransform;

   angularError := Abs(angularC);

   // Solve linear limit constraint.
   if m_enableLimit and (m_limitState <> e_inactiveLimit) then
   begin
      {$IFDEF OP_OVERLOAD}
      r1 := b2Mul(m_body1.m_xf.R, m_localAnchor1 - m_body1.GetLocalCenter);
      r2 := b2Mul(m_body2.m_xf.R, m_localAnchor2 - m_body2.GetLocalCenter);
      p1 := m_body1.m_sweep.c + r1;
      p2 := m_body2.m_sweep.c + r2;
      d := p2 - p1;
      {$ELSE}
      r1 := b2Mul(m_body1.m_xf.R, Subtract(m_localAnchor1, m_body1.GetLocalCenter));
      r2 := b2Mul(m_body2.m_xf.R, Subtract(m_localAnchor2, m_body2.GetLocalCenter));
      p1 := Add(m_body1.m_sweep.c, r1);
      p2 := Add(m_body2.m_sweep.c, r2);
      d := Subtract(p2, p1);
      {$ENDIF}
      ax1 := b2Mul(m_body1.m_xf.R, m_localXAxis1);

      translation := b2Dot(ax1, d);
      limitImpulse := 0.0;

      if m_limitState = e_equalLimits then
      begin
         // Prevent large angular corrections
         limitC := b2Clamp(translation, -b2_maxLinearCorrection, b2_maxLinearCorrection);
         limitImpulse := -m_motorMass * limitC;
         positionError := b2Max(positionError, Abs(angularC));
      end
      else if m_limitState = e_atLowerLimit then
      begin
         limitC := translation - m_lowerTranslation;
         positionError := b2Max(positionError, -limitC);

         // Prevent large linear corrections and allow some slop.
         limitC := b2Clamp(limitC + b2_linearSlop, -b2_maxLinearCorrection, 0.0);
         limitImpulse := -m_motorMass * limitC;
         oldLimitImpulse := m_limitPositionImpulse;
         m_limitPositionImpulse := b2Max(m_limitPositionImpulse + limitImpulse, 0.0);
         limitImpulse := m_limitPositionImpulse - oldLimitImpulse;
      end
      else if m_limitState = e_atUpperLimit then
      begin
         limitC := translation - m_upperTranslation;
         positionError := b2Max(positionError, limitC);

         // Prevent large linear corrections and allow some slop.
         limitC := b2Clamp(limitC - b2_linearSlop, 0.0, b2_maxLinearCorrection);
         limitImpulse := -m_motorMass * limitC;
         oldLimitImpulse := m_limitPositionImpulse;
         m_limitPositionImpulse := b2Min(m_limitPositionImpulse + limitImpulse, 0.0);
         limitImpulse := m_limitPositionImpulse - oldLimitImpulse;
      end;

      {$IFDEF OP_OVERLOAD}
      m_body1.m_sweep.c.AddBy((invMass1 * limitImpulse) * m_motorJacobian.linear1);
      m_body2.m_sweep.c.AddBy((invMass2 * limitImpulse) * m_motorJacobian.linear2);
      {$ELSE}
      AddBy(m_body1.m_sweep.c, Multiply(m_motorJacobian.linear1, invMass1 * limitImpulse));
      AddBy(m_body2.m_sweep.c, Multiply(m_motorJacobian.linear2, invMass2 * limitImpulse));
      {$ENDIF}

      m_body1.m_sweep.a := m_body1.m_sweep.a + invI1 * limitImpulse * m_motorJacobian.angular1;
      m_body2.m_sweep.a := m_body2.m_sweep.a + invI2 * limitImpulse * m_motorJacobian.angular2;

      m_body1.SynchronizeTransform;
      m_body2.SynchronizeTransform;
   end;

   Result := (positionError <= b2_linearSlop) and (angularError <= b2_angularSlop);
end;

function Tb2PrismaticJoint.GetJointTranslation: Float;
var
   d, axis: TVector2;
begin
   {$IFDEF OP_OVERLOAD}
   d := m_body2.GetWorldPoint(m_localAnchor2) - m_body1.GetWorldPoint(m_localAnchor1);
   {$ELSE}
   d := Subtract(m_body2.GetWorldPoint(m_localAnchor2), m_body1.GetWorldPoint(m_localAnchor1));
   {$ENDIF}
   axis := m_body1.GetWorldVector(m_localXAxis1);
   Result := b2Dot(d, axis);
end;

function Tb2PrismaticJoint.GetJointSpeed: Float;
var
   r1, r2, d, axis: TVector2;
   w1: Float;
begin
   {$IFDEF OP_OVERLOAD}
   r1 := b2Mul(m_body1.m_xf.R, m_localAnchor1 - m_body1.GetLocalCenter);
   r2 := b2Mul(m_body2.m_xf.R, m_localAnchor2 - m_body2.GetLocalCenter);
   d := (m_body2.m_sweep.c + r2) - (m_body1.m_sweep.c + r1);
   {$ELSE}
   r1 := b2Mul(m_body1.m_xf.R, Subtract(m_localAnchor1, m_body1.GetLocalCenter));
   r2 := b2Mul(m_body2.m_xf.R, Subtract(m_localAnchor2, m_body2.GetLocalCenter));
   d := Subtract(Add(m_body2.m_sweep.c, r2), Add(m_body1.m_sweep.c, r1));
   {$ENDIF}
   axis := m_body1.GetWorldVector(m_localXAxis1);

   w1 := m_body1.m_angularVelocity;

   {$IFDEF OP_OVERLOAD}
   Result := b2Dot(d, b2Cross(w1, axis)) +
      b2Dot(axis, m_body2.m_linearVelocity + b2Cross(m_body2.m_angularVelocity, r2) -
      m_body1.m_linearVelocity - b2Cross(w1, r1));
   {$ELSE}
   Result := b2Dot(d, b2Cross(w1, axis)) +
      b2Dot(axis, Subtract(Add(m_body2.m_linearVelocity, b2Cross(m_body2.m_angularVelocity, r2)),
      Add(m_body1.m_linearVelocity, b2Cross(w1, r1))));
   {$ENDIF}
end;

procedure Tb2PrismaticJoint.SetLimits(lower, upper: Float);
begin
   //b2Assert(lower <= upper);
	 m_lowerTranslation := lower;
	 m_upperTranslation := upper;
end;

{ Tb2MouseJointDef }

// p = attached point, m = mouse point
// C = p - m
// Cdot = v
//      = v + cross(w, r)
// J = [I r_skew]
// Identity used:
// w k % (rx i + ry j) = w * (-ry i + rx j)

constructor Tb2MouseJointDef.Create;
begin
		JointType := e_mouseJoint;
    SetZero(target);
		maxForce := 0.0;
		frequencyHz := 5.0;
		dampingRatio := 0.7;
		timeStep := 1.0 / 60.0;
end;

{ Tb2MouseJoint }

constructor Tb2MouseJoint.Create(def: Tb2MouseJointDef);
var
   omega, d, k: Float;
begin
   inherited Create(def);
   m_target := def.target;
   m_localAnchor := b2MulT(m_body2.m_xf, m_target);

   m_maxForce := def.maxForce;
   SetZero(m_impulse);

   omega := 2.0 * Pi * def.frequencyHz; // Frequency
   d := 2.0 * m_body2.m_mass * def.dampingRatio * omega; // Damping coefficient
   k := (def.timeStep * m_body2.m_mass) * (omega * omega); // Spring stiffness

   //b2Assert(d + k > B2_FLT_EPSILON);
   m_gamma := 1.0 / (d + k);
   m_beta := k / (d + k);
end;

function Tb2MouseJoint.GetAnchor1: TVector2;
begin
   Result := m_target;
end;

function Tb2MouseJoint.GetAnchor2: TVector2;
begin
   Result := m_body2.GetWorldPoint(m_localAnchor);
end;

function Tb2MouseJoint.GetReactionForce: TVector2;
begin
   Result := m_impulse;
end;

function Tb2MouseJoint.GetReactionTorque: Float;
begin
   Result := 0.0;
end;

procedure Tb2MouseJoint.InitVelocityConstraints(const step: Tb2TimeStep);
var
   r, P: TVector2;
   invMass, invI: Float;
   K1, K2, K: TMatrix22;
begin
   // Compute the effective mass matrix.
   {$IFDEF OP_OVERLOAD}
   r := b2Mul(m_body2.m_xf.R, m_localAnchor - m_body2.GetLocalCenter);
   {$ELSE}
   r := b2Mul(m_body2.m_xf.R, Subtract(m_localAnchor, m_body2.GetLocalCenter));
   {$ENDIF}

   // K    := [(1/m1 + 1/m2) * eye(2) - skew(r1) * invI1 * skew(r1) - skew(r2) * invI2 * skew(r2)]
   //      := [1/m1+1/m2     0    ] + invI1 * [r1.y*r1.y -r1.x*r1.y] + invI2 * [r1.y*r1.y -r1.x*r1.y]
   //        [    0     1/m1+1/m2]           [-r1.x*r1.y r1.x*r1.x]           [-r1.x*r1.y r1.x*r1.x]
   invMass := m_body2.m_invMass;
   invI := m_body2.m_invI;

   K1.col1.x := invMass;
   K1.col2.x := 0.0;
   K1.col1.y := 0.0;
   K1.col2.y := invMass;

   K2.col1.x :=  invI * r.y * r.y;
   K2.col2.x := -invI * r.x * r.y;
   K2.col1.y := K2.col2.x;
   K2.col2.y :=  invI * r.x * r.x;

   {$IFDEF OP_OVERLOAD}
   K := K1 + K2;
   {$ELSE}
   K := Add(K1, K2);
   {$ENDIF}
   K.col1.x := K.col1.x + m_gamma;
   K.col2.y := K.col2.y + m_gamma;

   {$IFDEF OP_OVERLOAD}
   m_mass := K.Invert;
   m_C := m_body2.m_sweep.c + r - m_target;
   {$ELSE}
   m_mass := Invert(K);
   m_C := Add(m_body2.m_sweep.c, r);
   SubtractBy(m_C, m_target);
   {$ENDIF}
   m_body2.m_angularVelocity := m_body2.m_angularVelocity * 0.98; // Cheat with some damping

   // Warm starting.
   {$IFDEF OP_OVERLOAD}
   P := step.dt * m_impulse;
   m_body2.m_linearVelocity.AddBy(invMass * P);
   {$ELSE}
   P := Multiply(m_impulse, step.dt);
   AddBy(m_body2.m_linearVelocity, Multiply(P, invMass));
   {$ENDIF}
   m_body2.m_angularVelocity := m_body2.m_angularVelocity + invI * b2Cross(r, P);
end;

procedure Tb2MouseJoint.SolveVelocityConstraints(const step: Tb2TimeStep);
var
   r, Cdot, force, oldForce, P: TVector2;
   forceMagnitude: Float;
begin
   oldForce := m_impulse;
   // Cdot := v + cross(w, r)
   {$IFDEF OP_OVERLOAD}
   r := b2Mul(m_body2.m_xf.R, m_localAnchor - m_body2.GetLocalCenter);
   Cdot := m_body2.m_linearVelocity + b2Cross(m_body2.m_angularVelocity, r);
   force := -step.inv_dt * b2Mul(m_mass, Cdot + (m_beta * step.inv_dt) * m_C +
      step.dt * (m_gamma * m_impulse));
   m_impulse.AddBy(force);
   forceMagnitude := m_impulse.Length;
   if forceMagnitude > m_maxForce then
      m_impulse.MultiplyBy(m_maxForce / forceMagnitude);
   force := m_impulse - oldForce;

   P := step.dt * force;
   m_body2.m_linearVelocity.AddBy(m_body2.m_invMass * P);
   {$ELSE}
   r := b2Mul(m_body2.m_xf.R, Subtract(m_localAnchor, m_body2.GetLocalCenter));
   Cdot := Add(m_body2.m_linearVelocity, b2Cross(m_body2.m_angularVelocity, r));
   force := Multiply(b2Mul(m_mass, Add(Cdot, Multiply(m_C, m_beta * step.inv_dt),
      Multiply(m_impulse, step.dt * m_gamma))), -step.inv_dt);
   AddBy(m_impulse, force);
   forceMagnitude := Length(m_impulse);
   if forceMagnitude > m_maxForce then
      MultiplyBy(m_impulse, m_maxForce / forceMagnitude);
   force := Subtract(m_impulse, oldForce);

   P := Multiply(force, step.dt);
   AddBy(m_body2.m_linearVelocity, Multiply(P, m_body2.m_invMass));
   {$ENDIF}

   m_body2.m_angularVelocity := m_body2.m_angularVelocity + m_body2.m_invI * b2Cross(r, P);
end;

function Tb2MouseJoint.SolvePositionConstraints: Boolean;
begin
   Result := True;
end;

procedure Tb2MouseJoint.SetTarget(const target: TVector2);
begin
   if m_body2.IsSleeping() then
      m_body2.WakeUp;
   m_target := target;
end;

{ Tb2PulleyJointDef }

// Pulley:
// length1 = norm(p1 - s1)
// length2 = norm(p2 - s2)
// C0 = (length1 + ratio * length2)_initial
// C = C0 - (length1 + ratio * length2) >= 0
// u1 = (p1 - s1) / norm(p1 - s1)
// u2 = (p2 - s2) / norm(p2 - s2)
// Cdot = -dot(u1, v1 + cross(w1, r1)) - ratio * dot(u2, v2 + cross(w2, r2))
// J = -[u1 cross(r1, u1) ratio * u2  ratio * cross(r2, u2)]
// K = J * invM * JT
//   = invMass1 + invI1 * cross(r1, u1)^2 + ratio^2 * (invMass2 + invI2 * cross(r2, u2)^2)
//
// Limit:
// C = maxLength - length
// u = (p - s) / norm(p - s)
// Cdot = -dot(u, v + cross(w, r))
// K = invMass + invI * cross(r, u)^2
// 0 <= impulse

const
   b2_minPulleyLength = 2.0;

constructor Tb2PulleyJointDef.Create;
begin
   JointType := e_pulleyJoint;
   {$IFDEF OP_OVERLOAD}
   groundAnchor1.SetValue(-1.0, 1.0);
   groundAnchor2.SetValue(1.0, 1.0);
   localAnchor1.SetValue(-1.0, 0.0);
   localAnchor2.SetValue(1.0, 0.0);
   {$ELSE}
   SetValue(groundAnchor1, -1.0, 1.0);
   SetValue(groundAnchor2, 1.0, 1.0);
   SetValue(localAnchor1, -1.0, 0.0);
   SetValue(localAnchor2, 1.0, 0.0);
   {$ENDIF}
   length1 := 0.0;
   maxLength1 := 0.0;
   length2 := 0.0;
   maxLength2 := 0.0;
   ratio := 1.0;
   collideConnected := True;
end;

procedure Tb2PulleyJointDef.Initialize(body1, body2: Tb2Body; const groundAnchor1,
   groundAnchor2, anchor1, anchor2: TVector2; ratio: Float);
var
   C: Float;
begin
   Self.body1 := body1;
   Self.body2 := body2;
   Self.groundAnchor1 := groundAnchor1;
   Self.groundAnchor2 := groundAnchor2;
   Self.localAnchor1 := body1.GetLocalPoint(anchor1);
   Self.localAnchor2 := body2.GetLocalPoint(anchor2);
   {$IFDEF OP_OVERLOAD}
   length1 := (anchor1 - groundAnchor1).Length;
   length2 := (anchor2 - groundAnchor2).Length;
   {$ELSE}
   length1 := Length(Subtract(anchor1, groundAnchor1));
   length2 := Length(Subtract(anchor2, groundAnchor2));
   {$ENDIF}
   Self.ratio := ratio;
   //b2Assert(ratio > B2_FLT_EPSILON);
   C := length1 + ratio * length2;
   maxLength1 := C - ratio * b2_minPulleyLength;
   maxLength2 := (C - b2_minPulleyLength) / ratio;
end;

{ Tb2PulleyJoint }

constructor Tb2PulleyJoint.Create(def: Tb2PulleyJointDef);
begin
   inherited Create(def);
   m_ground := m_body1.GetWorld.GetGroundBody;
   {$IFDEF OP_OVERLOAD}
   m_groundAnchor1 := def.groundAnchor1 - m_ground.m_xf.position;
   m_groundAnchor2 := def.groundAnchor2 - m_ground.m_xf.position;
   {$ELSE}
   m_groundAnchor1 := Subtract(def.groundAnchor1, m_ground.m_xf.position);
   m_groundAnchor2 := Subtract(def.groundAnchor2, m_ground.m_xf.position);
   {$ENDIF}
   m_localAnchor1 := def.localAnchor1;
   m_localAnchor2 := def.localAnchor2;

   //b2Assert(def.ratio != 0.0);
   m_ratio := def.ratio;
   m_constant := def.length1 + m_ratio * def.length2;

   m_maxLength1 := b2Min(def.maxLength1, m_constant - m_ratio * b2_minPulleyLength);
   m_maxLength2 := b2Min(def.maxLength2, (m_constant - b2_minPulleyLength) / m_ratio);

   m_force := 0.0;
   m_limitForce1 := 0.0;
   m_limitForce2 := 0.0;
end;

function Tb2PulleyJoint.GetAnchor1: TVector2;
begin
   Result := m_body1.GetWorldPoint(m_localAnchor1);
end;

function Tb2PulleyJoint.GetAnchor2: TVector2;
begin
   Result := m_body2.GetWorldPoint(m_localAnchor2);
end;

function Tb2PulleyJoint.GetReactionForce: TVector2;
begin
   {$IFDEF OP_OVERLOAD}
   Result := m_force * m_u2;
   {$ELSE}
   Result := Multiply(m_u2, m_force);
   {$ENDIF}
end;

function Tb2PulleyJoint.GetReactionTorque: Float;
begin
   Result := 0.0;
end;

procedure Tb2PulleyJoint.InitVelocityConstraints(const step: Tb2TimeStep);
var
   r1, r2, P1, P2: TVector2;
   length1, length2, C, cr1u1, cr2u2: Float;
begin
   {$IFDEF OP_OVERLOAD}
   r1 := b2Mul(m_body1.m_xf.R, m_localAnchor1 - m_body1.GetLocalCenter);
   r2 := b2Mul(m_body2.m_xf.R, m_localAnchor2 - m_body2.GetLocalCenter);

   // Get the pulley axes.
   m_u1 := m_body1.m_sweep.c + r1 - (m_ground.m_xf.position + m_groundAnchor1);
   m_u2 := m_body2.m_sweep.c + r2 - (m_ground.m_xf.position + m_groundAnchor2);

   length1 := m_u1.Length;
   length2 := m_u2.Length;

   if length1 > b2_linearSlop then
      m_u1.DivideBy(length1)
   else
      m_u1.SetZero;

   if length2 > b2_linearSlop then
      m_u2.DivideBy(length2)
   else
      m_u2.SetZero;
   {$ELSE}
   r1 := b2Mul(m_body1.m_xf.R, Subtract(m_localAnchor1, m_body1.GetLocalCenter));
   r2 := b2Mul(m_body2.m_xf.R, Subtract(m_localAnchor2, m_body2.GetLocalCenter));

   // Get the pulley axes.
   m_u1 := Subtract(Add(m_body1.m_sweep.c, r1), Add(m_ground.m_xf.position, m_groundAnchor1));
   m_u2 := Subtract(Add(m_body2.m_sweep.c, r2), Add(m_ground.m_xf.position, m_groundAnchor2));

   length1 := Length(m_u1);
   length2 := Length(m_u2);

   if length1 > b2_linearSlop then
      DivideBy(m_u1, length1)
   else
      SetZero(m_u1);

   if length2 > b2_linearSlop then
      DivideBy(m_u2, length2)
   else
      SetZero(m_u2);
   {$ENDIF}

   C := m_constant - length1 - m_ratio * length2;
   if C > 0.0 then
   begin
      m_state := e_inactiveLimit;
      m_force := 0.0;
   end
   else
   begin
      m_state := e_atUpperLimit;
      m_positionImpulse := 0.0;
   end;

   if length1 < m_maxLength1 then
   begin
      m_limitState1 := e_inactiveLimit;
      m_limitForce1 := 0.0;
   end
   else
   begin
      m_limitState1 := e_atUpperLimit;
      m_limitPositionImpulse1 := 0.0;
   end;

   if length2 < m_maxLength2 then
   begin
      m_limitState2 := e_inactiveLimit;
      m_limitForce2 := 0.0;
   end
   else
   begin
      m_limitState2 := e_atUpperLimit;
      m_limitPositionImpulse2 := 0.0;
   end;

   // Compute effective mass.
   cr1u1 := b2Cross(r1, m_u1);
   cr2u2 := b2Cross(r2, m_u2);

   m_limitMass1 := m_body1.m_invMass + m_body1.m_invI * cr1u1 * cr1u1;
   m_limitMass2 := m_body2.m_invMass + m_body2.m_invI * cr2u2 * cr2u2;
   m_pulleyMass := m_limitMass1 + m_ratio * m_ratio * m_limitMass2;
   //b2Assert(m_limitMass1 > B2_FLT_EPSILON);
   //b2Assert(m_limitMass2 > B2_FLT_EPSILON);
   //b2Assert(m_pulleyMass > B2_FLT_EPSILON);
   m_limitMass1 := 1.0 / m_limitMass1;
   m_limitMass2 := 1.0 / m_limitMass2;
   m_pulleyMass := 1.0 / m_pulleyMass;

   if step.warmStarting then
   begin
      // Warm starting.
      {$IFDEF OP_OVERLOAD}
      P1 := step.dt * (-m_force - m_limitForce1) * m_u1;
      P2 := step.dt * (-m_ratio * m_force - m_limitForce2) * m_u2;
      m_body1.m_linearVelocity.AddBy(m_body1.m_invMass * P1);
      m_body2.m_linearVelocity.AddBy(m_body2.m_invMass * P2);
      {$ELSE}
      P1 := Multiply(m_u1, -step.dt * (m_force + m_limitForce1));
      P2 := Multiply(m_u2, -step.dt * (m_ratio * m_force + m_limitForce2));
      AddBy(m_body1.m_linearVelocity, Multiply(P1, m_body1.m_invMass));
      AddBy(m_body2.m_linearVelocity, Multiply(P2, m_body2.m_invMass));
      {$ENDIF}
      m_body1.m_angularVelocity := m_body1.m_angularVelocity + m_body1.m_invI * b2Cross(r1, P1);
      m_body2.m_angularVelocity := m_body2.m_angularVelocity + m_body2.m_invI * b2Cross(r2, P2);
   end
   else
   begin
      m_force := 0.0;
      m_limitForce1 := 0.0;
      m_limitForce2 := 0.0;
   end;
end;

procedure Tb2PulleyJoint.SolveVelocityConstraints(const step: Tb2TimeStep);
var
   r1, r2, v1, v2, P1, P2: TVector2;
   Cdot, force, oldForce: Float;
begin
   {$IFDEF OP_OVERLOAD}
   r1 := b2Mul(m_body1.m_xf.R, m_localAnchor1 - m_body1.GetLocalCenter);
   r2 := b2Mul(m_body2.m_xf.R, m_localAnchor2 - m_body2.GetLocalCenter);
   {$ELSE}
   r1 := b2Mul(m_body1.m_xf.R, Subtract(m_localAnchor1, m_body1.GetLocalCenter));
   r2 := b2Mul(m_body2.m_xf.R, Subtract(m_localAnchor2, m_body2.GetLocalCenter));
   {$ENDIF}

   if m_state = e_atUpperLimit then
   begin
      {$IFDEF OP_OVERLOAD}
      v1 := m_body1.m_linearVelocity + b2Cross(m_body1.m_angularVelocity, r1);
      v2 := m_body2.m_linearVelocity + b2Cross(m_body2.m_angularVelocity, r2);
      {$ELSE}
      v1 := Add(m_body1.m_linearVelocity, b2Cross(m_body1.m_angularVelocity, r1));
      v2 := Add(m_body2.m_linearVelocity, b2Cross(m_body2.m_angularVelocity, r2));
      {$ENDIF}

      Cdot := -b2Dot(m_u1, v1) - m_ratio * b2Dot(m_u2, v2);
      force := -step.inv_dt * m_pulleyMass * Cdot;
      oldForce := m_force;
      m_force := b2Max(0.0, m_force + force);
      force := m_force - oldForce;

      {$IFDEF OP_OVERLOAD}
      P1 := -step.dt * force * m_u1;
      P2 := -step.dt * m_ratio * force * m_u2;
      m_body1.m_linearVelocity.AddBy(m_body1.m_invMass * P1);
      m_body2.m_linearVelocity.AddBy(m_body2.m_invMass * P2);
      {$ELSE}
      P1 := Multiply(m_u1, -step.dt * force);
      P2 := Multiply(m_u2, -step.dt * m_ratio * force);
      AddBy(m_body1.m_linearVelocity, Multiply(P1, m_body1.m_invMass));
      AddBy(m_body2.m_linearVelocity, Multiply(P2, m_body2.m_invMass));
      {$ENDIF}
      m_body1.m_angularVelocity := m_body1.m_angularVelocity + m_body1.m_invI * b2Cross(r1, P1);
      m_body2.m_angularVelocity := m_body2.m_angularVelocity + m_body2.m_invI * b2Cross(r2, P2);
   end;

   if m_limitState1 = e_atUpperLimit then
   begin
      {$IFDEF OP_OVERLOAD}
      v1 := m_body1.m_linearVelocity + b2Cross(m_body1.m_angularVelocity, r1);
      {$ELSE}
      v1 := Add(m_body1.m_linearVelocity, b2Cross(m_body1.m_angularVelocity, r1));
      {$ENDIF}

      Cdot := -b2Dot(m_u1, v1);
      force := -step.inv_dt * m_limitMass1 * Cdot;
      oldForce := m_limitForce1;
      m_limitForce1 := b2Max(0.0, m_limitForce1 + force);
      force := m_limitForce1 - oldForce;

      {$IFDEF OP_OVERLOAD}
      P1 := -step.dt * force * m_u1;
      m_body1.m_linearVelocity.AddBy(m_body1.m_invMass * P1);
      {$ELSE}
      P1 := Multiply(m_u1, -step.dt * force);
      AddBy(m_body1.m_linearVelocity, Multiply(P1, m_body1.m_invMass));
      {$ENDIF}
      m_body1.m_angularVelocity := m_body1.m_angularVelocity + m_body1.m_invI * b2Cross(r1, P1);
   end;

   if m_limitState2 = e_atUpperLimit then
   begin
      {$IFDEF OP_OVERLOAD}
      v2 := m_body2.m_linearVelocity + b2Cross(m_body2.m_angularVelocity, r2);
      {$ELSE}
      v2 := Add(m_body2.m_linearVelocity, b2Cross(m_body2.m_angularVelocity, r2));
      {$ENDIF}

      Cdot := -b2Dot(m_u2, v2);
      force := -step.inv_dt * m_limitMass2 * Cdot;
      oldForce := m_limitForce2;
      m_limitForce2 := b2Max(0.0, m_limitForce2 + force);
      force := m_limitForce2 - oldForce;

      {$IFDEF OP_OVERLOAD}
      P2 := -step.dt * force * m_u2;
      m_body2.m_linearVelocity.AddBy(m_body2.m_invMass * P2);
      {$ELSE}
      P2 := Multiply(m_u2, -step.dt * force);
      AddBy(m_body2.m_linearVelocity, Multiply(P2, m_body2.m_invMass));
      {$ENDIF}
      m_body2.m_angularVelocity := m_body2.m_angularVelocity + m_body2.m_invI * b2Cross(r2, P2);
   end;
end;

function Tb2PulleyJoint.SolvePositionConstraints: Boolean;
var
   s1, s2, r1, r2, P1, P2: TVector2;
   linearError, length1, length2, C, impulse, oldImpulse, oldLimitPositionImpulse: Float;
begin
   {$IFDEF OP_OVERLOAD}
   s1 := m_ground.m_xf.position + m_groundAnchor1;
   s2 := m_ground.m_xf.position + m_groundAnchor2;
   {$ELSE}
   s1 := Add(m_ground.m_xf.position, m_groundAnchor1);
   s2 := Add(m_ground.m_xf.position, m_groundAnchor2);
   {$ENDIF}

   linearError := 0.0;

   if m_state = e_atUpperLimit then
   begin
      {$IFDEF OP_OVERLOAD}
      r1 := b2Mul(m_body1.m_xf.R, m_localAnchor1 - m_body1.GetLocalCenter);
      r2 := b2Mul(m_body2.m_xf.R, m_localAnchor2 - m_body2.GetLocalCenter);

      // Get the pulley axes.
      m_u1 := m_body1.m_sweep.c + r1 - s1;
      m_u2 := m_body2.m_sweep.c + r2 - s2;

      length1 := m_u1.Length;
      length2 := m_u2.Length;

      if length1 > b2_linearSlop then
         m_u1.DivideBy(length1)
      else
         m_u1.SetZero;

      if length2 > b2_linearSlop then
         m_u2.DivideBy(length2)
      else
         m_u2.SetZero;
      {$ELSE}
      r1 := b2Mul(m_body1.m_xf.R, Subtract(m_localAnchor1, m_body1.GetLocalCenter));
      r2 := b2Mul(m_body2.m_xf.R, Subtract(m_localAnchor2, m_body2.GetLocalCenter));

      // Get the pulley axes.
      m_u1 := Subtract(Add(m_body1.m_sweep.c, r1), s1);
      m_u2 := Subtract(Add(m_body2.m_sweep.c, r2), s2);

      length1 := Length(m_u1);
      length2 := Length(m_u2);

      if length1 > b2_linearSlop then
         DivideBy(m_u1, length1)
      else
         SetZero(m_u1);

      if length2 > b2_linearSlop then
         DivideBy(m_u2, length2)
      else
         SetZero(m_u2);
      {$ENDIF}

      C := m_constant - length1 - m_ratio * length2;
      linearError := b2Max(linearError, -C);

      C := b2Clamp(C + b2_linearSlop, -b2_maxLinearCorrection, 0.0);
      impulse := -m_pulleyMass * C;
      oldImpulse := m_positionImpulse;
      m_positionImpulse := b2Max(0.0, m_positionImpulse + impulse);
      impulse := m_positionImpulse - oldImpulse;

      {$IFDEF OP_OVERLOAD}
      P1 := -impulse * m_u1;
      P2 := -m_ratio * impulse * m_u2;

      m_body1.m_sweep.c.AddBy(m_body1.m_invMass * P1);
      m_body2.m_sweep.c.AddBy(m_body2.m_invMass * P2);
      {$ELSE}
      P1 := Multiply(m_u1, -impulse);
      P2 := Multiply(m_u2, -m_ratio * impulse);

      AddBy(m_body1.m_sweep.c, Multiply(P1, m_body1.m_invMass));
      AddBy(m_body2.m_sweep.c, Multiply(P2, m_body2.m_invMass));
      {$ENDIF}
      m_body1.m_sweep.a := m_body1.m_sweep.a + m_body1.m_invI * b2Cross(r1, P1);
      m_body2.m_sweep.a := m_body2.m_sweep.a + m_body2.m_invI * b2Cross(r2, P2);

      m_body1.SynchronizeTransform;
      m_body2.SynchronizeTransform;
   end;

   if m_limitState1 = e_atUpperLimit then
   begin
      {$IFDEF OP_OVERLOAD}
      r1 := b2Mul(m_body1.m_xf.R, m_localAnchor1 - m_body1.GetLocalCenter);
      p1 := m_body1.m_sweep.c + r1;

      m_u1 := p1 - s1;
      length1 := m_u1.Length;

      if length1 > b2_linearSlop then
         m_u1.DivideBy(length1)
      else
         m_u1.SetZero;
      {$ELSE}
      r1 := b2Mul(m_body1.m_xf.R, Subtract(m_localAnchor1, m_body1.GetLocalCenter));
      p1 := Add(m_body1.m_sweep.c, r1);

      m_u1 := Subtract(p1, s1);
      length1 := Length(m_u1);

      if length1 > b2_linearSlop then
         DivideBy(m_u1, length1)
      else
         SetZero(m_u1);
      {$ENDIF}

      C := m_maxLength1 - length1;
      linearError := b2Max(linearError, -C);
      C := b2Clamp(C + b2_linearSlop, -b2_maxLinearCorrection, 0.0);
      impulse := -m_limitMass1 * C;
      oldLimitPositionImpulse := m_limitPositionImpulse1;
      m_limitPositionImpulse1 := b2Max(0.0, m_limitPositionImpulse1 + impulse);
      impulse := m_limitPositionImpulse1 - oldLimitPositionImpulse;

      {$IFDEF OP_OVERLOAD}
      P1 := -impulse * m_u1;
      m_body1.m_sweep.c.AddBy(m_body1.m_invMass * P1);
      {$ELSE}
      P1 := Multiply(m_u1, -impulse);
      AddBy(m_body1.m_sweep.c, Multiply(P1, m_body1.m_invMass));
      {$ENDIF}
      m_body1.m_sweep.a := m_body1.m_sweep.a + m_body1.m_invI * b2Cross(r1, P1);

      m_body1.SynchronizeTransform;
   end;

   if m_limitState2 = e_atUpperLimit then
   begin
      {$IFDEF OP_OVERLOAD}
      r2 := b2Mul(m_body2.m_xf.R, m_localAnchor2 - m_body2.GetLocalCenter);
      p2 := m_body2.m_sweep.c + r2;

      m_u2 := p2 - s2;
      length2 := m_u2.Length;

      if length2 > b2_linearSlop then
         m_u2.DivideBy(length2)
      else
         m_u2.SetZero;
      {$ELSE}
      r2 := b2Mul(m_body2.m_xf.R, Subtract(m_localAnchor2, m_body2.GetLocalCenter));
      p2 := Add(m_body2.m_sweep.c, r2);

      m_u2 := Subtract(p2, s2);
      length2 := Length(m_u2);

      if length2 > b2_linearSlop then
         DivideBy(m_u2, length2)
      else
         SetZero(m_u2);
      {$ENDIF}

      C := m_maxLength2 - length2;
      linearError := b2Max(linearError, -C);
      C := b2Clamp(C + b2_linearSlop, -b2_maxLinearCorrection, 0.0);
      impulse := -m_limitMass2 * C;
      oldLimitPositionImpulse := m_limitPositionImpulse2;
      m_limitPositionImpulse2 := b2Max(0.0, m_limitPositionImpulse2 + impulse);
      impulse := m_limitPositionImpulse2 - oldLimitPositionImpulse;

      {$IFDEF OP_OVERLOAD}
      P2 := -impulse * m_u2;
      m_body2.m_sweep.c.AddBy(m_body2.m_invMass * P2);
      {$ELSE}
      P2 := Multiply(m_u2, -impulse);
      AddBy(m_body2.m_sweep.c, Multiply(P2, m_body2.m_invMass));
      {$ENDIF}
      m_body2.m_sweep.a := m_body2.m_sweep.a + m_body2.m_invI * b2Cross(r2, P2);

      m_body2.SynchronizeTransform;
   end;

   Result := linearError < b2_linearSlop;
end;

function Tb2PulleyJoint.GetLength1: Float;
begin
   {$IFDEF OP_OVERLOAD}
   Result := (m_body1.GetWorldPoint(m_localAnchor1) - (m_ground.m_xf.position +
      m_groundAnchor1)).Length;
   {$ELSE}
   Result := Length(Subtract(m_body1.GetWorldPoint(m_localAnchor1),
      Add(m_ground.m_xf.position, m_groundAnchor1)));
   {$ENDIF}
end;

function Tb2PulleyJoint.GetLength2: Float;
begin
   {$IFDEF OP_OVERLOAD}
   Result := (m_body2.GetWorldPoint(m_localAnchor2) - (m_ground.m_xf.position +
      m_groundAnchor2)).Length;
   {$ELSE}
   Result := Length(Subtract(m_body2.GetWorldPoint(m_localAnchor2),
      Add(m_ground.m_xf.position, m_groundAnchor2)));
   {$ENDIF}
end;

function Tb2PulleyJoint.GetGroundAnchor1: TVector2;
begin
   {$IFDEF OP_OVERLOAD}
   Result := m_ground.m_xf.position + m_groundAnchor1;
   {$ELSE}
   Result := Add(m_ground.m_xf.position, m_groundAnchor1);
   {$ENDIF}
end;

function Tb2PulleyJoint.GetGroundAnchor2: TVector2;
begin
   {$IFDEF OP_OVERLOAD}
   Result := m_ground.m_xf.position + m_groundAnchor2;
   {$ELSE}
   Result := Add(m_ground.m_xf.position, m_groundAnchor2);
   {$ENDIF}
end;

{ Tb2RevoluteJointDef }

// Point-to-point constraint
// C = p2 - p1
// Cdot = v2 - v1
//      = v2 + cross(w2, r2) - v1 - cross(w1, r1)
// J = [-I -r1_skew I r2_skew ]
// Identity used:
// w k % (rx i + ry j) = w * (-ry i + rx j)

// Motor constraint
// Cdot = w2 - w1
// J = [0 0 -1 0 0 1]
// K = invI1 + invI2

constructor Tb2RevoluteJointDef.Create;
begin
   JointType := e_revoluteJoint;
   SetZero(localAnchor1);
   SetZero(localAnchor2);
   referenceAngle := 0.0;
   lowerAngle := 0.0;
   upperAngle := 0.0;
   maxMotorTorque := 0.0;
   motorSpeed := 0.0;
   enableLimit := False;
   enableMotor := False;
end;

procedure Tb2RevoluteJointDef.Initialize(body1, body2: Tb2Body; const anchor: TVector2);
begin
   Self.body1 := body1;
	 Self.body2 := body2;
	 localAnchor1 := body1.GetLocalPoint(anchor);
	 localAnchor2 := body2.GetLocalPoint(anchor);
	 referenceAngle := body2.GetAngle - body1.GetAngle;
end;

{ Tb2RevoluteJoint }

constructor Tb2RevoluteJoint.Create(def: Tb2RevoluteJointDef);
begin
   inherited Create(def);
   m_localAnchor1 := def.localAnchor1;
   m_localAnchor2 := def.localAnchor2;
   m_referenceAngle := def.referenceAngle;

   SetZero(m_pivotForce);
   m_motorForce := 0.0;
   m_limitForce := 0.0;
   m_limitPositionImpulse := 0.0;

   m_lowerAngle := def.lowerAngle;
   m_upperAngle := def.upperAngle;
   m_maxMotorTorque := def.maxMotorTorque;
   m_motorSpeed := def.motorSpeed;
   m_enableLimit := def.enableLimit;
   m_enableMotor := def.enableMotor;
end;

function Tb2RevoluteJoint.GetAnchor1: TVector2;
begin
   Result := m_body1.GetWorldPoint(m_localAnchor1);
end;

function Tb2RevoluteJoint.GetAnchor2: TVector2;
begin
   Result := m_body2.GetWorldPoint(m_localAnchor2);
end;

function Tb2RevoluteJoint.GetReactionForce: TVector2;
begin
   Result := m_pivotForce;
end;

function Tb2RevoluteJoint.GetReactionTorque: Float;
begin
   Result := m_limitForce;
end;

procedure Tb2RevoluteJoint.InitVelocityConstraints(const step: Tb2TimeStep);
var
   r1, r2: TVector2;
   K1, K2, K3, K: TMatrix22;
   jointAngle: Float;
   tmp: Float;
begin
   // Compute the effective mass matrix.
   {$IFDEF OP_OVERLOAD}
   r1 := b2Mul(m_body1.m_xf.R, m_localAnchor1 - m_body1.GetLocalCenter);
   r2 := b2Mul(m_body2.m_xf.R, m_localAnchor2 - m_body2.GetLocalCenter);
   {$ELSE}
   r1 := b2Mul(m_body1.m_xf.R, Subtract(m_localAnchor1, m_body1.GetLocalCenter));
   r2 := b2Mul(m_body2.m_xf.R, Subtract(m_localAnchor2, m_body2.GetLocalCenter));
   {$ENDIF}

   // K    := [(1/m1 + 1/m2) * eye(2) - skew(r1) * invI1 * skew(r1) - skew(r2) * invI2 * skew(r2)]
   //      := [1/m1+1/m2     0    ] + invI1 * [r1.y*r1.y -r1.x*r1.y] + invI2 * [r1.y*r1.y -r1.x*r1.y]
   //        [    0     1/m1+1/m2]           [-r1.x*r1.y r1.x*r1.x]           [-r1.x*r1.y r1.x*r1.x]

   K1.col1.x := m_body1.m_invMass + m_body2.m_invMass;
   K1.col2.x := 0.0;
   K1.col1.y := 0.0;
   K1.col2.y := K1.col1.x;

   tmp := m_body1.m_invI;
   K2.col1.x := tmp * r1.y * r1.y;
   tmp := tmp * r1.x;
   K2.col2.x := -tmp * r1.y;
   K2.col1.y := K2.col2.x;
   K2.col2.y := tmp * r1.x;

   tmp := m_body2.m_invI;
   K3.col1.x := tmp * r2.y * r2.y;
   tmp := tmp * r2.x;
   K3.col2.x := -tmp * r2.y;
   K3.col1.y := K3.col2.x;
   K3.col2.y := tmp * r2.x;

   {$IFDEF OP_OVERLOAD}
   K := K1 + K2 + K3;
   m_pivotMass := K.Invert;
   {$ELSE}
   K := Add(K1, K2, K3);
   m_pivotMass := Invert(K);
   {$ENDIF}
   m_motorMass := 1.0 / (m_body1.m_invI + m_body2.m_invI);

   if not m_enableMotor then
      m_motorForce := 0.0;

   if m_enableLimit then
   begin
      jointAngle := m_body2.m_sweep.a - m_body1.m_sweep.a - m_referenceAngle;
      if Abs(m_upperAngle - m_lowerAngle) < 2.0 * b2_angularSlop then
         m_limitState := e_equalLimits
      else if jointAngle <= m_lowerAngle then
      begin
         if m_limitState <> e_atLowerLimit then
            m_limitForce := 0.0;
         m_limitState := e_atLowerLimit;
      end
      else if jointAngle >= m_upperAngle then
      begin
         if m_limitState <> e_atUpperLimit then
            m_limitForce := 0.0;
         m_limitState := e_atUpperLimit;
      end
      else
      begin
         m_limitState := e_inactiveLimit;
         m_limitForce := 0.0;
      end;
   end
   else
      m_limitForce := 0.0;

   if step.warmStarting then
   begin
      {$IFDEF OP_OVERLOAD}
      m_body1.m_linearVelocity.SubtractBy(step.dt * m_body1.m_invMass * m_pivotForce);
      m_body2.m_linearVelocity.AddBy(step.dt * m_body2.m_invMass * m_pivotForce);
      {$ELSE}
      SubtractBy(m_body1.m_linearVelocity, Multiply(m_pivotForce, step.dt * m_body1.m_invMass));
      AddBy(m_body2.m_linearVelocity, Multiply(m_pivotForce, step.dt * m_body2.m_invMass));
      {$ENDIF}
      m_body1.m_angularVelocity := m_body1.m_angularVelocity - step.dt * m_body1.m_invI *
         (b2Cross(r1, m_pivotForce) + m_motorForce + m_limitForce);
      m_body2.m_angularVelocity := m_body2.m_angularVelocity + step.dt * m_body2.m_invI *
         (b2Cross(r2, m_pivotForce) + m_motorForce + m_limitForce);
   end
   else
   begin
      SetZero(m_pivotForce);
      m_motorForce := 0.0;
      m_limitForce := 0.0;
   end;

   m_limitPositionImpulse := 0.0;
end;

procedure Tb2RevoluteJoint.SolveVelocityConstraints(const step: Tb2TimeStep);
var
   r1, r2, pivotCdot, pivotForce, P: TVector2;
   motorCdot, motorForce, oldMotorForce, limitCdot, limitForce, oldLimitForce, tmp: Float;
begin
   {$IFDEF OP_OVERLOAD}
   r1 := b2Mul(m_body1.m_xf.R, m_localAnchor1 - m_body1.GetLocalCenter);
   r2 := b2Mul(m_body2.m_xf.R, m_localAnchor2 - m_body2.GetLocalCenter);

   // Solve point-to-point constraint
   pivotCdot := m_body2.m_linearVelocity + b2Cross(m_body2.m_angularVelocity, r2) -
      m_body1.m_linearVelocity - b2Cross(m_body1.m_angularVelocity, r1);
   pivotForce := -step.inv_dt * b2Mul(m_pivotMass, pivotCdot);
   m_pivotForce.AddBy(pivotForce);

   P := step.dt * pivotForce;
   m_body1.m_linearVelocity.SubtractBy(m_body1.m_invMass * P);
   m_body2.m_linearVelocity.AddBy(m_body2.m_invMass * P);
   {$ELSE}
   r1 := b2Mul(m_body1.m_xf.R, Subtract(m_localAnchor1, m_body1.GetLocalCenter));
   r2 := b2Mul(m_body2.m_xf.R, Subtract(m_localAnchor2, m_body2.GetLocalCenter));

   // Solve point-to-point constraint
   pivotCdot := Subtract(m_body2.m_linearVelocity, m_body1.m_linearVelocity);
   AddBy(pivotCdot, b2Cross(m_body2.m_angularVelocity, r2));
   SubtractBy(pivotCdot, b2Cross(m_body1.m_angularVelocity, r1));

   pivotForce := Multiply(b2Mul(m_pivotMass, pivotCdot), -step.inv_dt);
   AddBy(m_pivotForce, pivotForce);

   P := Multiply(pivotForce, step.dt);
   SubtractBy(m_body1.m_linearVelocity, Multiply(P, m_body1.m_invMass));
   AddBy(m_body2.m_linearVelocity, Multiply(P, m_body2.m_invMass));
   {$ENDIF}
   m_body1.m_angularVelocity := m_body1.m_angularVelocity - m_body1.m_invI * b2Cross(r1, P);
   m_body2.m_angularVelocity := m_body2.m_angularVelocity + m_body2.m_invI * b2Cross(r2, P);

   if m_enableMotor and (m_limitState <> e_equalLimits) then
   begin
      motorCdot := m_body2.m_angularVelocity - m_body1.m_angularVelocity - m_motorSpeed;
      motorForce := -step.inv_dt * m_motorMass * motorCdot;
      oldMotorForce := m_motorForce;
      m_motorForce := b2Clamp(m_motorForce + motorForce, -m_maxMotorTorque, m_maxMotorTorque);
      motorForce := m_motorForce - oldMotorForce;

      tmp := step.dt * motorForce;
      m_body1.m_angularVelocity := m_body1.m_angularVelocity - m_body1.m_invI * tmp;
      m_body2.m_angularVelocity := m_body2.m_angularVelocity + m_body2.m_invI * tmp;
   end;

   if m_enableLimit and (m_limitState <> e_inactiveLimit) then
   begin
      limitCdot := m_body2.m_angularVelocity - m_body1.m_angularVelocity;
      limitForce := -step.inv_dt * m_motorMass * limitCdot;

      if m_limitState = e_equalLimits then
         m_limitForce := m_limitForce + limitForce
      else if m_limitState = e_atLowerLimit then
      begin
         oldLimitForce := m_limitForce;
         m_limitForce := b2Max(m_limitForce + limitForce, 0.0);
         limitForce := m_limitForce - oldLimitForce;
      end
      else if m_limitState = e_atUpperLimit then
      begin
         oldLimitForce := m_limitForce;
         m_limitForce := b2Min(m_limitForce + limitForce, 0.0);
         limitForce := m_limitForce - oldLimitForce;
      end;

      tmp := step.dt * limitForce;
      m_body1.m_angularVelocity := m_body1.m_angularVelocity - m_body1.m_invI * tmp;
      m_body2.m_angularVelocity := m_body2.m_angularVelocity + m_body2.m_invI * tmp;
   end;
end;

function Tb2RevoluteJoint.SolvePositionConstraints: Boolean;
var
   positionError, angularError, angle, limitImpulse, limitC, oldLimitImpulse: Float;
   r1, r2, ptpC, impulse: TVector2;
   K1, K2, K3, K: TMatrix22;
   tmp: Float;
begin
   positionError := 0.0;

   // Solve point-to-point position error.
   {$IFDEF OP_OVERLOAD}
   r1 := b2Mul(m_body1.m_xf.R, m_localAnchor1 - m_body1.GetLocalCenter);
   r2 := b2Mul(m_body2.m_xf.R, m_localAnchor2 - m_body2.GetLocalCenter);

   ptpC := m_body2.m_sweep.c + r2 - (m_body1.m_sweep.c + r1);
   positionError := ptpC.Length;
   {$ELSE}
   r1 := b2Mul(m_body1.m_xf.R, Subtract(m_localAnchor1, m_body1.GetLocalCenter));
   r2 := b2Mul(m_body2.m_xf.R, Subtract(m_localAnchor2, m_body2.GetLocalCenter));

   ptpC := Subtract(Add(m_body2.m_sweep.c, r2), Add(m_body1.m_sweep.c, r1));
   positionError := Length(ptpC);
   {$ENDIF}

   // Prevent overly large corrections.
   //b2Vec2 dpMax(b2_maxLinearCorrection, b2_maxLinearCorrection);
   //ptpC := b2Clamp(ptpC, -dpMax, dpMax);

   K1.col1.x := m_body1.m_invMass + m_body2.m_invMass;
   K1.col2.x := 0.0;
   K1.col1.y := 0.0;
   K1.col2.y := K1.col1.x;

   tmp := m_body1.m_invI;
   K2.col1.x := tmp * r1.y * r1.y;
   tmp := tmp * r1.x;
   K2.col2.x := -tmp * r1.y;
   K2.col1.y := K2.col2.x;
   K2.col2.y := tmp * r1.x;

   tmp := m_body2.m_invI;
   K3.col1.x := tmp * r2.y * r2.y;
   tmp := tmp * r2.x;
   K3.col2.x := -tmp * r2.y;
   K3.col1.y := K3.col2.x;
   K3.col2.y := tmp * r2.x;

   {$IFDEF OP_OVERLOAD}
   K := K1 + K2 + K3;
   impulse := K.Solve(-ptpC);

   m_body1.m_sweep.c.SubtractBy(m_body1.m_invMass * impulse);
   m_body2.m_sweep.c.AddBy(m_body2.m_invMass * impulse);
   {$ELSE}
   K := Add(K1, K2, K3);
   impulse := Solve(K, Negative(ptpC));

   SubtractBy(m_body1.m_sweep.c, Multiply(impulse, m_body1.m_invMass));
   AddBy(m_body2.m_sweep.c, Multiply(impulse, m_body2.m_invMass));
   {$ENDIF}

   m_body1.m_sweep.a := m_body1.m_sweep.a - m_body1.m_invI * b2Cross(r1, impulse);
   m_body2.m_sweep.a := m_body2.m_sweep.a + m_body2.m_invI * b2Cross(r2, impulse);

   m_body1.SynchronizeTransform;
   m_body2.SynchronizeTransform;

   // Handle limits.
   angularError := 0.0;

   if m_enableLimit and (m_limitState <> e_inactiveLimit) then
   begin
      angle := m_body2.m_sweep.a - m_body1.m_sweep.a - m_referenceAngle;
      limitImpulse := 0.0;

      if m_limitState = e_equalLimits then
      begin
         // Prevent large angular corrections
         limitC := b2Clamp(angle, -b2_maxAngularCorrection, b2_maxAngularCorrection);
         limitImpulse := -m_motorMass * limitC;
         angularError := Abs(limitC);
      end
      else if m_limitState = e_atLowerLimit then
      begin
         limitC := angle - m_lowerAngle;
         angularError := b2Max(0.0, -limitC);

         // Prevent large angular corrections and allow some slop.
         limitC := b2Clamp(limitC + b2_angularSlop, -b2_maxAngularCorrection, 0.0);
         limitImpulse := -m_motorMass * limitC;
         oldLimitImpulse := m_limitPositionImpulse;
         m_limitPositionImpulse := b2Max(m_limitPositionImpulse + limitImpulse, 0.0);
         limitImpulse := m_limitPositionImpulse - oldLimitImpulse;
      end
      else if m_limitState = e_atUpperLimit then
      begin
         limitC := angle - m_upperAngle;
         angularError := b2Max(0.0, limitC);

         // Prevent large angular corrections and allow some slop.
         limitC := b2Clamp(limitC - b2_angularSlop, 0.0, b2_maxAngularCorrection);
         limitImpulse := -m_motorMass * limitC;
         oldLimitImpulse := m_limitPositionImpulse;
         m_limitPositionImpulse := b2Min(m_limitPositionImpulse + limitImpulse, 0.0);
         limitImpulse := m_limitPositionImpulse - oldLimitImpulse;
      end;

      m_body1.m_sweep.a := m_body1.m_sweep.a - m_body1.m_invI * limitImpulse;
      m_body2.m_sweep.a := m_body2.m_sweep.a + m_body2.m_invI * limitImpulse;

      m_body1.SynchronizeTransform;
      m_body2.SynchronizeTransform;
   end;

   Result := (positionError <= b2_linearSlop) and (angularError <= b2_angularSlop);
end;

function Tb2RevoluteJoint.GetJointAngle: Float;
begin
	 Result := m_body2.m_sweep.a - m_body1.m_sweep.a - m_referenceAngle;
end;

function Tb2RevoluteJoint.GetJointSpeed: Float;
begin
	 Result := m_body2.m_angularVelocity - m_body1.m_angularVelocity;
end;

{ Tb2GearJointDef }

// Gear Joint:
// C0 = (coordinate1 + ratio * coordinate2)_initial
// C = C0 - (cordinate1 + ratio * coordinate2) = 0
// Cdot = -(Cdot1 + ratio * Cdot2)
// J = -[J1 ratio * J2]
// K = J * invM * JT
//   = J1 * invM1 * J1T + ratio * ratio * J2 * invM2 * J2T
//
// Revolute:
// coordinate = rotation
// Cdot = angularVelocity
// J = [0 0 1]
// K = J * invM * JT = invI
//
// Prismatic:
// coordinate = dot(p - pg, ug)
// Cdot = dot(v + cross(w, r), ug)
// J = [ug cross(r, ug)]
// K = J * invM * JT = invMass + invI * cross(r, ug)^2

constructor Tb2GearJointDef.Create;
begin
	 JointType := e_gearJoint;
	 joint1 := nil;
	 joint2 := nil;
	 ratio := 1.0;
end;

{ Tb2GearJoint }

constructor Tb2GearJoint.Create(def: Tb2GearJointDef);
var
   type1, type2: Tb2JointType;
   coordinate1, coordinate2: Float;
begin
   inherited Create(def);
   type1 := def.joint1.GetType;
   type2 := def.joint2.GetType;

   //b2Assert(type1 == e_revoluteJoint || type1 == e_prismaticJoint);
   //b2Assert(type2 == e_revoluteJoint || type2 == e_prismaticJoint);
   //b2Assert(def.joint1.GetBody1().IsStatic());
   //b2Assert(def.joint2.GetBody1().IsStatic());

   m_revolute1 := nil;
   m_prismatic1 := nil;
   m_revolute2 := nil;
   m_prismatic2 := nil;

   m_ground1 := def.joint1.GetBody1;
   m_body1 := def.joint1.GetBody2;
   if type1 = e_revoluteJoint then
   begin
      m_revolute1 := Tb2RevoluteJoint(def.joint1);
      m_groundAnchor1 := m_revolute1.m_localAnchor1;
      m_localAnchor1 := m_revolute1.m_localAnchor2;
      coordinate1 := m_revolute1.GetJointAngle;
   end
   else
   begin
      m_prismatic1 := Tb2PrismaticJoint(def.joint1);
      m_groundAnchor1 := m_prismatic1.m_localAnchor1;
      m_localAnchor1 := m_prismatic1.m_localAnchor2;
      coordinate1 := m_prismatic1.GetJointTranslation;
   end;

   m_ground2 := def.joint2.GetBody1;
   m_body2 := def.joint2.GetBody2;
   if type2 = e_revoluteJoint then
   begin
      m_revolute2 := Tb2RevoluteJoint(def.joint2);
      m_groundAnchor2 := m_revolute2.m_localAnchor1;
      m_localAnchor2 := m_revolute2.m_localAnchor2;
      coordinate2 := m_revolute2.GetJointAngle;
   end
   else
   begin
      m_prismatic2 := Tb2PrismaticJoint(def.joint2);
      m_groundAnchor2 := m_prismatic2.m_localAnchor1;
      m_localAnchor2 := m_prismatic2.m_localAnchor2;
      coordinate2 := m_prismatic2.GetJointTranslation;
   end;

   m_ratio := def.ratio;
   m_constant := coordinate1 + m_ratio * coordinate2;
   m_force := 0.0;
end;

function Tb2GearJoint.GetAnchor1: TVector2;
begin
   Result := m_body1.GetWorldPoint(m_localAnchor1);
end;

function Tb2GearJoint.GetAnchor2: TVector2;
begin
   Result := m_body2.GetWorldPoint(m_localAnchor2);
end;

function Tb2GearJoint.GetReactionForce: TVector2;
begin
	 // TODO_ERIN not tested
   {$IFDEF OP_OVERLOAD}
	 Result := m_force * m_J.linear2;
   {$ELSE}
   Result := Multiply(m_J.linear2, m_force);
   {$ENDIF}
end;

function Tb2GearJoint.GetReactionTorque: Float;
var
   r, F: TVector2;
begin
   // TODO_ERIN not tested
   {$IFDEF OP_OVERLOAD}
   r := b2Mul(m_body2.m_xf.R, m_localAnchor2 - m_body2.GetLocalCenter);
   F := m_force * m_J.linear2;
   {$ELSE}
   r := b2Mul(m_body2.m_xf.R, Subtract(m_localAnchor2, m_body2.GetLocalCenter));
   F := Multiply(m_J.linear2, m_force);
   {$ENDIF}
   Result := m_force * m_J.angular2 - b2Cross(r, F);
end;

procedure Tb2GearJoint.InitVelocityConstraints(const step: Tb2TimeStep);
var
   K, crug, P: Float;
   ug, r: TVector2;
begin
   K := 0.0;
   {$IFDEF OP_OVERLOAD}
   m_J.SetZero;
   {$ELSE}
   SetZero(m_J);
   {$ENDIF}

   if Assigned(m_revolute1) then
   begin
      m_J.angular1 := -1.0;
      K := K + m_body1.m_invI;
   end
   else
   begin
      ug := b2Mul(m_ground1.m_xf.R, m_prismatic1.m_localXAxis1);
      {$IFDEF OP_OVERLOAD}
      r := b2Mul(m_body1.m_xf.R, m_localAnchor1 - m_body1.GetLocalCenter);
      m_J.linear1 := -ug;
      {$ELSE}
      r := b2Mul(m_body1.m_xf.R, Subtract(m_localAnchor1, m_body1.GetLocalCenter));
      m_J.linear1 := Negative(ug);
      {$ENDIF}
      crug := b2Cross(r, ug);
      m_J.angular1 := -crug;
      K := K + m_body1.m_invMass + m_body1.m_invI * crug * crug;
   end;

   if Assigned(m_revolute2) then
   begin
      m_J.angular2 := -m_ratio;
      K := K + m_ratio * m_ratio * m_body2.m_invI;
   end
   else
   begin
      ug := b2Mul(m_ground2.m_xf.R, m_prismatic2.m_localXAxis1);
      {$IFDEF OP_OVERLOAD}
      r := b2Mul(m_body2.m_xf.R, m_localAnchor2 - m_body2.GetLocalCenter);
      m_J.linear2 := -m_ratio * ug;
      {$ELSE}
      r := b2Mul(m_body2.m_xf.R, Subtract(m_localAnchor2, m_body2.GetLocalCenter));
      m_J.linear2 := Multiply(ug, -m_ratio);
      {$ENDIF}
      crug := b2Cross(r, ug);
      m_J.angular2 := -m_ratio * crug;
      K := K + m_ratio * m_ratio * (m_body2.m_invMass + m_body2.m_invI * crug * crug);
   end;

   // Compute effective mass.
   //b2Assert(K > 0.0);
   m_mass := 1.0 / K;

   if step.warmStarting then
   begin
      // Warm starting.
      P := step.dt * m_force;
      {$IFDEF OP_OVERLOAD}
      m_body1.m_linearVelocity.AddBy(m_body1.m_invMass * P * m_J.linear1);
      m_body2.m_linearVelocity.AddBy(m_body2.m_invMass * P * m_J.linear2);
      {$ELSE}
      AddBy(m_body1.m_linearVelocity, Multiply(m_J.linear1, m_body1.m_invMass * P));
      AddBy(m_body2.m_linearVelocity, Multiply(m_J.linear2, m_body2.m_invMass * P));
      {$ENDIF}
      m_body1.m_angularVelocity := m_body1.m_angularVelocity + m_body1.m_invI * P * m_J.angular1;
      m_body2.m_angularVelocity := m_body2.m_angularVelocity + m_body2.m_invI * P * m_J.angular2;
   end
   else
      m_force := 0.0;
end;

procedure Tb2GearJoint.SolveVelocityConstraints(const step: Tb2TimeStep);
var
   Cdot, force, P: Float;
begin
   {$IFDEF OP_OVERLOAD}
   Cdot := m_J.Compute(m_body1.m_linearVelocity, m_body2.m_linearVelocity,
   m_body1.m_angularVelocity, m_body2.m_angularVelocity);
   {$ELSE}
   Cdot := Compute(m_J,	m_body1.m_linearVelocity, m_body2.m_linearVelocity,
   m_body1.m_angularVelocity, m_body2.m_angularVelocity);
   {$ENDIF}

   force := -step.inv_dt * m_mass * Cdot;
   m_force := m_force + force;

   P := step.dt * force;
   {$IFDEF OP_OVERLOAD}
   m_body1.m_linearVelocity.AddBy(m_body1.m_invMass * P * m_J.linear1);
   m_body2.m_linearVelocity.AddBy(m_body2.m_invMass * P * m_J.linear2);
   {$ELSE}
   AddBy(m_body1.m_linearVelocity, Multiply(m_J.linear1, m_body1.m_invMass * P));
   AddBy(m_body2.m_linearVelocity, Multiply(m_J.linear2, m_body2.m_invMass * P));
   {$ENDIF}
   m_body1.m_angularVelocity := m_body1.m_angularVelocity + m_body1.m_invI * P * m_J.angular1;
   m_body2.m_angularVelocity := m_body2.m_angularVelocity + m_body2.m_invI * P * m_J.angular2;
end;

function Tb2GearJoint.SolvePositionConstraints: Boolean;
var
   linearError, coordinate1, coordinate2, C, impulse: Float;
begin
   linearError := 0.0;

   if Assigned(m_revolute1) then
      coordinate1 := m_revolute1.GetJointAngle()
   else
      coordinate1 := m_prismatic1.GetJointTranslation;

   if Assigned(m_revolute2) then
      coordinate2 := m_revolute2.GetJointAngle()
   else
      coordinate2 := m_prismatic2.GetJointTranslation;

   C := m_constant - (coordinate1 + m_ratio * coordinate2);
   impulse := -m_mass * C;

   {$IFDEF OP_OVERLOAD}
   m_body1.m_sweep.c.AddBy(m_body1.m_invMass * impulse * m_J.linear1);
   m_body2.m_sweep.c.AddBy(m_body2.m_invMass * impulse * m_J.linear2);
   {$ELSE}
   AddBy(m_body1.m_sweep.c, Multiply(m_J.linear1, m_body1.m_invMass * impulse));
   AddBy(m_body2.m_sweep.c, Multiply(m_J.linear2, m_body2.m_invMass * impulse));
   {$ENDIF}
   m_body1.m_sweep.a := m_body1.m_sweep.a + m_body1.m_invI * impulse * m_J.angular1;
   m_body2.m_sweep.a := m_body2.m_sweep.a + m_body2.m_invI * impulse * m_J.angular2;

   m_body1.SynchronizeTransform;
   m_body2.SynchronizeTransform;

   Result := linearError < b2_linearSlop;
end;

{ Tb2FixedJointDef }

constructor Tb2FixedJointDef.Create;
begin
	 JointType := e_fixedJoint;
end;

procedure Tb2FixedJointDef.Initialize(body1, body2: Tb2Body);
begin
   Self.body1 := body1;
	 Self.body2 := body2;
end;

{ Tb2FixedJoint }

constructor Tb2FixedJoint.Create(def: Tb2FixedJointDef);
begin
   inherited Create(def);
   // Get initial delta position and angle
   {$IFDEF OP_OVERLOAD}
   m_dp := b2MulT(m_body1.m_xf.R, m_body2.m_xf.position - m_body1.m_xf.position);
   {$ELSE}
   m_dp := b2MulT(m_body1.m_xf.R, Subtract(m_body2.m_xf.position, m_body1.m_xf.position));
   {$ENDIF}
   m_a := m_body2.GetAngle - m_body1.GetAngle;
   m_R0 := b2MulT(m_body1.m_xf.R, m_body2.m_xf.R);

   // Reset accumulated lambda
   m_lambda[0] := 0.0;
   m_lambda[1] := 0.0;
   m_lambda[2] := 0.0;
end;

procedure Tb2FixedJoint.CalculateMC;
var
   invM12, invI1, a, b, c, d, e, f, c1, c2, c3, den: Float;
begin
   UPhysics2DTypes.SinCos(m_body1.m_sweep.a, m_s, m_c);

   // Calculate vector A w1 := d/dt (R(a1) d)
   m_Ax := -m_s * m_d.x - m_c * m_d.y;
   m_Ay := m_c * m_d.x - m_s * m_d.y;

   // Calculate effective constraint mass: mC := (J M^-1 J^T)^-1
   invM12 := m_body1.m_invMass + m_body2.m_invMass;
   invI1 := m_body1.m_invI;
   a := invM12 + m_Ax * m_Ax * invI1;
   b := m_Ax * m_Ay * invI1;
   c := m_Ax * invI1;
   d := invM12 + m_Ay * m_Ay * invI1;
   e := m_Ay * invI1;
   f := m_body1.m_invI + invI1;
   c1 := d * f - e * e;
   c2 := c * e - b * f;
   c3 := b * e - c * d;
   den := a * c1 + b * c2 + c * c3;
   m_mc[0][0] := c1 / den;
   m_mc[1][0] := c2 / den;
   m_mc[2][0] := c3 / den;
   m_mc[0][1] := m_mc[1][0];
   m_mc[1][1] := (a * f - c * c ) / den;
   m_mc[2][1] := (b * c - a * e ) / den;
   m_mc[0][2] := m_mc[2][0];
   m_mc[1][2] := m_mc[2][1];
   m_mc[2][2] := (a * d - b * b ) / den;
end;

function Tb2FixedJoint.GetAnchor1: TVector2;
begin
	 // Return arbitrary position (we have to implement this abstract virtual function)
	 Result := m_body1.GetWorldCenter;
end;

function Tb2FixedJoint.GetAnchor2: TVector2;
begin
	 // Return arbitrary position (we have to implement this abstract virtual function)
	 Result := m_body2.GetWorldCenter;
end;

function Tb2FixedJoint.GetReactionForce: TVector2;
begin
   {$IFDEF OP_OVERLOAD}
   Result := m_inv_dt * MakeVector(m_lambda[0], m_lambda[1]);
   {$ELSE}
   Result := Multiply(MakeVector(m_lambda[0], m_lambda[1]), m_inv_dt);
   {$ENDIF}
end;

function Tb2FixedJoint.GetReactionTorque: Float;
begin
   Result := m_inv_dt * m_lambda[2];
end;

procedure Tb2FixedJoint.InitVelocityConstraints(const step: Tb2TimeStep);
begin
   // Store step
   m_inv_dt := step.inv_dt;

   // Get d for this step (transform from delta between positions to delta between center of masses)
   {$IFDEF OP_OVERLOAD}
   m_d := m_dp - m_body1.m_sweep.localCenter + b2Mul(m_R0, m_body2.m_sweep.localCenter);
   {$ELSE}
   m_d := Add(Subtract(m_dp, m_body1.m_sweep.localCenter), b2Mul(m_R0, m_body2.m_sweep.localCenter));
   {$ENDIF}

   // Calculate effective joint mass (stays constant during velocity solving)
   CalculateMC;

   if step.warmStarting then
   begin
      // Apply initial impulse
      with m_body1 do
      begin
         m_linearVelocity.x := m_linearVelocity.x - m_invMass * m_lambda[0];
         m_linearVelocity.y := m_linearVelocity.y - m_invMass * m_lambda[1];
         m_angularVelocity := m_angularVelocity - m_invI * (m_lambda[0] * m_Ax + m_lambda[1] * m_Ay + m_lambda[2]);
      end;
      with m_body2 do
      begin
         m_linearVelocity.x := m_linearVelocity.x + m_invMass * m_lambda[0];
         m_linearVelocity.y := m_linearVelocity.y + m_invMass * m_lambda[1];
         m_angularVelocity := m_angularVelocity + m_invI * m_lambda[2];
      end;
   end
   else
   begin
      // Reset accumulated lambda
      m_lambda[0] := 0.0;
      m_lambda[1] := 0.0;
      m_lambda[2] := 0.0;
   end;
end;

procedure Tb2FixedJoint.SolveVelocityConstraints(const step: Tb2TimeStep);
var
   i: Integer;
   Cdot, lambda: array[0..2] of Float;
begin
   // Assert that angle is still the same so the effective joint mass is still valid
   //assert(m_body1.m_sweep.a == m_a1);

   // Calculate Cdot
   Cdot[0] := m_body2.m_linearVelocity.x - m_body1.m_linearVelocity.x - m_Ax * m_body1.m_angularVelocity;
   Cdot[1] := m_body2.m_linearVelocity.y - m_body1.m_linearVelocity.y - m_Ay * m_body1.m_angularVelocity;
   Cdot[2] := m_body2.m_angularVelocity - m_body1.m_angularVelocity;

   // Calculate lambda
   for i := 0 to 2 do
      lambda[i] := -(m_mc[i][0] * Cdot[0] + m_mc[i][1] * Cdot[1] + m_mc[i][2] * Cdot[2]);

   // Apply impulse
   with m_body1 do
   begin
      m_linearVelocity.x := m_linearVelocity.x - m_invMass * lambda[0];
      m_linearVelocity.y := m_linearVelocity.y - m_invMass * lambda[1];
      m_angularVelocity := m_angularVelocity - m_invI * (lambda[0] * m_Ax + lambda[1] * m_Ay + lambda[2]);
   end;
   with m_body2 do
   begin
      m_linearVelocity.x := m_linearVelocity.x + m_invMass * lambda[0];
      m_linearVelocity.y := m_linearVelocity.y + m_invMass * lambda[1];
      m_angularVelocity := m_angularVelocity + m_invI * lambda[2];
   end;

   // Accumulate total lambda
   for i := 0 to 2 do
     m_lambda[i] := m_lambda[i] + lambda[i];
end;

function Tb2FixedJoint.SolvePositionConstraints: Boolean;
var
   i: Integer;
   C, lambda: array[0..2] of Float;
begin
   // Recalculate effective constraint mass if angle changed enough
   if Abs(m_body1.m_sweep.a - m_a1) > 1e-3 then
     CalculateMC;

   // Calculate C
   C[0] := m_body2.m_sweep.c.x - m_body1.m_sweep.c.x - m_c * m_d.x + m_s * m_d.y;
   C[1] := m_body2.m_sweep.c.y - m_body1.m_sweep.c.y - m_s * m_d.x - m_c * m_d.y;
   C[2] := m_body2.m_sweep.a - m_a1 - m_a;

   // Calculate lambda
   for i := 0 to 2 do
     lambda[i] := -(m_mc[i][0] * C[0] + m_mc[i][1] * C[1] + m_mc[i][2] * C[2]);

   // Apply impulse
   with m_body1, m_sweep do
   begin
      c.x := c.x - m_invMass * lambda[0];
      c.y := c.y - m_invMass * lambda[1];
      a := a - m_invI * (lambda[0] * m_Ax + lambda[1] * m_Ay + lambda[2]);
   end;
   with m_body2, m_sweep do
   begin
      c.x := c.x + m_invMass * lambda[0];
      c.y := c.y + m_invMass * lambda[1];
      a := a + m_invI * lambda[2];
   end;

   // Push the changes to the transforms
   m_body1.SynchronizeTransform;
   m_body2.SynchronizeTransform;

   // Constraint is satisfied if all constraint equations are nearly zero
   Result := (Abs(C[0]) < b2_linearSlop) and (Abs(C[1]) < b2_linearSlop) and (Abs(C[2]) < b2_angularSlop);
end;

initialization
   b2_defaultFilter := Tb2ContactFilter.Create;
   g_GJK_Iterations := 0;

{$IFDEF CLASSVAR_AVAIL}
   Tb2BroadPhase.s_validate := False;
{$ELSE}
   b2BroadPhase_s_validate := False;
{$ENDIF}

finalization
   b2_defaultFilter.Free;

{$IFDEF DEBUG_LOG}
   DEBUG.Free;
{$ENDIF}
end.

